@@ -1,6 +1,6 @@
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008, 2010 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -8,23 +8,23 @@
 import math
 import os.path
 from urllib.parse import unquote
-import cairo
+from cairo import OPERATOR_OVER, LINE_CAP_ROUND #pylint: disable=no-name-in-module
 from gi.repository import Gtk, Gdk, Gio
 
 from xpra.os_util import bytestostr, strtobytes, is_X11, monotonic_time, WIN32, OSX, POSIX
 from xpra.util import (
-    AdHocStruct, typedict, envint, envbool, nonl, csv, first_time,
+    AdHocStruct, typedict, envint, envbool, csv, first_time,
     WORKSPACE_UNSET, WORKSPACE_ALL, WORKSPACE_NAMES, MOVERESIZE_DIRECTION_STRING, SOURCE_INDICATION_STRING,
     MOVERESIZE_CANCEL,
     MOVERESIZE_SIZE_TOPLEFT, MOVERESIZE_SIZE_TOP, MOVERESIZE_SIZE_TOPRIGHT,
     MOVERESIZE_SIZE_RIGHT,
     MOVERESIZE_SIZE_BOTTOMRIGHT,  MOVERESIZE_SIZE_BOTTOM, MOVERESIZE_SIZE_BOTTOMLEFT,
-    MOVERESIZE_SIZE_LEFT, MOVERESIZE_MOVE,
+    MOVERESIZE_SIZE_LEFT, MOVERESIZE_MOVE, MOVERESIZE_MOVE_KEYBOARD,
     )
 from xpra.gtk_common.gobject_util import no_arg_signal, one_arg_signal
 from xpra.gtk_common.gtk_util import (
     get_pixbuf_from_data, get_default_root_window,
-    enable_alpha,
+    set_visual,
     BUTTON_MASK,
     GRAB_STATUS_STRING,
     WINDOW_EVENT_MASK,
@@ -39,6 +39,7 @@ focuslog = Logger("focus", "grab")
 workspacelog = Logger("workspace")
 log = Logger("window")
 keylog = Logger("keyboard")
+keyeventlog = Logger("keyboard", "events")
 iconlog = Logger("icon")
 metalog = Logger("metadata")
 statelog = Logger("state")
@@ -92,8 +93,10 @@ if USE_X11_BINDINGS:
         log.error(" %s", e)
 
 
+AWT_DIALOG_WORKAROUND = envbool("XPRA_AWT_DIALOG_WORKAROUND", WIN32)
 BREAK_MOVERESIZE = os.environ.get("XPRA_BREAK_MOVERESIZE", "Escape").split(",")
 MOVERESIZE_X11 = envbool("XPRA_MOVERESIZE_X11", POSIX)
+MOVERESIZE_GDK = envbool("XPRA_MOVERESIZE_GDK", True)
 CURSOR_IDLE_TIMEOUT = envint("XPRA_CURSOR_IDLE_TIMEOUT", 6)
 DISPLAY_HAS_SCREEN_INDEX = POSIX and os.environ.get("DISPLAY", "").split(":")[-1].find(".")>=0
 DRAGNDROP = envbool("XPRA_DRAGNDROP", True)
@@ -104,7 +107,6 @@ REFRESH_MAXIMIZED = envbool("XPRA_REFRESH_MAXIMIZED", True)
 
 WINDOW_OVERFLOW_TOP = envbool("XPRA_WINDOW_OVERFLOW_TOP", False)
 AWT_RECENTER = envbool("XPRA_AWT_RECENTER", True)
-SAVE_WINDOW_ICONS = envbool("XPRA_SAVE_WINDOW_ICONS", False)
 UNDECORATED_TRANSIENT_IS_OR = envint("XPRA_UNDECORATED_TRANSIENT_IS_OR", 1)
 XSHAPE = envbool("XPRA_XSHAPE", True)
 LAZY_SHAPE = envbool("XPRA_LAZY_SHAPE", True)
@@ -163,6 +165,39 @@ GDK_SCROLL_MAP = {
     Gdk.ScrollDirection.RIGHT    : 7,
     }
 
+OR_TYPE_HINTS = (
+    Gdk.WindowTypeHint.DIALOG,
+    Gdk.WindowTypeHint.MENU,
+    Gdk.WindowTypeHint.TOOLBAR,
+    #Gdk.WindowTypeHint.SPLASHSCREEN,
+    #Gdk.WindowTypeHint.UTILITY,
+    #Gdk.WindowTypeHint.DOCK,
+    #Gdk.WindowTypeHint.DESKTOP,
+    Gdk.WindowTypeHint.DROPDOWN_MENU,
+    Gdk.WindowTypeHint.POPUP_MENU,
+    Gdk.WindowTypeHint.TOOLTIP,
+    #Gdk.WindowTypeHint.NOTIFICATION,
+    Gdk.WindowTypeHint.COMBO,
+    Gdk.WindowTypeHint.DND,
+    )
+
+WINDOW_NAME_TO_HINT = {
+    "NORMAL"        : Gdk.WindowTypeHint.NORMAL,
+    "DIALOG"        : Gdk.WindowTypeHint.DIALOG,
+    "MENU"          : Gdk.WindowTypeHint.MENU,
+    "TOOLBAR"       : Gdk.WindowTypeHint.TOOLBAR,
+    "SPLASH"        : Gdk.WindowTypeHint.SPLASHSCREEN,
+    "UTILITY"       : Gdk.WindowTypeHint.UTILITY,
+    "DOCK"          : Gdk.WindowTypeHint.DOCK,
+    "DESKTOP"       : Gdk.WindowTypeHint.DESKTOP,
+    "DROPDOWN_MENU" : Gdk.WindowTypeHint.DROPDOWN_MENU,
+    "POPUP_MENU"    : Gdk.WindowTypeHint.POPUP_MENU,
+    "TOOLTIP"       : Gdk.WindowTypeHint.TOOLTIP,
+    "NOTIFICATION"  : Gdk.WindowTypeHint.NOTIFICATION,
+    "COMBO"         : Gdk.WindowTypeHint.COMBO,
+    "DND"           : Gdk.WindowTypeHint.DND
+    }
+
 
 def wn(w):
     return WORKSPACE_NAMES.get(w, w)
@@ -193,15 +228,16 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
             window_type = Gtk.WindowType.TOPLEVEL
         self.on_realize_cb = {}
         Gtk.Window.__init__(self, type = window_type)
+        self.set_app_paintable(True)
         self.init_drawing_area()
         self.set_decorated(self._is_decorated(metadata))
-        self.set_app_paintable(True)
         self._window_state = {}
         self._resize_counter = 0
         self._can_set_workspace = HAS_X11_BINDINGS and CAN_SET_WORKSPACE
         self._current_frame_extents = None
         self._screen = -1
         self._frozen = False
+        self._focus_latest = None
         self.window_state_timer = None
         self.send_iconify_timer = None
         self.remove_pointer_overlay_timer = None
@@ -219,11 +255,20 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
 
     def init_drawing_area(self):
         widget = Gtk.DrawingArea()
+        widget.set_app_paintable(True)
+        widget.set_size_request(*self._size)
         widget.show()
         self.drawing_area = widget
         self.init_widget_events(widget)
         self.add(widget)
 
+    def repaint(self, x, y, w, h):
+        #self.queue_draw_area(0, 0, *self._size)
+        widget = self.drawing_area
+        if widget:
+            widget.queue_draw_area(x, y, w, h)
+
+
     def init_widget_events(self, widget):
         widget.add_events(WINDOW_EVENT_MASK)
         def motion(_w, event):
@@ -242,6 +287,13 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
             self._do_scroll_event(event)
             return True
         widget.connect("scroll-event", scroll)
+        widget.connect("draw", self.draw_widget)
+
+    def draw_widget(self, widget, context):
+        raise NotImplementedError()
+
+    def get_drawing_area_geometry(self):
+        raise NotImplementedError()
 
 
     ######################################################################
@@ -381,7 +433,6 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
                 log.warn("Warning: missing gdk bindings:")
                 log.warn(" %s", e)
             else:
-                self._focus_latest = None
                 grablog("adding event receiver so we can get FocusIn and FocusOut events whilst grabbing the keyboard")
                 add_event_receiver(self.get_window(), self)
         #other platforms should bet getting regular focus events instead:
@@ -395,29 +446,37 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
         self.connect("focus-out-event", focus_out)
         if not self._override_redirect:
             self.connect("notify::has-toplevel-focus", self._focus_change)
+        def grab_broken(win, event):
+            grablog("grab_broken%s", (win, event))
+            self._client._window_with_grab = None
+        self.connect("grab-broken-event", grab_broken)
 
     def _focus_change(self, *args):
         assert not self._override_redirect
         htf = self.has_toplevel_focus()
         focuslog("%s focus_change%s has-toplevel-focus=%s, _been_mapped=%s", self, args, htf, self._been_mapped)
         if self._been_mapped:
-            self._client.update_focus(self._id, htf)
+            self._focus_latest = htf
+            self.send_latest_focus()
 
     def recheck_focus(self):
         self.recheck_focus_timer = 0
-        #we receive pairs of FocusOut + FocusIn following a keyboard grab,
-        #so we recheck the focus status via this timer to skip unnecessary churn
+        self.send_latest_focus()
+
+    def send_latest_focus(self):
         focused = self._client._focused
         focuslog("recheck_focus() wid=%i, focused=%s, latest=%s", self._id, focused, self._focus_latest)
-        hasfocus = focused==self._id
-        if hasfocus==self._focus_latest:
-            #we're already up to date
-            return
-        if not self._focus_latest:
-            self._client.window_ungrab()
-            self._client.update_focus(self._id, False)
+        if self._focus_latest:
+            self._focus()
         else:
-            self._client.update_focus(self._id, True)
+            self._unfocus()
+
+    def _focus(self):
+        super()._focus()
+
+    def _unfocus(self):
+        self._client.window_ungrab()
+        super()._unfocus()
 
     def cancel_focus_timer(self):
         rft = self.recheck_focus_timer
@@ -426,12 +485,12 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
             self.source_remove(rft)
 
     def schedule_recheck_focus(self):
+        #we receive pairs of FocusOut + FocusIn following a keyboard grab,
+        #so we recheck the focus status via this timer to skip unnecessary churn
         if FOCUS_RECHECK_DELAY<0:
             self.recheck_focus()
-            return
-        if self.recheck_focus_timer==0:
+        elif self.recheck_focus_timer==0:
             self.recheck_focus_timer = self.timeout_add(FOCUS_RECHECK_DELAY, self.recheck_focus)
-        return True
 
     def do_xpra_focus_out_event(self, event):
         focuslog("do_xpra_focus_out_event(%s)", event)
@@ -441,13 +500,15 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
                 focuslog("dropped NotifyInferior focus event")
                 return True
         self._focus_latest = False
-        return self.schedule_recheck_focus()
+        self.schedule_recheck_focus()
+        return True
 
     def do_xpra_focus_in_event(self, event):
         focuslog("do_xpra_focus_in_event(%s) been_mapped=%s", event, self._been_mapped)
         if self._been_mapped:
             self._focus_latest = True
-            return self.schedule_recheck_focus()
+            self.schedule_recheck_focus()
+        return True
 
 
     def init_max_window_size(self):
@@ -563,7 +624,7 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
                     if self.window_offset:
                         x += self.window_offset[0]
                         y += self.window_offset[1]
-            if not self.is_OR() and self.get_decorated():
+            elif self.get_decorated():
                 #try to adjust for window frame size if we can figure it out:
                 #Note: we cannot just call self.get_window_frame_size() here because
                 #the window is not realized yet, and it may take a while for the window manager
@@ -727,17 +788,19 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
         #by default, only RGB (no transparency):
         #rgb_formats = tuple(BACKING_CLASS.RGB_MODES)
         self._client_properties["encodings.rgb_formats"] = ["RGB", "RGBX"]
-        if not self._has_alpha or not bc.HAS_ALPHA:
-            self._client_properties["encoding.transparency"] = False
-            return
-        if self._has_alpha and not self.get_realized():
-            if enable_alpha(self):
-                self._client_properties["encodings.rgb_formats"] = ["RGBA", "RGB", "RGBX"]
-                self._window_alpha = True
+        #only set the visual if we need to enable alpha:
+        #(breaks the headerbar otherwise!)
+        if not self.get_realized() and self._has_alpha:
+            if set_visual(self, True):
+                if self._has_alpha:
+                    self._client_properties["encodings.rgb_formats"] = ["RGBA", "RGB", "RGBX"]
+                self._window_alpha = self._has_alpha
             else:
-                alphalog("enable_alpha()=False")
+                alphalog("failed to set RGBA visual")
                 self._has_alpha = False
                 self._client_properties["encoding.transparency"] = False
+        if not self._has_alpha or not bc.HAS_ALPHA:
+            self._client_properties["encoding.transparency"] = False
 
 
     def freeze(self):
@@ -756,31 +819,14 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
         self.deiconify()
 
 
-    def show(self):
-        Gtk.Window.show(self)
-
-
     def window_state_updated(self, widget, event):
         statelog("%s.window_state_updated(%s, %s) changed_mask=%s, new_window_state=%s",
                  self, widget, repr(event), event.changed_mask, event.new_window_state)
         state_updates = {}
-        if event.changed_mask & Gdk.WindowState.FULLSCREEN:
-            state_updates["fullscreen"] = bool(event.new_window_state & Gdk.WindowState.FULLSCREEN)
-        if event.changed_mask & Gdk.WindowState.ABOVE:
-            state_updates["above"] = bool(event.new_window_state & Gdk.WindowState.ABOVE)
-        if event.changed_mask & Gdk.WindowState.BELOW:
-            state_updates["below"] = bool(event.new_window_state & Gdk.WindowState.BELOW)
-        if event.changed_mask & Gdk.WindowState.STICKY:
-            state_updates["sticky"] = bool(event.new_window_state & Gdk.WindowState.STICKY)
-        if event.changed_mask & Gdk.WindowState.ICONIFIED:
-            state_updates["iconified"] = bool(event.new_window_state & Gdk.WindowState.ICONIFIED)
-        if event.changed_mask & Gdk.WindowState.MAXIMIZED:
-            #this may get sent now as part of map_event code below (and it is irrelevant for the unmap case),
-            #or when we get the configure event - which should come straight after
-            #if we're changing the maximized state
-            state_updates["maximized"] = bool(event.new_window_state & Gdk.WindowState.MAXIMIZED)
-        if event.changed_mask & Gdk.WindowState.FOCUSED:
-            state_updates["focused"] = bool(event.new_window_state & Gdk.WindowState.FOCUSED)
+        for flag in ("fullscreen", "above", "below", "sticky", "iconified", "maximized", "focused"):
+            wstate = getattr(Gdk.WindowState, flag.upper()) #ie: Gdk.WindowState.FULLSCREEN
+            if event.changed_mask & wstate:
+                state_updates[flag] = bool(event.new_window_state & wstate)
         self.update_window_state(state_updates)
 
     def update_window_state(self, state_updates):
@@ -796,7 +842,7 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
                 self.window_offset, (w, h), (ww, wh))
             if self._backing.offsets!=(0, 0, 0, 0):
                 self.center_backing(w, h)
-                self.queue_draw_area(0, 0, ww, wh)
+                self.repaint(0, 0, ww, wh)
         #decide if this is really an update by comparing with our local state vars:
         #(could just be a notification of a state change we already know about)
         actual_updates = {}
@@ -809,11 +855,8 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
                 statelog("%s=%s (was %s)", var, value, cur)
         server_updates = dict((k,v) for k,v in actual_updates.items() if k in self._client.server_window_states)
         #iconification is handled a bit differently...
-        try:
-            iconified = server_updates.pop("iconified")
-        except KeyError:
-            iconified = None
-        else:
+        iconified = server_updates.pop("iconified", None)
+        if iconified is not None:
             statelog("iconified=%s", iconified)
             #handle iconification as map events:
             if iconified:
@@ -831,14 +874,14 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
         if "maximized" in state_updates:
             if REPAINT_MAXIMIZED>0:
                 def repaint_maximized():
-                    widget = self.drawing_area
-                    if not self._backing or not widget:
+                    if not self._backing:
                         return
                     ww, wh = self.get_size()
-                    widget.queue_draw_area(0, 0, ww, wh)
+                    self.repaint(0, 0, ww, wh)
                 self.timeout_add(REPAINT_MAXIMIZED, repaint_maximized)
             if REFRESH_MAXIMIZED:
                 self._client.send_refresh(self._id)
+
         self._window_state.update(server_updates)
         self.emit("state-updated")
         #if we have state updates, send them back to the server using a configure window packet:
@@ -895,7 +938,7 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
             except UnicodeDecodeError:
                 v = bytestostr(command)
         def do_set_command():
-            metalog("do_set_command() str(%s)='%s' (type=%s)", command, nonl(v), type(command))
+            metalog("do_set_command() str(%s)='%r' (type=%s)", command, v, type(command))
             prop_set(self.get_window(), "WM_COMMAND", "latin1", v)
         self.when_realized("command", do_set_command)
 
@@ -1030,6 +1073,27 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
         self.when_realized("strut", do_set_strut)
 
 
+    def set_window_type(self, window_types):
+        pass
+        hints = 0
+        for window_type in window_types:
+            #win32 workaround:
+            if AWT_DIALOG_WORKAROUND and window_type=="DIALOG" and self._metadata.boolget("skip-taskbar"):
+                wm_class = self._metadata.strtupleget("class-instance", (None, None), 2, 2)
+                if wm_class and len(wm_class)==2 and wm_class[0] and wm_class[0].startswith("sun-awt-X11"):
+                    #replace "DIALOG" with "NORMAL":
+                    if "NORMAL" in window_types:
+                        continue
+                    window_type = "NORMAL"
+            hint = WINDOW_NAME_TO_HINT.get(window_type, None)
+            if hint is not None:
+                hints |= hint
+            else:
+                log("ignoring unknown window type hint: %s", window_type)
+        log("set_window_type(%s) hints=%s", window_types, hints)
+        if hints:
+            self.set_type_hint(hints)
+
     def set_modal(self, modal):
         #with gtk2 setting the window as modal would prevent
         #all other windows we manage from receiving input
@@ -1213,7 +1277,7 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
             client_properties["workspace"] = window_workspace
         self.send_control_refresh(suspend_resume, client_properties)
 
-    def send_control_refresh(self, suspend_resume, client_properties={}, refresh=False):
+    def send_control_refresh(self, suspend_resume, client_properties=None, refresh=False):
         statelog("send_control_refresh%s", (suspend_resume, client_properties, refresh))
         #we can tell the server using a "buffer-refresh" packet instead
         #and also take care of tweaking the batch config
@@ -1266,7 +1330,7 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
             return
         gdkwin = self.get_window()
         workspacelog("do_set_workspace: gdkwindow: %#x, mapped=%s, visible=%s",
-                     gdkwin.get_xid(), self.is_mapped(), gdkwin.is_visible())
+                     gdkwin.get_xid(), self.get_mapped(), gdkwin.is_visible())
         root = get_default_root_window()
         with xlog:
             send_wm_workspace(root, gdkwin, workspace)
@@ -1307,12 +1371,12 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
         if gdkwin:
             d = gdkwin.get_display()
             if d:
-                d.keyboard_ungrab(0)
+                d.keyboard_ungrab(Gdk.CURRENT_TIME)
         return True
 
     def keyboard_grab(self, *args):
         grablog("keyboard_grab%s", args)
-        r = Gdk.keyboard_grab(self.get_window(), True, 0)
+        r = Gdk.keyboard_grab(self.get_window(), True, Gdk.CURRENT_TIME)
         self._client.keyboard_grabbed = r==Gdk.GrabStatus.SUCCESS
         grablog("keyboard_grab%s Gdk.keyboard_grab(%s, True)=%s, keyboard_grabbed=%s",
                 args, self.get_window(), GRAB_STATUS_STRING.get(r), self._client.keyboard_grabbed)
@@ -1334,7 +1398,7 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
                       em.POINTER_MOTION_HINT_MASK |
                       em.ENTER_NOTIFY_MASK |
                       em.LEAVE_NOTIFY_MASK)
-        r = Gdk.pointer_grab(gdkwin, True, event_mask, gdkwin, None, 0)
+        r = Gdk.pointer_grab(gdkwin, True, event_mask, gdkwin, None, Gdk.CURRENT_TIME)
         self._client.pointer_grabbed = r==Gdk.GrabStatus.SUCCESS
         grablog("pointer_grab%s Gdk.pointer_grab(%s, True)=%s, pointer_grabbed=%s",
                 args, self.get_window(), GRAB_STATUS_STRING.get(r), self._client.pointer_grabbed)
@@ -1347,7 +1411,7 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
         if gdkwin:
             d = gdkwin.get_display()
             if d:
-                d.pointer_ungrab(0)
+                d.pointer_ungrab(Gdk.CURRENT_TIME)
         return True
 
     def toggle_pointer_grab(self):
@@ -1427,7 +1491,7 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
         if value:
             #repaint the scale value (in window coordinates):
             x, y, w, h = abs_coords(*value[2:5])
-            self.queue_draw_area(x, y, w, h)
+            self.repaint(x, y, w, h)
             #clear it shortly after:
             self.cancel_remove_pointer_overlay_timer()
             def remove_pointer_overlay():
@@ -1436,7 +1500,7 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
             self.remove_pointer_overlay_timer = self.timeout_add(CURSOR_IDLE_TIMEOUT*1000, remove_pointer_overlay)
         if prev:
             x, y, w, h = abs_coords(*prev[2:5])
-            self.queue_draw_area(x, y, w, h)
+            self.repaint(x, y, w, h)
 
 
     def _do_button_press_event(self, event):
@@ -1460,6 +1524,7 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
         x_root, y_root, direction, button, start_buttons, wx, wy, ww, wh = self.moveresize_event
         dirstr = MOVERESIZE_DIRECTION_STRING.get(direction, direction)
         buttons = self._event_buttons(event)
+        geomlog("motion_moveresize(%s) direction=%s, buttons=%s", event, dirstr, buttons)
         if start_buttons is None:
             #first time around, store the buttons
             start_buttons = buttons
@@ -1468,6 +1533,7 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
             geomlog("%s for window button %i is no longer pressed (buttons=%s) cancelling moveresize",
                     dirstr, button, buttons)
             self.moveresize_event = None
+            self.cancel_moveresize_timer()
         else:
             x = event.x_root
             y = event.y_root
@@ -1528,6 +1594,12 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
                 if self.moveresize_timer is None:
                     self.moveresize_timer = self.timeout_add(20, self.do_moveresize)
 
+    def cancel_moveresize_timer(self):
+        mrt = self.moveresize_timer
+        if mrt:
+            self.moveresize_timer = None
+            self.source_remove(mrt)
+
     def do_moveresize(self):
         self.moveresize_timer = None
         mrd = self.moveresize_data
@@ -1554,16 +1626,42 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
 
 
     def initiate_moveresize(self, x_root, y_root, direction, button, source_indication):
-        statelog("initiate_moveresize%s",
+        geomlog("initiate_moveresize%s",
                  (x_root, y_root, MOVERESIZE_DIRECTION_STRING.get(direction, direction),
                   button, SOURCE_INDICATION_STRING.get(source_indication, source_indication)))
+        #the values we get are bogus!
+        #x, y = x_root, y_root
+        #use the current position instead:
+        p = self.get_root_window().get_pointer()[-3:-1]
+        x, y = p[0], p[1]
         if MOVERESIZE_X11 and HAS_X11_BINDINGS:
-            self.initiate_moveresize_X11(x_root, y_root, direction, button, source_indication)
+            self.initiate_moveresize_X11(x, y, direction, button, source_indication)
             return
         if direction==MOVERESIZE_CANCEL:
             self.moveresize_event = None
             self.moveresize_data = None
+            self.cancel_moveresize_timer()
+        elif MOVERESIZE_GDK:
+            if direction in (MOVERESIZE_MOVE, MOVERESIZE_MOVE_KEYBOARD):
+                self.begin_move_drag(button, x, y, 0)
+            else:
+                edge = {
+                    MOVERESIZE_SIZE_TOPLEFT     : Gdk.WindowEdge.NORTH_WEST,
+                    MOVERESIZE_SIZE_TOP         : Gdk.WindowEdge.NORTH,
+                    MOVERESIZE_SIZE_TOPRIGHT    : Gdk.WindowEdge.NORTH_EAST,
+                    MOVERESIZE_SIZE_RIGHT       : Gdk.WindowEdge.EAST,
+                    MOVERESIZE_SIZE_BOTTOMRIGHT : Gdk.WindowEdge.SOUTH_EAST,
+                    MOVERESIZE_SIZE_BOTTOM      : Gdk.WindowEdge.SOUTH,
+                    MOVERESIZE_SIZE_BOTTOMLEFT  : Gdk.WindowEdge.SOUTH_WEST,
+                    MOVERESIZE_SIZE_LEFT        : Gdk.WindowEdge.WEST,
+                    #MOVERESIZE_SIZE_KEYBOARD,
+                    }.get(direction)
+                geomlog("edge(%s)=%s", MOVERESIZE_DIRECTION_STRING.get(direction), edge)
+                if direction is not None:
+                    etime = Gtk.get_current_event_time()
+                    self.begin_resize_drag(edge, button, x, y, etime)
         else:
+            #handle it ourselves:
             #use window coordinates (which include decorations)
             wx, wy = self.get_window().get_root_origin()
             ww, wh = self.get_size()
@@ -1589,7 +1687,7 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
                 #root is a gdk window, so we need to ensure we have one
                 #backing our gtk window to be able to call set_transient_for on it
                 log("%s.apply_transient_for(%s) gdkwindow=%s, mapped=%s",
-                    self, wid, self.get_window(), self.is_mapped())
+                    self, wid, self.get_window(), self.get_mapped())
                 self.get_window().set_transient_for(get_default_root_window())
             self.when_realized("transient-for-root", set_root_transient)
         else:
@@ -1644,7 +1742,7 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
         w = c.cx(ww)
         h = c.cy(wh)
         #add grey semi-opaque layer on top:
-        context.set_operator(cairo.OPERATOR_OVER)
+        context.set_operator(OPERATOR_OVER)
         context.set_source_rgba(0.2, 0.2, 0.2, 0.4)
         #we can't use the area as rectangle with:
         #context.rectangle(area)
@@ -1655,7 +1753,7 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
         #add spinner:
         dim = min(w/3.0, h/3.0, 100.0)
         context.set_line_width(dim/10.0)
-        context.set_line_cap(cairo.LINE_CAP_ROUND)
+        context.set_line_cap(LINE_CAP_ROUND)
         context.translate(w/2, h/2)
         from xpra.client.spinner import cv
         count = int(monotonic_time()*4.0)
@@ -1673,7 +1771,7 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
         #with normal windows, we just queue a draw request
         #and let the expose event paint the spinner
         w, h = self.get_size()
-        self.queue_draw_area(0, 0, w, h)
+        self.repaint(0, 0, w, h)
 
 
     def do_map_event(self, event):
@@ -1684,6 +1782,11 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
             if self._iconified:
                 self.deiconify()
             self.process_map_event()
+        #use the drawing area to enforce the minimum size:
+        #(as this also honoured correctly with CSD,
+        # whereas set_geometry_hints is not..)
+        minw, minh = self.size_constraints.intpair("minimum-size", (0, 0))
+        self.drawing_area.set_size_request(minw, minh)
 
     def process_map_event(self):
         x, y, w, h = self.get_drawing_area_geometry()
@@ -1718,13 +1821,13 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
         cx = self._client.cx
         cy = self._client.cy
         sx, sy, sw, sh = cx(x), cy(y), cx(w), cy(h)
+        if self._backing is None:
+            #we may have cleared the backing, so we must re-create one:
+            self._set_backing_size(w, h)
         packet = ["map-window", self._id, sx, sy, sw, sh, props, state]
         self.send(*packet)
         self._pos = (x, y)
         self._size = (w, h)
-        if self._backing is None:
-            #we may have cleared the backing, so we must re-create one:
-            self._set_backing_size(w, h)
         if not self._override_redirect:
             htf = self.has_toplevel_focus()
             focuslog("mapped: has-toplevel-focus=%s", htf)
@@ -1741,6 +1844,12 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
         return frame
 
 
+    def may_send_client_properties(self):
+        #if there are client properties the server should know about,
+        #we currently have no other way to send them to the server:
+        if self._client_properties:
+            self.send_configure_event(True)
+
     def send_configure(self):
         self.send_configure_event()
 
@@ -1759,7 +1868,6 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
         ox, oy = self._pos
         dx, dy = x-ox, y-oy
         self._pos = (x, y)
-        self.send_configure_event(skip_geometry)
         if dx!=0 or dy!=0:
             #window has moved, also move any child OR window:
             for window in self._override_redirect_windows:
@@ -1767,12 +1875,12 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
                 window.move(x+dx, y+dy)
         geomlog("configure event: current size=%s, new size=%s, backing=%s, iconified=%s",
                 self._size, (w, h), self._backing, self._iconified)
-        if (w, h) != self._size or (self._backing is None and not self._iconified):
-            self._size = (w, h)
-            self._set_backing_size(w, h)
-        elif self._backing and not self._iconified:
+        self._size = (w, h)
+        self._set_backing_size(w, h)
+        self.send_configure_event(skip_geometry)
+        if self._backing and not self._iconified:
             geomlog("configure event: size unchanged, queueing redraw")
-            self.queue_draw_area(0, 0, w, h)
+            self.repaint(0, 0, w, h)
 
     def send_configure_event(self, skip_geometry=False):
         assert skip_geometry or not self.is_OR()
@@ -1807,10 +1915,19 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
 
     def _set_backing_size(self, ww, wh):
         b = self._backing
+        bw = self._client.cx(ww)
+        bh = self._client.cy(wh)
+        if max(ww, wh)>=32000 or min(ww, wh)<0:
+            raise Exception("invalid window size %ix%i" % (ww, wh))
+        if max(bw, bh)>=32000:
+            raise Exception("invalid window backing size %ix%i" % (bw, bh))
         if b:
-            b.init(ww, wh, self._client.cx(ww), self._client.cy(wh))
+            prev_render_size = b.render_size
+            b.init(ww, wh, bw, bh)
+            if prev_render_size!=b.render_size:
+                self._client_properties["encoding.render-size"] = b.render_size
         else:
-            self.new_backing(self._client.cx(ww), self._client.cy(wh))
+            self.new_backing(bw, bh)
 
     def resize(self, w, h, resize_counter=0):
         ww, wh = self.get_size()
@@ -1819,7 +1936,7 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
         self._resize_counter = resize_counter
         if (w, h)==(ww, wh):
             self._backing.offsets = 0, 0, 0, 0
-            self.queue_draw_area(0, 0, w, h)
+            self.repaint(0, 0, w, h)
             return
         if not self._fullscreen and not self._maximized:
             Gtk.Window.resize(self, w, h)
@@ -1829,7 +1946,8 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
             self.center_backing(w, h)
         geomlog("backing offsets=%s, window offset=%s", self._backing.offsets, self.window_offset)
         self._set_backing_size(w, h)
-        self.queue_draw_area(0, 0, ww, wh)
+        self.repaint(0, 0, ww, wh)
+        self.may_send_client_properties()
 
     def center_backing(self, w, h):
         ww, wh = self.get_size()
@@ -1929,6 +2047,8 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
         window.move_resize(ax, ay, w, h)
         #re-init the backing with the new size
         self._set_backing_size(w, h)
+        self.repaint(0, 0, w, h)
+        self.may_send_client_properties()
 
 
     def noop_destroy(self):
@@ -1940,10 +2060,7 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
         self.cancel_show_pointer_overlay_timer()
         self.cancel_remove_pointer_overlay_timer()
         self.cancel_focus_timer()
-        mrt = self.moveresize_timer
-        if mrt:
-            self.moveresize_timer = None
-            self.source_remove(mrt)
+        self.cancel_moveresize_timer()
         self.on_realize_cb = {}
         ClientWindowBase.destroy(self)
         Gtk.Window.destroy(self)
@@ -2014,7 +2131,7 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
                 keylog.warn(" %s", e)
             key_event.string = ""
         key_event.pressed = pressed
-        keylog("parse_key_event(%s, %s)=%s", event, pressed, key_event)
+        keyeventlog("parse_key_event(%s, %s)=%s", event, pressed, key_event)
         return key_event
 
     def do_key_press_event(self, event):
@@ -2022,16 +2139,21 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
         if self.moveresize_event and key_event.keyname in BREAK_MOVERESIZE:
             #cancel move resize if there is one:
             self.moveresize_event = None
-        self._client.handle_key_action(self, key_event)
+            self.cancel_moveresize_timer()
+        return self._client.handle_key_action(self, key_event)
 
     def do_key_release_event(self, event):
         key_event = self.parse_key_event(event, False)
-        self._client.handle_key_action(self, key_event)
+        return self._client.handle_key_action(self, key_event)
 
 
     def _do_scroll_event(self, event):
         if self._client.readonly:
             return
+        if event.direction==Gdk.ScrollDirection.SMOOTH:
+            mouselog("smooth scroll event: %s", event)
+            self._client.wheel_event(self._id, event.delta_x, -event.delta_y)
+            return
         button_mapping = GDK_SCROLL_MAP.get(event.direction, -1)
         mouselog("do_scroll_event device=%s, direction=%s, button_mapping=%s",
                  self._device_info(event), event.direction, button_mapping)
