@@ -3,6 +3,8 @@
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
+from gi.repository import GLib
+
 from xpra.platform.paths import get_icon_filename
 from xpra.scripts.parsing import sound_option
 from xpra.net.compression import Compressed
@@ -19,10 +21,10 @@ DELTA_THRESHOLD = envint("XPRA_AV_SYNC_DELTA_THRESHOLD", 40)
 DEFAULT_AV_SYNC_DELAY = envint("XPRA_DEFAULT_AV_SYNC_DELAY", 150)
 
 
-"""
-Utility superclass for clients that handle audio
-"""
 class AudioClient(StubClientMixin):
+    """
+    Utility mixin for clients that handle audio
+    """
     __signals__ = ["speaker-changed", "microphone-changed"]
 
     def __init__(self):
@@ -140,6 +142,21 @@ class AudioClient(StubClientMixin):
             self.stop_receiving_sound()
 
 
+    def get_info(self) -> dict:
+        info = {
+            "speaker" : self.speaker_enabled,
+            "microphone" : self.microphone_enabled,
+            "properties" : dict(self.sound_properties),
+            }
+        ss = self.sound_source
+        if ss:
+            info["src"] = ss.get_info()
+        ss = self.sound_sink
+        if ss:
+            info["sink"] = ss.get_info()
+        return {"audio" : info}
+
+
     def get_caps(self) -> dict:
         d = {}
         updict(d, "av-sync", self.get_avsync_capabilities())
@@ -186,9 +203,12 @@ class AudioClient(StubClientMixin):
                  csv(self.server_sound_decoders), csv(self.server_sound_encoders),
                  self.server_sound_receive, self.server_sound_send)
         if self.server_sound_send and self.speaker_enabled:
+            self.show_progress(90, "starting speaker forwarding")
             self.start_receiving_sound()
         if self.server_sound_receive and self.microphone_enabled:
-            self.start_sending_sound()
+            #call via idle_add because we may query X11 properties
+            #to find the pulseaudio server:
+            GLib.idle_add(self.start_sending_sound)
         return True
 
 
@@ -338,7 +358,7 @@ class AudioClient(StubClientMixin):
             self.emit("microphone-changed")
         self.sound_source = None
         if ss is None:
-            log.warn("Warning: cannot stop sound source which has not been started")
+            log.warn("Warning: cannot stop audio capture which has not been started")
             return
         #tell the server to stop:
         self.send("sound-data", ss.codec or "", "", {
