@@ -1,12 +1,12 @@
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2021 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import os
 
-from xpra.codecs.codec_constants import PREFERED_ENCODING_ORDER, PROBLEMATIC_ENCODINGS
+from xpra.codecs.codec_constants import PREFERRED_ENCODING_ORDER
 from xpra.codecs.loader import load_codec, codec_versions, has_codec, get_codec
 from xpra.codecs.video_helper import getVideoHelper, NO_GFX_CSC_OPTIONS
 from xpra.scripts.config import parse_bool_or_int
@@ -22,20 +22,51 @@ PAINT_FLUSH = envbool("XPRA_PAINT_FLUSH", True)
 MAX_SOFT_EXPIRED = envint("XPRA_MAX_SOFT_EXPIRED", 5)
 SEND_TIMESTAMPS = envbool("XPRA_SEND_TIMESTAMPS", False)
 VIDEO_MAX_SIZE = tuple(int(x) for x in os.environ.get("XPRA_VIDEO_MAX_SIZE", "4096,4096").replace("x", ",").split(","))
+SCROLL_ENCODING = envbool("XPRA_SCROLL_ENCODING", True)
 
 #we assume that any server will support at least those:
 DEFAULT_ENCODINGS = os.environ.get("XPRA_DEFAULT_ENCODINGS", "rgb32,rgb24,jpeg,png").split(",")
 
 
-"""
-Mixin for adding encodings to a client
-"""
+def get_core_encodings():
+    """
+        This method returns the actual encodings supported.
+        ie: ["rgb24", "vp8", "webp", "png", "png/L", "png/P", "jpeg", "h264", "vpx"]
+        It is often overriden in the actual client class implementations,
+        where extra encodings can be added (generally just 'rgb32' for transparency),
+        or removed if the toolkit implementation class is more limited.
+    """
+    #we always support rgb:
+    core_encodings = ["rgb24", "rgb32"]
+    for codec in ("dec_pillow", "dec_webp", "dec_jpeg"):
+        if has_codec(codec):
+            c = get_codec(codec)
+            encs = c.get_encodings()
+            log("%s.get_encodings()=%s", codec, encs)
+            for e in encs:
+                if e not in core_encodings:
+                    core_encodings.append(e)
+    if SCROLL_ENCODING:
+        core_encodings.append("scroll")
+    #we enable all the video decoders we know about,
+    #what will actually get used by the server will still depend on the csc modes supported
+    video_decodings = getVideoHelper().get_decodings()
+    log("video_decodings=%s", video_decodings)
+    for encoding in video_decodings:
+        if encoding not in core_encodings:
+            core_encodings.append(encoding)
+    #remove duplicates and use prefered encoding order:
+    return [x for x in PREFERRED_ENCODING_ORDER if x in set(core_encodings)]
+
+
 class Encodings(StubClientMixin):
+    """
+    Mixin for adding encodings to a client
+    """
 
     def __init__(self):
         StubClientMixin.__init__(self)
         self.allowed_encodings = []
-        self.core_encodings = None
         self.encoding = None
         self.quality = -1
         self.min_quality = 0
@@ -46,7 +77,6 @@ class Encodings(StubClientMixin):
 
         self.server_encodings = []
         self.server_core_encodings = []
-        self.server_encodings_problematic = PROBLEMATIC_ENCODINGS
         self.server_encodings_with_speed = ()
         self.server_encodings_with_quality = ()
         self.server_encodings_with_lossless_mode = ()
@@ -82,7 +112,7 @@ class Encodings(StubClientMixin):
     def cleanup(self):
         try:
             getVideoHelper().cleanup()
-        except Exception:
+        except Exception:   # pragma: no cover
             log.error("error on video cleanup", exc_info=True)
 
 
@@ -124,15 +154,13 @@ class Encodings(StubClientMixin):
         updict(caps, "encoding",        self.get_encodings_caps())
         return caps
 
-    def parse_server_capabilities(self, caps : typedict) -> bool:
-        self._parse_server_capabilities(caps)
+    def parse_server_capabilities(self, c : typedict) -> bool:
+        self._parse_server_capabilities(c)
         return True
 
     def _parse_server_capabilities(self, c):
         self.server_encodings = c.strtupleget("encodings", DEFAULT_ENCODINGS)
         self.server_core_encodings = c.strtupleget("encodings.core", self.server_encodings)
-        #server is telling us to try to avoid those:
-        self.server_encodings_problematic = c.strtupleget("encodings.problematic", PROBLEMATIC_ENCODINGS)
         #old servers only supported x264:
         self.server_encodings_with_speed = c.strtupleget("encodings.with_speed", ("h264",))
         self.server_encodings_with_quality = c.strtupleget("encodings.with_quality", ("jpeg", "webp", "h264"))
@@ -172,7 +200,6 @@ class Encodings(StubClientMixin):
             "video_max_size"            : self.video_max_size,
             "max-soft-expired"          : MAX_SOFT_EXPIRED,
             "send-timestamps"           : SEND_TIMESTAMPS,
-            "supports_delta"            : tuple(x for x in ("png", "rgb24", "rgb32") if x in self.get_core_encodings()),
             }
         if self.video_scaling is not None:
             caps["scaling.control"] = self.video_scaling
@@ -236,15 +263,12 @@ class Encodings(StubClientMixin):
             and the actual encoding used ("rgb24" or "rgb32") depends on the window's bit depth.
             ("rgb32" if there is an alpha channel, and if the client supports it)
         """
-        cenc = list(self.get_core_encodings())
+        cenc = self.get_core_encodings()
         if ("rgb24" in cenc or "rgb32" in cenc) and "rgb" not in cenc:
             cenc.append("rgb")
-        return [x for x in PREFERED_ENCODING_ORDER if x in cenc and x not in ("rgb32", "rgb24")]
-
-    def get_core_encodings(self):
-        if self.core_encodings is None:
-            self.core_encodings = self.do_get_core_encodings()
-        return self.core_encodings
+        if "grayscale" not in cenc:
+            cenc.append("grayscale")
+        return [x for x in PREFERRED_ENCODING_ORDER if x in cenc and x not in ("rgb32", "rgb24")]
 
     def get_cursor_encodings(self):
         e = ["raw"]
@@ -258,33 +282,8 @@ class Encodings(StubClientMixin):
             e.append("png")
         return e
 
-    def do_get_core_encodings(self):
-        """
-            This method returns the actual encodings supported.
-            ie: ["rgb24", "vp8", "webp", "png", "png/L", "png/P", "jpeg", "h264", "vpx"]
-            It is often overriden in the actual client class implementations,
-            where extra encodings can be added (generally just 'rgb32' for transparency),
-            or removed if the toolkit implementation class is more limited.
-        """
-        #we always support rgb:
-        core_encodings = ["rgb24", "rgb32"]
-        for codec in ("dec_pillow", "dec_webp", "dec_jpeg"):
-            if has_codec(codec):
-                c = get_codec(codec)
-                for e in c.get_encodings():
-                    if e not in core_encodings:
-                        core_encodings.append(e)
-        #we enable all the video decoders we know about,
-        #what will actually get used by the server will still depend on the csc modes supported
-        video_decodings = getVideoHelper().get_decodings()
-        log("video_decodings=%s", video_decodings)
-        for encoding in video_decodings:
-            if encoding not in core_encodings:
-                core_encodings.append(encoding)
-        #remove duplicates and use prefered encoding order:
-        core_encodings = [x for x in PREFERED_ENCODING_ORDER
-                          if x in set(core_encodings) and x in self.allowed_encodings]
-        return core_encodings
+    def get_core_encodings(self):
+        return [x for x in get_core_encodings() if x in self.allowed_encodings]
 
     def set_encoding(self, encoding):
         log("set_encoding(%s)", encoding)
@@ -297,7 +296,7 @@ class Encodings(StubClientMixin):
                 log.error(" the only encodings allowed are:")
                 log.error(" %s", csv(self.server_encodings))
                 return
-        self.encoding = encoding
+            self.encoding = encoding
         self.send("encoding", self.encoding)
 
     def send_quality(self):
