@@ -1,16 +1,16 @@
 # This file is part of Xpra.
-# Copyright (C) 2010-2018 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-from xpra.log import Logger
-printlog = Logger("printing")
-filelog = Logger("file")
-
-from xpra.util import envbool, csv
+from xpra.util import envbool, csv, typedict
 from xpra.net.file_transfer import FileTransferHandler
 from xpra.client.mixins.stub_client_mixin import StubClientMixin
+from xpra.make_thread import start_thread
+from xpra.log import Logger
 
+printlog = Logger("printing")
+filelog = Logger("file")
 
 DELETE_PRINTER_FILE = envbool("XPRA_DELETE_PRINTER_FILE", True)
 SKIP_STOPPED_PRINTERS = envbool("XPRA_SKIP_STOPPED_PRINTERS", True)
@@ -24,22 +24,24 @@ class FilePrintMixin(StubClientMixin, FileTransferHandler):
         self.printer_attributes = []
         self.send_printers_timer = 0
         self.exported_printers = None
+        self.remote_request_file = False
 
     def init(self, opts):
         #printing and file transfer:
         FileTransferHandler.init_opts(self, opts)
 
     def init_authenticated_packet_handlers(self):
-        self.set_packet_handlers(self._packet_handlers, {
+        for packet_type, handler in {
             "open-url"          : self._process_open_url,
             "send-file"         : self._process_send_file,
             "send-data-request" : self._process_send_data_request,
             "send-data-response": self._process_send_data_response,
             "ack-file-chunk"    : self._process_ack_file_chunk,
             "send-file-chunk"   : self._process_send_file_chunk,
-            })
+            }.items():
+            self.add_packet_handler(packet_type, handler, False)
 
-    def get_caps(self):
+    def get_caps(self) -> dict:
         return self.get_file_transfer_features()
 
     def cleanup(self):
@@ -47,18 +49,20 @@ class FilePrintMixin(StubClientMixin, FileTransferHandler):
         self.cleanup_printing()
         FileTransferHandler.cleanup(self)
 
-    def parse_server_capabilities(self):
-        self.parse_printing_capabilities()
-        self.parse_file_transfer_caps(self.server_capabilities)
+    def parse_server_capabilities(self, c : typedict) -> bool:
+        self.parse_printing_capabilities(c)
+        self.parse_file_transfer_caps(c)
+        self.remote_request_file = c.boolget("request-file", False)
         return True
 
-    def parse_printing_capabilities(self):
+    def parse_printing_capabilities(self, caps : typedict):
         printlog("parse_printing_capabilities() client printing support=%s", self.printing)
         if self.printing:
-            server_printing = self.server_capabilities.boolget("printing")
+            server_printing = caps.boolget("printing")
             printlog("parse_printing_capabilities() server printing support=%s", server_printing)
             if server_printing:
-                self.printer_attributes = self.server_capabilities.strlistget("printer.attributes", ["printer-info", "device-uri"])
+                self.printer_attributes = caps.strtupleget("printer.attributes",
+                                                        ("printer-info", "device-uri"))
                 self.timeout_add(1000, self.init_printing)
 
 
@@ -72,11 +76,7 @@ class FilePrintMixin(StubClientMixin, FileTransferHandler):
             printlog.error(" %s", e)
             self.printing = False
         else:
-            try:
-                self.do_send_printers()
-            except Exception:
-                printlog.error("Error sending the list of printers:", exc_info=True)
-                self.printing = False
+            self.send_printers()
         printlog("init_printing() enabled=%s", self.printing)
 
     def cleanup_printing(self):
@@ -111,6 +111,10 @@ class FilePrintMixin(StubClientMixin, FileTransferHandler):
             self.source_remove(spt)
 
     def do_send_printers(self):
+        self.send_printers_timer = None
+        start_thread(self.send_printers_thread, "send-printers", True)
+
+    def send_printers_thread(self):
         try:
             self.send_printers_timer = None
             from xpra.platform.printing import get_printers, get_mimetypes
@@ -157,18 +161,19 @@ class FilePrintMixin(StubClientMixin, FileTransferHandler):
                 printlog("do_send_printers() exported printers unchanged: %s", self.exported_printers)
                 return
             #show summary of what has changed:
-            added = [k for k in exported_printers.keys() if k not in self.exported_printers]
+            added = tuple(k for k in exported_printers if k not in self.exported_printers)
             if added:
                 printlog("do_send_printers() new printers: %s", added)
-            removed = [k for k in self.exported_printers.keys() if k not in exported_printers]
+            removed = tuple(k for k in self.exported_printers if k not in exported_printers)
             if removed:
                 printlog("do_send_printers() printers removed: %s", removed)
-            modified = [k for k,v in exported_printers.items() if self.exported_printers.get(k)!=v and k not in added]
+            modified = tuple(k for k,v in exported_printers.items() if
+                        self.exported_printers.get(k)!=v and k not in added)
             if modified:
                 printlog("do_send_printers() printers modified: %s", modified)
             printlog("do_send_printers() printers=%s", exported_printers.keys())
-            printlog("do_send_printers() exported printers=%s", csv(str(x) for x in exported_printers.keys()))
+            printlog("do_send_printers() exported printers=%s", csv(str(x) for x in exported_printers))
             self.exported_printers = exported_printers
             self.send("printers", self.exported_printers)
-        except:
+        except Exception:
             printlog.error("do_send_printers()", exc_info=True)
