@@ -1,54 +1,59 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # This file is part of Xpra.
-# Copyright (C) 2013, 2014 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2013-2020 Antoine Martin <antoine@xpra.org>
 
 import re
 import sys
 from collections import deque
 
-from xpra.gtk_common.gobject_compat import import_gtk, import_gdk, is_gtk3, import_pango, import_gobject
-from xpra.gtk_common.gtk_util import TableBuilder, label, get_xwindow, GetClipboard
-from xpra.platform.paths import get_icon
+from xpra.platform import program_context
+from xpra.platform.gui import force_focus
+from xpra.util import csv
+from xpra.gtk_common.gtk_util import TableBuilder, label, get_icon_pixbuf
 from xpra.platform.features import CLIPBOARDS
 
-gtk = import_gtk()
-gdk = import_gdk()
-pango = import_pango()
-gobject = import_gobject()
+import gi
+gi.require_version("Gdk", "3.0")
+gi.require_version("Gtk", "3.0")
+gi.require_version("Pango", "1.0")
+from gi.repository import Pango, Gtk, Gdk, GLib
 
 
-class ClipboardInstance(object):
+class ClipboardInstance:
     def __init__(self, selection, _log):
-        self.clipboard = GetClipboard(selection)
+        atom = Gdk.Atom.intern(selection, False)
+        self.clipboard = Gtk.Clipboard.get(atom)
         self.selection = selection
         self._log = _log
         self.owned_label = label()
-        self.get_targets = gtk.combo_box_new_text()
+        self.get_targets = Gtk.ComboBoxText()
         self.get_targets.set_sensitive(False)
         self.get_targets.connect("changed", self.get_target_changed)
-        self.set_targets = gtk.combo_box_new_text()
+        self.set_targets = Gtk.ComboBoxText()
         self.set_targets.append_text("STRING")
         self.set_targets.append_text("UTF8_STRING")
         self.set_targets.set_active(0)
         self.set_targets.connect("changed", self.set_target_changed)
         self.value_label = label()
-        self.value_entry = gtk.Entry()
+        self.value_entry = Gtk.Entry()
         self.value_entry.set_max_length(100)
         self.value_entry.set_width_chars(32)
-        self.clear_label_btn = gtk.Button("X")
+        def b(l):
+            return Gtk.Button(label=l)
+        self.clear_label_btn = b("X")
         self.clear_label_btn.connect("clicked", self.clear_label)
-        self.clear_entry_btn = gtk.Button("X")
+        self.clear_entry_btn = b("X")
         self.clear_entry_btn.connect("clicked", self.clear_entry)
-        self.get_get_targets_btn = gtk.Button("Get Targets")
+        self.get_get_targets_btn = b("Get Targets")
         self.get_get_targets_btn.connect("clicked", self.do_get_targets)
-        self.get_target_btn = gtk.Button("Get Target")
+        self.get_target_btn = b("Get Target")
         self.get_target_btn.connect("clicked", self.do_get_target)
         self.get_target_btn.set_sensitive(False)
-        self.set_target_btn = gtk.Button("Set Target")
+        self.set_target_btn = b("Set Target")
         self.set_target_btn.connect("clicked", self.do_set_target)
-        self.get_string_btn = gtk.Button("Get String")
+        self.get_string_btn = b("Get String")
         self.get_string_btn.connect("clicked", self.do_get_string)
-        self.set_string_btn = gtk.Button("Set String")
+        self.set_string_btn = b("Set String")
         self.set_string_btn.connect("clicked", self.do_set_string)
         self.clipboard.connect("owner-change", self.owner_changed)
         self.log("ready")
@@ -66,7 +71,7 @@ class ClipboardInstance(object):
         self.value_label.set_text("")
 
     def get_targets_callback(self, _c, targets, *_args):
-        self.log("got targets: %s" % str(targets))
+        self.log("got targets: %s" % csv(str(x) for x in targets))
         if hasattr(targets, "name"):
             self.log("target is atom: %s" % targets.name())
             targets = []
@@ -116,7 +121,7 @@ class ClipboardInstance(object):
         except TypeError:
             try:
                 s = self.ellipsis("\\".join([str(x) for x in bytearray(selection_data.data)]))
-            except:
+            except Exception:
                 s = "!ERROR! binary data?"
         self.log("Got selection data: '%s'" % s)
         self.value_label.set_text(s)
@@ -158,42 +163,36 @@ class ClipboardInstance(object):
         self.clipboard.set_text(self.ellipsis(self.value_entry.get_text()))
 
     def owner_changed(self, _cb, event):
-        r = {}
-        if not is_gtk3():
-            r = {
-                gtk.gdk.OWNER_CHANGE_CLOSE      : "close",
-                gtk.gdk.OWNER_CHANGE_DESTROY    : "destroy",
-                gtk.gdk.OWNER_CHANGE_NEW_OWNER  : "new owner",
-                }
         owner = self.clipboard.get_owner()
         #print("xid=%s, owner=%s" % (self.value_entry.get_window().xid, event.owner))
         weownit = (owner is not None)
         if weownit:
             owner_info="(us)"
         else:
-            owner_info = hex(event.owner)
+            owner_info = hex(event.owner or 0)
         self.log("Owner changed, reason: %s, new owner=%s" % (
-                        r.get(event.reason, event.reason), owner_info))
+                        event.reason, owner_info))
 
 
 
-class ClipboardStateInfoWindow(object):
+class ClipboardStateInfoWindow:
 
     def    __init__(self):
-        self.window = gtk.Window()
+        self.window = Gtk.Window()
         self.window.connect("destroy", self.destroy)
         self.window.set_default_size(640, 300)
         self.window.set_border_width(20)
+        self.window.set_position(Gtk.WindowPosition.CENTER)
         self.window.set_title("Clipboard Test Tool")
 
-        vbox = gtk.VBox(False, 0)
+        vbox = Gtk.VBox(False, 0)
         vbox.set_spacing(15)
 
         self.log = deque(maxlen=25)
         for x in range(25):
             self.log.append("")
-        self.events = gtk.Label()
-        fixed = pango.FontDescription('monospace 9')
+        self.events = Gtk.Label()
+        fixed = Pango.FontDescription('monospace 9')
         self.events.modify_font(fixed)
 
         #how many clipboards to show:
@@ -206,11 +205,11 @@ class ClipboardStateInfoWindow(object):
         tb.add_row(*labels)
         for selection in self.clipboards:
             cs = ClipboardInstance(selection, self.add_event)
-            get_actions = gtk.HBox()
+            get_actions = Gtk.HBox()
             for x in (cs.get_get_targets_btn, cs.get_target_btn, cs.get_string_btn):
                 get_actions.pack_start(x)
             tb.add_row(label(selection), cs.value_label, cs.clear_label_btn, cs.get_targets, get_actions)
-            set_actions = gtk.HBox()
+            set_actions = Gtk.HBox()
             for x in (cs.set_target_btn, cs.set_string_btn):
                 set_actions.pack_start(x)
             tb.add_row(None, cs.value_entry, cs.clear_entry_btn, cs.set_targets, set_actions)
@@ -219,12 +218,12 @@ class ClipboardStateInfoWindow(object):
 
         self.window.add(vbox)
         self.window.show_all()
-        icon = get_icon("clipboard.png")
+        icon = get_icon_pixbuf("clipboard.png")
         if icon:
             self.window.set_icon(icon)
         try:
-            self.add_event("ALL", "window=%s, xid=%#x" % (self.window, get_xwindow(self.window.get_window())))
-        except:
+            self.add_event("ALL", "window=%s, xid=%#x" % (self.window, self.window.get_window().get_xid()))
+        except Exception:
             self.add_event("ALL", "window=%s" % self.window)
 
     def add_event(self, selection, message):
@@ -235,16 +234,31 @@ class ClipboardStateInfoWindow(object):
         self.events.set_text("\n".join(self.log))
 
     def destroy(self, *_args):
-        gtk.main_quit()
+        Gtk.main_quit()
+
+    def show_with_focus(self):
+        force_focus()
+        self.window.show_all()
+        self.window.present()
 
 
 def main():
-    from xpra.platform import program_context
     from xpra.log import enable_color
+    from xpra.platform.gui import init, set_default_icon
     with program_context("Clipboard-Test", "Clipboard Test Tool"):
         enable_color()
-        ClipboardStateInfoWindow()
-        gtk.main()
+
+        set_default_icon("clipboard.png")
+        init()
+
+        from xpra.gtk_common.gobject_compat import register_os_signals
+        def signal_handler(*_args):
+            Gtk.main_quit()
+        register_os_signals(signal_handler, "test window")
+
+        w = ClipboardStateInfoWindow()
+        GLib.idle_add(w.show_with_focus)
+        Gtk.main()
 
 
 if __name__ == "__main__":
