@@ -1,5 +1,5 @@
 # This file is part of Xpra.
-# Copyright (C) 2011-2018 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2011-2019 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008, 2009, 2010 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -7,7 +7,7 @@
 import os
 from struct import pack
 
-from xpra.util import envint, envbool, xor
+from xpra.util import envint, envbool
 from xpra.log import Logger
 from xpra.os_util import hexstr, get_hex_uuid
 from xpra.net.digest import get_salt
@@ -20,7 +20,7 @@ ENCRYPT_FIRST_PACKET = envbool("XPRA_ENCRYPT_FIRST_PACKET", False)
 DEFAULT_IV = os.environ.get("XPRA_CRYPTO_DEFAULT_IV", "0000000000000000")
 DEFAULT_SALT = os.environ.get("XPRA_CRYPTO_DEFAULT_SALT", "0000000000000000")
 DEFAULT_ITERATIONS = envint("XPRA_CRYPTO_DEFAULT_ITERATIONS", 1000)
-DEFAULT_BLOCKSIZE = envint("XPRA_CRYPTO_BLOCKSIZE", 32)      #fixme: can we derive this?
+DEFAULT_BLOCKSIZE = envint("XPRA_CRYPTO_BLOCKSIZE", 32)
 
 #other option "PKCS#7", "legacy"
 PADDING_LEGACY = "legacy"
@@ -32,17 +32,13 @@ PREFERRED_PADDING = os.environ.get("XPRA_CRYPTO_PREFERRED_PADDING", PADDING_PKCS
 assert PREFERRED_PADDING in ALL_PADDING_OPTIONS, "invalid preferred padding: %s" % PREFERRED_PADDING
 assert INITIAL_PADDING in ALL_PADDING_OPTIONS, "invalid padding: %s" % INITIAL_PADDING
 #make sure the preferred one is first in the list:
-PADDING_OPTIONS = [PREFERRED_PADDING]
-for x in ALL_PADDING_OPTIONS:
-    if x not in PADDING_OPTIONS:
-        PADDING_OPTIONS.append(x)
-
-
-try:
-    from xpra.codecs.xor.cyxor import xor_str           #@UnresolvedImport
-    xor = xor_str
-except Exception:
-    log("no accelerated xor", exc_info=True)
+def get_padding_options():
+    options = [PREFERRED_PADDING]
+    for x in ALL_PADDING_OPTIONS:
+        if x not in options:
+            options.append(x)
+    return options
+PADDING_OPTIONS = get_padding_options()
 
 
 ENCRYPTION_CIPHERS = []
@@ -94,10 +90,9 @@ def validate_backend(try_backend):
 def pad(padding, size):
     if padding==PADDING_LEGACY:
         return b" "*size
-    elif padding==PADDING_PKCS7:
+    if padding==PADDING_PKCS7:
         return pack("B", size)*size
-    else:
-        raise Exception("invalid padding: %s" % padding)
+    raise Exception("invalid padding: %s" % padding)
 
 def choose_padding(options):
     for x in options:
@@ -111,11 +106,11 @@ def get_iv():
     #IV = "0000000000000000"
     return IV or get_hex_uuid()[:16]
 
-def get_iterations():
+def get_iterations() -> int:
     return DEFAULT_ITERATIONS
 
 
-def new_cipher_caps(proto, cipher, encryption_key, padding_options):
+def new_cipher_caps(proto, cipher, encryption_key, padding_options) -> dict:
     assert backend
     iv = get_iv()
     key_salt = get_salt()
@@ -131,7 +126,7 @@ def new_cipher_caps(proto, cipher, encryption_key, padding_options):
          "cipher.padding.options"       : PADDING_OPTIONS,
          }
 
-def get_crypto_caps():
+def get_crypto_caps() -> dict:
     if not backend:
         return {}
     caps = {
@@ -145,7 +140,7 @@ def get_encryptor(ciphername, iv, password, key_salt, iterations):
     log("get_encryptor(%s, %s, %s, %s, %s)", ciphername, iv, password, hexstr(key_salt), iterations)
     if not ciphername:
         return None, 0
-    assert iterations>=100
+    assert iterations>=100 and iterations<=10000
     assert ciphername=="AES"
     assert password and iv
     block_size = DEFAULT_BLOCKSIZE
@@ -156,7 +151,7 @@ def get_decryptor(ciphername, iv, password, key_salt, iterations):
     log("get_decryptor(%s, %s, %s, %s, %s)", ciphername, iv, password, hexstr(key_salt), iterations)
     if not ciphername:
         return None, 0
-    assert iterations>=100
+    assert iterations>=100 and iterations<=10000
     assert ciphername=="AES"
     assert password and iv
     block_size = DEFAULT_BLOCKSIZE
