@@ -8,31 +8,39 @@ import os
 import shlex
 import socket
 import signal
+from time import sleep
 from subprocess import PIPE, Popen
 
-from xpra.scripts.main import InitException, InitExit, shellquote, host_target_string
+from xpra.scripts.main import (
+    InitException, InitExit,
+    shellquote, host_target_string,
+    get_pinentry_command, run_pinentry_getpin, run_pinentry_confirm,
+    )
 from xpra.platform.paths import get_ssh_known_hosts_files
 from xpra.platform import get_username
 from xpra.scripts.config import parse_bool
 from xpra.net.bytestreams import SocketConnection, SOCKET_TIMEOUT, ConnectionClosedException
+from xpra.make_thread import start_thread
 from xpra.exit_codes import (
     EXIT_SSH_KEY_FAILURE, EXIT_SSH_FAILURE,
     EXIT_CONNECTION_FAILED,
     )
 from xpra.os_util import (
-    bytestostr, osexpand, load_binary_file,
+    bytestostr, osexpand, load_binary_file, monotonic_time,
     nomodule_context, umask_context, is_main_thread,
+    use_gui_prompt,
     WIN32, OSX, POSIX,
     )
-from xpra.util import envint, envbool, nonl, engs, csv
+from xpra.util import envint, envbool, envfloat, engs, csv
 from xpra.log import Logger, is_debug_enabled
 
 log = Logger("network", "ssh")
 
-INITENV_COMMAND = os.environ.get("XPRA_INITENV_COMMAND", "xpra initenv")
+INITENV_COMMAND = os.environ.get("XPRA_INITENV_COMMAND", "")    #"xpra initenv"
 WINDOW_SIZE = envint("XPRA_SSH_WINDOW_SIZE", 2**27-1)
 TIMEOUT = envint("XPRA_SSH_TIMEOUT", 60)
 SKIP_UI = envbool("XPRA_SKIP_UI", False)
+PINENTRY = envbool("XPRA_SSH_PINENTRY", POSIX and not OSX)
 
 VERIFY_HOSTKEY = envbool("XPRA_SSH_VERIFY_HOSTKEY", True)
 VERIFY_STRICT = envbool("XPRA_SSH_VERIFY_STRICT", False)
@@ -46,6 +54,9 @@ PASSWORD_AUTH = envbool("XPRA_SSH_PASSWORD_AUTH", True)
 PASSWORD_RETRY = envint("XPRA_SSH_PASSWORD_RETRY", 2)
 assert PASSWORD_RETRY>=0
 MAGIC_QUOTES = envbool("XPRA_SSH_MAGIC_QUOTES", True)
+TEST_COMMAND_TIMEOUT = envint("XPRA_SSH_TEST_COMMAND_TIMEOUT", 10)
+EXEC_STDOUT_TIMEOUT = envfloat("XPRA_SSH_EXEC_STDOUT_TIMEOUT", 2)
+EXEC_STDERR_TIMEOUT = envfloat("XPRA_SSH_EXEC_STDERR_TIMEOUT", 0)
 
 
 def keymd5(k) -> str:
@@ -62,68 +73,79 @@ def force_focus():
     from xpra.platform.gui import force_focus as _force_focus
     _force_focus()
 
-def dialog_run(dialog) -> int:
-    from gi.repository import GLib
-    if is_main_thread():
-        force_focus()
-        dialog.show()
-        try:
-            return dialog.run()
-        finally:
-            dialog.destroy()
+def dialog_run(run_fn) -> int:
+    import gi
+    gi.require_version('Gtk', '3.0')
+    from gi.repository import GLib, Gtk
+    log("dialog_run(%s) is_main_thread=%s, main_level=%i", run_fn, is_main_thread(), Gtk.main_level())
+    if is_main_thread() or Gtk.main_level()==0:
+        return run_fn()
+    log("dialog_run(%s) main_depth=%s", run_fn, GLib.main_depth())
     #do a little dance if we're not running in the main thread:
     #block this thread and wait for the main thread to run the dialog
     from threading import Event
     e = Event()
     code = []
     def main_thread_run():
+        log("main_thread_run() calling %s", run_fn)
+        try:
+            code.append(run_fn())
+        finally:
+            e.set()
+    GLib.idle_add(main_thread_run)
+    log("dialog_run(%s) waiting for main thread to run", run_fn)
+    e.wait()
+    log("dialog_run(%s) code=%s", run_fn, code)
+    return code[0]
+
+def do_run_dialog(dialog):
+    try:
         force_focus()
         dialog.show()
         try:
-            r = dialog.run()
+            return dialog.run()
         finally:
             dialog.destroy()
-        code.append(r)
-        e.set()
-    GLib.idle_add(main_thread_run)
-    e.wait()
-    log("dialog_run(%s) code=%s", dialog, code)
-    return code[0]
+    finally:
+        dialog.destroy()
 
 def dialog_pass(title="Password Input", prompt="enter password", icon="") -> str:
-    from xpra.client.gtk_base.pass_dialog import PasswordInputDialogWindow
-    dialog = PasswordInputDialogWindow(title, prompt, icon)
-    try:
-        if dialog_run(dialog)==0:
-            return dialog.get_password()
-        return None
-    finally:
-        dialog.destroy()
+    log("dialog_pass%s PINENTRY=%s", (title, prompt, icon), PINENTRY)
+    if PINENTRY:
+        pinentry_cmd = get_pinentry_command()
+        if pinentry_cmd:
+            return run_pinentry_getpin(pinentry_cmd, title, prompt)
+    def password_input_run():
+        from xpra.client.gtk_base.pass_dialog import PasswordInputDialogWindow
+        dialog = PasswordInputDialogWindow(title, prompt, icon)
+        return do_run_dialog(dialog)
+    return dialog_run(password_input_run)
 
 def dialog_confirm(title, prompt, qinfo=(), icon="", buttons=(("OK", 1),)) -> int:
-    from xpra.client.gtk_base.confirm_dialog import ConfirmDialogWindow
-    dialog = ConfirmDialogWindow(title, prompt, qinfo, icon, buttons)
-    try:
-        r = dialog_run(dialog)
-    finally:
-        dialog.destroy()
-    return r
+    def confirm_run():
+        from xpra.client.gtk_base.confirm_dialog import ConfirmDialogWindow
+        dialog = ConfirmDialogWindow(title, prompt, qinfo, icon, buttons)
+        return do_run_dialog(dialog)
+    return dialog_run(confirm_run)
 
 
-def confirm_key(info=()) -> bool:
+def confirm_key(info=(), title="Confirm Key", prompt="Are you sure you want to continue connecting?") -> bool:
     if SKIP_UI:
         return False
     from xpra.platform.paths import get_icon_filename
-    from xpra.os_util import use_tty
-    if not use_tty():
+    if PINENTRY:
+        pinentry_cmd = get_pinentry_command()
+        if pinentry_cmd:
+            messages = list(info)+["", prompt]
+            return run_pinentry_confirm(pinentry_cmd, title, "%0A".join(messages))
+    if use_gui_prompt():
         icon = get_icon_filename("authentication", "png") or ""
-        prompt = "Are you sure you want to continue connecting?"
-        code = dialog_confirm("Confirm Key", prompt, info, icon, buttons=[("yes", 200), ("NO", 201)])
+        code = dialog_confirm(title, prompt, info, icon, buttons=[("yes", 200), ("NO", 201)])
         log("dialog return code=%s", code)
         r = code==200
         log.info("host key %sconfirmed", ["not ", ""][r])
         return r
-    log("confirm_key(%s) will use stdin prompt", nonl(info))
+    log("confirm_key(%r) will use stdin prompt", info)
     prompt = "Are you sure you want to continue connecting (yes/NO)? "
     sys.stderr.write(os.linesep.join(info)+os.linesep+prompt)
     try:
@@ -136,8 +158,7 @@ def input_pass(prompt) -> str:
     if SKIP_UI:
         return None
     from xpra.platform.paths import get_icon_filename
-    from xpra.os_util import use_tty
-    if not use_tty():
+    if PINENTRY or use_gui_prompt():
         icon = get_icon_filename("authentication", "png") or ""
         return dialog_pass("Password Input", prompt, icon)
     from getpass import getpass
@@ -149,29 +170,28 @@ def input_pass(prompt) -> str:
 
 class SSHSocketConnection(SocketConnection):
 
-    def __init__(self, ssh_channel, sock, sockname, peername, target, info=None):
-        super().__init__(ssh_channel, sockname, peername, target, "ssh", info)
+    def __init__(self, ssh_channel, sock, sockname, peername, target, info=None, socket_options=None):
         self._raw_socket = sock
+        super().__init__(ssh_channel, sockname, peername, target, "ssh", info, socket_options)
+
+    def get_raw_socket(self):
+        return self._raw_socket
 
     def start_stderr_reader(self):
-        from xpra.make_thread import start_thread
         start_thread(self._stderr_reader, "ssh-stderr-reader", daemon=True)
 
     def _stderr_reader(self):
         #stderr = self._socket.makefile_stderr(mode="rb", bufsize=1)
         chan = self._socket
         stderr = chan.makefile_stderr("rb", 1)
-        errs = []
         while self.active:
             v = stderr.readline()
             if not v:
-                log("SSH EOF on stderr of %s", chan.get_name())
-                return
-            errs.append(bytestostr(v.rstrip(b"\n\r")))
-        if errs:
-            log.warn("remote SSH stderr:")
-            for e in errs:
-                log.warn(" %s", e)
+                log.info("SSH EOF on stderr of %s", chan.get_name())
+                break
+            s = bytestostr(v.rstrip(b"\n\r"))
+            if s:
+                log.info(" SSH: %r", s)
 
     def peek(self, n):
         if not self._raw_socket:
@@ -267,7 +287,11 @@ def ssh_paramiko_connect_to(display_desc):
         if os.path.exists(user_config_file):
             with open(user_config_file) as f:
                 ssh_config.parse(f)
-            log("parsed user config '%s': %i hosts found", user_config_file, len(ssh_config.get_hostnames()))
+            log("parsed user config '%s'", user_config_file)
+            try:
+                log("%i hosts found", len(ssh_config.get_hostnames()))
+            except KeyError:
+                pass
             host_config = ssh_config.lookup(host)
             if host_config:
                 log("got host config for '%s': %s", host, host_config)
@@ -276,8 +300,8 @@ def ssh_paramiko_connect_to(display_desc):
                 cport = host_config.get("port", port)
                 try:
                     port = int(cport)
-                except ValueError:
-                    raise InitExit(EXIT_SSH_FAILURE, "invalid ssh port specified: '%s'" % cport)
+                except (TypeError, ValueError):
+                    raise InitExit(EXIT_SSH_FAILURE, "invalid ssh port specified: '%s'" % cport) from None
                 proxycommand = host_config.get("proxycommand")
                 if proxycommand:
                     log("found proxycommand='%s' for host '%s'", proxycommand, chost)
@@ -305,7 +329,7 @@ def ssh_paramiko_connect_to(display_desc):
                     chan = paramiko_run_remote_xpra(transport, proxy_command, remote_xpra, socket_dir, display_as_args)
                     peername = (chost, cport)
                     conn = SSHProxyCommandConnection(chan, peername, peername, socket_info)
-                    conn.target = host_target_string("ssh", cusername, chost, cport, display)
+                    conn.target = host_target_string("ssh", cusername, chost, port, display)
                     conn.timeout = SOCKET_TIMEOUT
                     conn.start_stderr_reader()
                     conn.process = (sock.process, "ssh", cmd)
@@ -347,7 +371,7 @@ def ssh_paramiko_connect_to(display_desc):
                 transport.start_client()
             except SSHException as e:
                 log("start_client()", exc_info=True)
-                raise InitExit(EXIT_SSH_FAILURE, "SSH negotiation failed: %s" % e)
+                raise InitExit(EXIT_SSH_FAILURE, "SSH negotiation failed: %s" % e) from None
             do_ssh_paramiko_connect_to(transport, host,
                                        username, password,
                                        host_config or ssh_config.lookup("*"),
@@ -459,8 +483,8 @@ def do_ssh_paramiko_connect_to(transport, host, username, password, host_config=
                     dnscheck = check_host_key(host, host_key)
                 except ImportError as e:
                     log("verifyhostkeydns failed", exc_info=True)
-                    log.warn("Warning: cannot check SSHFP DNS records")
-                    log.warn(" %s", e)
+                    log.info("cannot check SSHFP DNS records")
+                    log.info(" %s", e)
             log("dnscheck=%s", dnscheck)
             def adddnscheckinfo(q):
                 if dnscheck is not True:
@@ -506,7 +530,7 @@ keymd5(host_key),
             else:
                 assert (not keys) or (host_key.get_name() not in keys)
                 if not keys:
-                    log.warn("Warning: unknown SSH host")
+                    log.warn("Warning: unknown SSH host '%s'", host)
                 else:
                     log.warn("Warning: unknown %s SSH host key", keyname())
                 qinfo = [
@@ -564,7 +588,8 @@ keymd5(host_key),
                         log("authenticated using agent and key '%s'", keymd5(agent_key))
                         break
                 except SSHException:
-                    log("agent key '%s' rejected", keymd5(agent_key), exc_info=True)
+                    log.info("SSH agent key '%s' rejected for user '%s'", keymd5(agent_key), username)
+                    log("%s%s", transport.auth_publickey, (username, agent_key), exc_info=True)
             if not transport.is_authenticated():
                 log.info("agent authentication failed, tried %i key%s", len(agent_keys), engs(agent_keys))
 
@@ -577,7 +602,14 @@ keymd5(host_key),
             log("trying '%s'", keyfile_path)
             key = None
             import paramiko
-            for pkey_classname in ("RSA", "DSS", "ECDSA", "Ed25519"):
+            try_key_formats = ()
+            for kf in ("RSA", "DSS", "ECDSA", "Ed25519"):
+                if keyfile_path.lower().endswith(kf.lower()):
+                    try_key_formats = (kf, )
+                    break
+            if not try_key_formats:
+                try_key_formats = ("RSA", "DSS", "ECDSA", "Ed25519")
+            for pkey_classname in try_key_formats:
                 pkey_class = getattr(paramiko, "%sKey" % pkey_classname, None)
                 if pkey_class is None:
                     log("no %s key type", pkey_classname)
@@ -721,7 +753,7 @@ keymd5(host_key),
 
 def paramiko_run_test_command(transport, cmd):
     from paramiko import SSHException
-    log("run_test_command('%s')", cmd)
+    log("paramiko_run_test_command(transport, %r)", cmd)
     try:
         chan = transport.open_session(window_size=None, max_packet_size=0, timeout=60)
         chan.set_name("find %s" % cmd)
@@ -729,10 +761,29 @@ def paramiko_run_test_command(transport, cmd):
         log("open_session", exc_info=True)
         raise InitExit(EXIT_SSH_FAILURE, "failed to open SSH session: %s" % e) from None
     chan.exec_command(cmd)
-    out = chan.makefile().read()
-    err = chan.makefile_stderr().read()
+    log("exec_command returned")
+    start = monotonic_time()
+    while not chan.exit_status_ready():
+        if monotonic_time()-start>TEST_COMMAND_TIMEOUT:
+            chan.close()
+            raise InitException("SSH test command '%s' timed out" % cmd)
+        log("exit status is not ready yet, sleeping")
+        sleep(0.01)
     code = chan.recv_exit_status()
-    log("exec_command('%s')=%s, out=%s, err=%s", cmd, code, out, err)
+    log("exec_command('%s')=%s", cmd, code)
+    def chan_read(read_fn):
+        try:
+            return read_fn()
+        except socket.error:
+            log("chan_read(%s)", read_fn, exc_info=True)
+            return b""
+    #don't wait too long for the data:
+    chan.settimeout(EXEC_STDOUT_TIMEOUT)
+    out = chan_read(chan.makefile().readline)
+    log("exec_command out=%r", out)
+    chan.settimeout(EXEC_STDERR_TIMEOUT)
+    err = chan_read(chan.makefile_stderr().readline)
+    log("exec_command err=%r", err)
     chan.close()
     return out, err, code
 
@@ -750,14 +801,13 @@ def paramiko_run_remote_xpra(transport, xpra_proxy_command=None, remote_xpra=Non
             if r[2]!=0:
                 continue
             #let's find where xpra is installed:
-            r = rtc("FOR /F \"usebackq tokens=3*\" %A IN (`REG QUERY \"HKEY_LOCAL_MACHINE\Software\Xpra\" /v InstallPath`) DO (echo %A %B)")
+            r = rtc("FOR /F \"usebackq tokens=3*\" %A IN (`REG QUERY \"HKEY_LOCAL_MACHINE\\Software\\Xpra\" /v InstallPath`) DO (echo %A %B)")  #pylint: disable=line-too-long
             if r[2]==0:
                 #found in registry:
                 lines = r[0].splitlines()
                 installpath = bytestostr(lines[-1])
                 xpra_cmd = "%s\\%s" % (installpath, xpra_cmd)
                 xpra_cmd = xpra_cmd.replace("\\", "\\\\")
-                log("using '%s'", xpra_cmd)
         elif xpra_cmd.endswith(".exe"):
             #assume this path exists
             pass
@@ -766,6 +816,16 @@ def paramiko_run_remote_xpra(transport, xpra_proxy_command=None, remote_xpra=Non
             r = rtc("which %s" % xpra_cmd)
             if r[2]!=0:
                 continue
+            out = r[0]
+            if out:
+                #use the actual path returned by 'command -v':
+                if not isinstance(out, str):
+                    out = bytestostr(out)
+                try:
+                    xpra_cmd = out.splitlines()[-1].rstrip("\n\r ").lstrip("\t ")
+                except Exception as e:
+                    log("cannot get command from %r: %s", xpra_cmd, e)
+        log("adding xpra_cmd='%s'", xpra_cmd)
         if xpra_cmd in tried:
             continue
         tried.add(xpra_cmd)
@@ -790,15 +850,15 @@ def paramiko_run_remote_xpra(transport, xpra_proxy_command=None, remote_xpra=Non
             log("channel exec_command(%s)" % cmd)
             chan.exec_command(cmd)
             return chan
-    raise Exception("all SSH remote proxy commands have failed")
+    raise Exception("all SSH remote proxy commands have failed - is xpra installed on the remote host?")
 
 
 def ssh_connect_failed(_message):
     #by the time ssh fails, we may have entered the gtk main loop
     #(and more than once thanks to the clipboard code..)
-    if "gtk" in sys.modules or "gi.repository.Gtk" in sys.modules:
-        from xpra.gtk_common.quit import gtk_main_quit_really
-        gtk_main_quit_really()
+    if "gi.repository.Gtk" in sys.modules:
+        from gi.repository import Gtk
+        Gtk.main_quit()
 
 
 def ssh_exec_connect_to(display_desc, opts=None, debug_cb=None, ssh_fail_cb=ssh_connect_failed):
@@ -815,10 +875,12 @@ def ssh_exec_connect_to(display_desc, opts=None, debug_cb=None, ssh_fail_cb=ssh_
             startupinfo = STARTUPINFO()
             startupinfo.dwFlags |= STARTF_USESHOWWINDOW
             startupinfo.wShowWindow = 0     #aka win32.con.SW_HIDE
-            kwargs["startupinfo"] = startupinfo
             flags = CREATE_NEW_PROCESS_GROUP | CREATE_NEW_CONSOLE
-            kwargs["creationflags"] = flags
-            kwargs["stderr"] = PIPE
+            kwargs.update({
+                "startupinfo"   : startupinfo,
+                "creationflags" : flags,
+                "stderr"        : PIPE,
+                })
         elif not display_desc.get("exit_ssh", False) and not OSX:
             kwargs["start_new_session"] = True
         remote_xpra = display_desc["remote_xpra"]
@@ -846,8 +908,8 @@ def ssh_exec_connect_to(display_desc, opts=None, debug_cb=None, ssh_fail_cb=ssh_
             remote_cmd = INITENV_COMMAND + ";" + remote_cmd
         #how many times we need to escape the remote command string
         #depends on how many times the ssh command is parsed
-        if MAGIC_QUOTES:
-            nssh = sum(int(x=="ssh") for x in cmd)
+        nssh = sum(int(x=="ssh") for x in cmd)
+        if nssh>=2 and MAGIC_QUOTES:
             for _ in range(nssh):
                 remote_cmd = shlex.quote(remote_cmd)
         else:
@@ -964,4 +1026,24 @@ def ssh_exec_connect_to(display_desc, opts=None, debug_cb=None, ssh_fail_cb=ssh_
     conn.endpoint = host_target_string("ssh", username, host, port, display)
     conn.timeout = 0            #taken care of by abort_test
     conn.process = (child, "ssh", cmd)
+    if kwargs.get("stderr")==PIPE:
+        def stderr_reader():
+            errs = []
+            while child.poll() is None:
+                try:
+                    v = child.stderr.readline()
+                except OSError:
+                    log("stderr_reader()", exc_info=True)
+                    break
+                if not v:
+                    log("SSH EOF on stderr of %s", cmd)
+                    break
+                s = bytestostr(v.rstrip(b"\n\r"))
+                if s:
+                    errs.append(s)
+            if errs:
+                log.warn("remote SSH stderr:")
+                for e in errs:
+                    log.warn(" %s", e)
+        start_thread(stderr_reader, "ssh-stderr-reader", daemon=True)
     return conn
