@@ -1,7 +1,7 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2013-2018 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2013-2021 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -13,52 +13,13 @@ import uuid
 import time
 import struct
 import binascii
-
-#hide some ugly python3 compat:
-try:
-    import _thread as thread            #@UnresolvedImport @UnusedImport (python3)
-except ImportError:
-    import thread                       #@Reimport @UnusedImport
-assert thread
-
-try:
-    from queue import Queue             #@UnresolvedImport @UnusedImport (python3)
-except ImportError:
-    from Queue import Queue             #@Reimport @UnusedImport
-assert Queue
-
-try:
-    import builtins                     #@UnresolvedImport @UnusedImport (python3)
-except ImportError:
-    import __builtin__ as builtins      #@Reimport @UnusedImport
-assert builtins
-_buffer = builtins.__dict__.get("buffer")
-
+import threading
 
 SIGNAMES = {}
 for signame in (sig for sig in dir(signal) if sig.startswith("SIG") and not sig.startswith("SIG_")):
     SIGNAMES[getattr(signal, signame)] = signame
 
 
-#use cStringIO, fallback to StringIO,
-#and python3 is making life more difficult yet again:
-try:
-    from io import BytesIO as BytesIOClass              #@UnusedImport
-except ImportError:
-    try:
-        from cStringIO import StringIO as BytesIOClass  #@Reimport @UnusedImport
-    except ImportError:
-        from StringIO import StringIO as BytesIOClass   #@Reimport @UnusedImport
-assert BytesIOClass
-try:
-    from StringIO import StringIO as StringIOClass      #@UnusedImport
-except ImportError:
-    try:
-        from cStringIO import StringIO as StringIOClass #@Reimport @UnusedImport
-    except ImportError:
-        from io import StringIO as StringIOClass        #@Reimport @UnusedImport
-assert StringIOClass
-
 WIN32 = sys.platform.startswith("win")
 OSX = sys.platform.startswith("darwin")
 LINUX = sys.platform.startswith("linux")
@@ -67,30 +28,97 @@ OPENBSD = sys.platform.startswith("openbsd")
 FREEBSD  = sys.platform.startswith("freebsd")
 
 POSIX = os.name=="posix"
-PYTHON2 = sys.version_info[0]==2
-PYTHON3 = sys.version_info[0]==3
 
 BITS = struct.calcsize(b"P")*8
 
 
-if PYTHON2:
-    def strtobytes(x):
-        return str(x)
-    def bytestostr(x):
-        return str(x)
-    def hexstr(v):
-        return binascii.hexlify(str(v))
-else:
-    def strtobytes(x):
-        if isinstance(x, bytes):
-            return x
-        return str(x).encode("latin1")
-    def bytestostr(x):
-        if isinstance(x, bytes):
-            return x.decode("latin1")
-        return str(x)
-    def hexstr(v):
-        return bytestostr(binascii.hexlify(strtobytes(v)))
+main_thread = threading.current_thread()
+def is_main_thread():
+    return threading.current_thread()==main_thread
+
+
+def get_frame_info(ignore_threads=()):
+    info = {
+        "count"        : threading.active_count() - len(ignore_threads),
+        }
+    try:
+        import traceback
+        def nn(x):
+            if x is None:
+                return ""
+            return str(x)
+        thread_ident = {}
+        for t in threading.enumerate():
+            if t not in ignore_threads:
+                thread_ident[t.ident] = t.getName()
+            else:
+                thread_ident[t.ident] = None
+        thread_ident.update({
+                threading.current_thread().ident    : "info",
+                main_thread.ident                   : "main",
+                })
+        frames = sys._current_frames()  #pylint: disable=protected-access
+        stack = None
+        for i,frame_pair in enumerate(frames.items()):
+            stack = traceback.extract_stack(frame_pair[1])
+            tident = thread_ident.get(frame_pair[0], "unknown")
+            if tident is None:
+                continue
+            #sanitize stack to prevent None values (which cause encoding errors with the bencoder)
+            sanestack = []
+            for e in stack:
+                sanestack.append(tuple([nn(x) for x in e]))
+            info[i] = {
+                ""          : tident,
+                "stack"     : sanestack,
+                }
+        del frames, stack
+    except Exception as e:
+        get_util_logger().error("failed to get frame info: %s", e)
+    return info
+
+def get_info_env():
+    filtered_env = os.environ.copy()
+    if filtered_env.get('XPRA_PASSWORD'):
+        filtered_env['XPRA_PASSWORD'] = "*****"
+    if filtered_env.get('XPRA_ENCRYPTION_KEY'):
+        filtered_env['XPRA_ENCRYPTION_KEY'] = "*****"
+    return filtered_env
+
+def get_sysconfig_info():
+    import sysconfig
+    sysinfo = {}
+    log = get_util_logger()
+    for attr in (
+        "platform",
+        "python-version",
+        "config-vars",
+        "paths",
+        ):
+        fn = "get_%s" % attr.replace("-", "_")
+        getter = getattr(sysconfig, fn, None)
+        if getter:
+            try:
+                sysinfo[attr] = getter()  #pylint: disable=not-callable
+            except ModuleNotFoundError:
+                log("sysconfig.%s", fn, exc_info=True)
+                if attr=="config-vars" and WIN32:
+                    continue
+                log.warn("Warning: failed to collect %s sysconfig information", attr)
+            except Exception:
+                log.error("Error calling sysconfig.%s", fn, exc_info=True)
+    return sysinfo
+
+def strtobytes(x) -> bytes:
+    if isinstance(x, bytes):
+        return x
+    return str(x).encode("latin1")
+def bytestostr(x) -> str:
+    if isinstance(x, (bytes, bytearray)):
+        return x.decode("latin1")
+    return str(x)
+def hexstr(v) -> str:
+    return bytestostr(binascii.hexlify(strtobytes(v)))
 
 
 util_logger = None
@@ -101,35 +129,27 @@ def get_util_logger():
         util_logger = Logger("util")
     return util_logger
 
-def memoryview_to_bytes(v):
+def memoryview_to_bytes(v) -> bytes:
     if isinstance(v, bytes):
         return v
     if isinstance(v, memoryview):
         return v.tobytes()
-    if _buffer and isinstance(v, _buffer):
-        return bytes(v)
     if isinstance(v, bytearray):
         return bytes(v)
     return strtobytes(v)
 
 
-def setsid():
-    #run in a new session
-    if POSIX:
-        os.setsid()
-
-
-def getuid():
+def getuid() -> int:
     if POSIX:
         return os.getuid()
     return 0
 
-def getgid():
+def getgid() -> int:
     if POSIX:
         return os.getgid()
     return 0
 
-def get_shell_for_uid(uid):
+def get_shell_for_uid(uid) -> str:
     if POSIX:
         from pwd import getpwuid
         try:
@@ -138,7 +158,7 @@ def get_shell_for_uid(uid):
             pass
     return ""
 
-def get_username_for_uid(uid):
+def get_username_for_uid(uid) -> str:
     if POSIX:
         from pwd import getpwuid
         try:
@@ -147,7 +167,7 @@ def get_username_for_uid(uid):
             pass
     return ""
 
-def get_home_for_uid(uid):
+def get_home_for_uid(uid) -> str:
     if POSIX:
         from pwd import getpwuid
         try:
@@ -162,7 +182,7 @@ def get_groups(username):
         return [gr.gr_name for gr in grp.getgrall() if username in gr.gr_mem]
     return []
 
-def get_group_id(group):
+def get_group_id(group) -> int:
     try:
         import grp      #@UnresolvedImport
         gr = grp.getgrnam(group)
@@ -176,11 +196,8 @@ def platform_release(release):
         SYSTEMVERSION_PLIST = "/System/Library/CoreServices/SystemVersion.plist"
         try:
             import plistlib
-            if PYTHON2:
-                pl = plistlib.readPlist('/System/Library/CoreServices/SystemVersion.plist')
-            else:
-                with open(SYSTEMVERSION_PLIST, "rb") as f:
-                    pl = plistlib.load(f)           #@UndefinedVariable
+            with open(SYSTEMVERSION_PLIST, "rb") as f:
+                pl = plistlib.load(f)           #@UndefinedVariable
             return pl['ProductUserVisibleVersion']
         except Exception as e:
             get_util_logger().debug("platform_release(%s)", release, exc_info=True)
@@ -190,7 +207,7 @@ def platform_release(release):
     return release
 
 
-def platform_name(sys_platform, release=None):
+def platform_name(sys_platform=sys.platform, release=None) -> str:
     if not sys_platform:
         return "unknown"
     PLATFORMS = {"win32"    : "Microsoft Windows",
@@ -214,25 +231,25 @@ def platform_name(sys_platform, release=None):
     return rel(sys_platform)
 
 
-def get_rand_chars(l=16, chars=b"0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"):
+def get_rand_chars(l=16, chars=b"0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ") -> bytes:
     import random
-    return b"".join(chars[random.randint(0, len(chars)-1)] for _ in range(l))
+    return b"".join(chars[random.randint(0, len(chars)-1):][:1] for _ in range(l))
 
-def get_hex_uuid():
+def get_hex_uuid() -> str:
     return uuid.uuid4().hex
 
-def get_int_uuid():
+def get_int_uuid() -> int:
     return uuid.uuid4().int
 
-def get_machine_id():
+def get_machine_id() -> str:
     """
         Try to get uuid string which uniquely identifies this machine.
         Warning: only works on posix!
         (which is ok since we only used it on posix at present)
     """
-    v = u""
+    v = ""
     if POSIX:
-        for filename in ["/etc/machine-id", "/var/lib/dbus/machine-id"]:
+        for filename in ("/etc/machine-id", "/var/lib/dbus/machine-id"):
             v = load_binary_file(filename)
             if v is not None:
                 break
@@ -240,7 +257,7 @@ def get_machine_id():
         v = uuid.getnode()
     return bytestostr(v).strip("\n\r")
 
-def get_user_uuid():
+def get_user_uuid() -> str:
     """
         Try to generate a uuid string which is unique to this user.
         (relies on get_machine_id to uniquely identify a machine)
@@ -270,54 +287,96 @@ except (ImportError, AssertionError):
     monotonic_time = time.time
 
 
-def is_X11():
+def is_X11() -> bool:
     if OSX or WIN32:
         return False
-    if PYTHON2:
+    try:
+        from xpra.x11.gtk3.gdk_bindings import is_X11_Display   #@UnresolvedImport
+        return is_X11_Display()
+    except ImportError:
+        get_util_logger().debug("failed to load x11 bindings", exc_info=True)
         return True
-    from xpra.x11.gtk3.gdk_bindings import is_X11_Display   #@UnresolvedImport
-    return is_X11_Display()
 
-def is_Wayland():
-    return os.environ.get("WAYLAND_DISPLAY") or os.environ.get("XDG_SESSION_TYPE")=="wayland"
+saved_env = os.environ.copy()
+def is_Wayland() -> bool:
+    return _is_Wayland(saved_env)
+
+def _is_Wayland(env : dict) -> bool:
+    backend = env.get("GDK_BACKEND", "")
+    if backend=="wayland":
+        return True
+    return backend!="x11" and (
+        bool(env.get("WAYLAND_DISPLAY")) or env.get("XDG_SESSION_TYPE")=="wayland"
+        )
 
 
-def is_distribution_variant(variant=b"Debian", os_file="/etc/os-release"):
+def is_distribution_variant(variant=b"Debian") -> bool:
     if not POSIX:
         return False
     try:
-        if get_linux_distribution()[0]==variant:
-            return True
-    except:
+        v = load_os_release_file()
+        return any(l.find(variant)>=0 for l in v.splitlines() if l.startswith(b"NAME="))
+    except Exception:
         pass
     try:
-        v = load_binary_file(os_file)
-        return any(l.find(variant)>=0 for l in v.splitlines() if l.startswith(b"NAME="))
-    except:
-        return False
+        if variant==b"RedHat" and get_linux_distribution()[0].startswith(variant):
+            return True
+        if get_linux_distribution()[0]==variant:
+            return True
+    except Exception:
+        pass
+    return False
 
-def is_Ubuntu():
+def get_distribution_version_id() -> bool:
+    if not POSIX:
+        return ""
+    try:
+        v = load_os_release_file()
+        for line in v.splitlines():
+            l = line.decode()
+            if l.startswith("VERSION_ID="):
+                return l.split("=", 1)[1].strip('"')
+    except Exception:
+        pass
+    return ""
+
+os_release_file_data = False
+def load_os_release_file() -> bytes:
+    global os_release_file_data
+    if os_release_file_data is False:
+        try:
+            os_release_file_data = load_binary_file("/etc/os-release")
+        except OSError: # pragma: no cover
+            os_release_file_data = None
+    return os_release_file_data
+
+def is_Ubuntu() -> bool:
     return is_distribution_variant(b"Ubuntu")
 
-def is_Debian():
+def is_Debian() -> bool:
     return is_distribution_variant(b"Debian")
 
-def is_Raspbian():
+def is_Raspbian() -> bool:
     return is_distribution_variant(b"Raspbian")
 
-def is_Fedora():
+def is_Fedora() -> bool:
     return is_distribution_variant(b"Fedora")
 
-def is_Arch():
+def is_Arch() -> bool:
     return is_distribution_variant(b"Arch")
 
-def is_CentOS():
+def is_CentOS() -> bool:
     return is_distribution_variant(b"CentOS")
 
-def is_RedHat():
+def is_RedHat() -> bool:
     return is_distribution_variant(b"RedHat")
 
 
+def is_arm() -> bool:
+    import platform
+    return platform.uname()[4].startswith("arm")
+
+
 _linux_distribution = None
 def get_linux_distribution():
     global _linux_distribution
@@ -327,14 +386,14 @@ def get_linux_distribution():
         import subprocess
         cmd = ["lsb_release", "-a"]
         try:
-            p = subprocess.Popen(cmd, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+            p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
             out = p.communicate()[0]
             assert p.returncode==0 and out
-        except:
+        except Exception:
             try:
-                from xpra.scripts.config import python_platform
-                _linux_distribution = python_platform.linux_distribution()
-            except:
+                import platform
+                _linux_distribution = platform.linux_distribution()  #@UndefinedVariable, pylint: disable=deprecated-method, no-member
+            except Exception:
                 _linux_distribution = ("unknown", "unknown", "unknown")
         else:
             d = {}
@@ -348,27 +407,30 @@ def get_linux_distribution():
                 return tuple([bytestostr(x) for x in v])
     return _linux_distribution
 
-def getUbuntuVersion():
-    distro = get_linux_distribution()
-    if distro and len(distro)==3 and distro[0]=="Ubuntu":
-        ur = distro[1]  #ie: "12.04"
-        try:
-            return tuple(int(x) for x in ur.split("."))  #ie: (12, 4)
-        except ValueError:
-            pass
-    return ()
+def is_unity() -> bool:
+    d = os.environ.get("XDG_CURRENT_DESKTOP", "").lower()
+    return d.find("unity")>=0 or d.find("ubuntu")>=0
 
-def is_unity():
-    return os.environ.get("XDG_CURRENT_DESKTOP", "").lower().find("unity")>=0
-
-def is_gnome():
+def is_gnome() -> bool:
+    if os.environ.get("XDG_SESSION_DESKTOP", "").split("-", 1)[0] in ("i3", "ubuntu", ):
+        #"i3-gnome" is not really gnome... ie: the systray does work!
+        return False
     return os.environ.get("XDG_CURRENT_DESKTOP", "").lower().find("gnome")>=0
 
-def is_kde():
+def is_kde() -> bool:
     return os.environ.get("XDG_CURRENT_DESKTOP", "").lower().find("kde")>=0
 
 
-def is_WSL():
+def get_loaded_kernel_modules(*modlist):
+    loaded = []
+    if LINUX and os.path.exists("/sys/module"):
+        for mod in modlist:
+            if os.path.exists("/sys/module/%s" % mod):  # pragma: no cover
+                loaded.append(mod)
+    return loaded
+
+
+def is_WSL() -> bool:
     if not POSIX:
         return False
     r = None
@@ -379,61 +441,88 @@ def is_WSL():
     return r is not None and r.find(b"Microsoft")>=0
 
 
-def get_generic_os_name():
+def get_generic_os_name() -> str:
+    return do_get_generic_os_name().lower()
+
+def do_get_generic_os_name() -> str:
     for k,v in {
-            "linux"     : "linux",
-            "darwin"    : "osx",
-            "win"       : "win32",
-            "freebsd"   : "freebsd",
+            "linux"     : "Linux",
+            "darwin"    : "MacOS",
+            "win"       : "Win32",
+            "freebsd"   : "FreeBSD",
             }.items():
         if sys.platform.startswith(k):
             return v
-    return sys.platform
-
-def get_cpu_count():
-    #sensible default:
-    cpus = 2
-    try:
-        try:
-            #python3:
-            cpus = os.cpu_count()
-        except AttributeError:
-            #python2:
-            import multiprocessing
-            cpus = multiprocessing.cpu_count()
-    except:
-        pass
-    return cpus
+    return sys.platform     # pragma: no cover
 
 
-def filedata_nocrlf(filename):
+def filedata_nocrlf(filename) -> str:
     v = load_binary_file(filename)
     if v is None:
         get_util_logger().error("failed to load '%s'", filename)
         return None
-    return v.strip("\n\r")
+    return v.strip(b"\n\r")
 
-def load_binary_file(filename):
+def load_binary_file(filename) -> bytes:
     if not os.path.exists(filename):
         return None
     try:
         with open(filename, "rb") as f:
             return f.read()
-    except Exception as e:
+    except Exception as e:  # pragma: no cover
         get_util_logger().warn("Warning: failed to load '%s':", filename)
         get_util_logger().warn(" %s", e)
         return None
 
+def parse_encoded_bin_data(data):
+    if not data:
+        return None
+    header = bytestostr(data).lower()[:10]
+    if header.startswith("0x"):
+        return binascii.unhexlify(data[2:])
+    import base64
+    if header.startswith("b64:"):
+        return base64.b64decode(data[4:])
+    if header.startswith("base64:"):
+        return base64.b64decode(data[7:])
+    try:
+        return binascii.unhexlify(data)
+    except (TypeError, binascii.Error):
+        try:
+            return base64.b64decode(data)
+        except Exception:
+            pass
+    return None
+
+
 #here so we can override it when needed
 def force_quit(status=1):
-    os._exit(status)
+    os._exit(status)  #pylint: disable=protected-access
+
+
+def no_idle(fn, *args, **kwargs):
+    fn(*args, **kwargs)
+def register_SIGUSR_signals(idle_add=no_idle):
+    if not os.name=="posix":
+        return
+    from xpra.util import dump_all_frames, dump_gc_frames
+    def sigusr1(*_args):
+        log = get_util_logger().info
+        log("SIGUSR1")
+        idle_add(dump_all_frames, log)
+    def sigusr2(*_args):
+        log = get_util_logger().info
+        log("SIGUSR2")
+        idle_add(dump_gc_frames, log)
+    signal.signal(signal.SIGUSR1, sigusr1)
+    signal.signal(signal.SIGUSR2, sigusr2)
 
 
 def livefds():
     live = set()
     try:
         MAXFD = os.sysconf("SC_OPEN_MAX")
-    except ValueError:
+    except (ValueError, AttributeError):
         MAXFD = 256
     for fd in range(0, MAXFD):
         try:
@@ -445,18 +534,6 @@ def livefds():
                 live.add(fd)
     return live
 
-def close_fds(excluding=(0, 1, 2)):
-    try:
-        MAXFD = os.sysconf("SC_OPEN_MAX")
-    except ValueError:
-        MAXFD = 256
-    for i in range(0, MAXFD):
-        if i not in excluding:
-            try:
-                os.close(i)
-            except (IOError, OSError):
-                pass
-
 def get_all_fds():
     fd_dirs = ["/dev/fd", "/proc/self/fd"]
     fds = []
@@ -486,28 +563,41 @@ def close_all_fds(exceptions=()):
 
 def use_tty():
     from xpra.util import envbool
-    NOTTY = envbool("XPRA_NOTTY", False)
-    stdin = sys.stdin
-    return not os.environ.get("MSYSCON") and not NOTTY and stdin and stdin.isatty()
+    if envbool("XPRA_NOTTY", False):
+        return False
+    from xpra.platform.gui import use_stdin
+    return use_stdin()
+
+def use_gui_prompt():
+    return WIN32 or OSX or not use_tty()
 
 
 def shellsub(s, subs=None):
     """ shell style string substitution using the dictionary given """
     if subs:
         for var,value in subs.items():
-            s = s.replace("$%s" % var, str(value))
-            s = s.replace("${%s}" % var, str(value))
+            try:
+                if isinstance(s, bytes):
+                    s = s.replace(("$%s" % var).encode(), str(value).encode())
+                    s = s.replace(("${%s}" % var).encode(), str(value).encode())
+                else:
+                    s = s.replace("$%s" % var, str(value))
+                    s = s.replace("${%s}" % var, str(value))
+            except (TypeError, ValueError):
+                raise Exception("failed to substitute '%s' with value '%s' (%s) in '%s'" % (
+                    var, value, type(value), s)) from None
     return s
 
 
-def osexpand(s, actual_username="", uid=0, gid=0, subs={}):
+def osexpand(s, actual_username="", uid=0, gid=0, subs=None):
+    if not s:
+        return s
     def expanduser(s):
         if actual_username and s.startswith("~/"):
             #replace "~/" with "~$actual_username/"
             return os.path.expanduser("~%s/%s" % (actual_username, s[2:]))
         return os.path.expanduser(s)
-    from collections import OrderedDict
-    d = OrderedDict(subs)
+    d = dict(subs or {})
     d.update({
         "PID"   : os.getpid(),
         "HOME"  : expanduser("~/"),
@@ -519,13 +609,15 @@ def osexpand(s, actual_username="", uid=0, gid=0, subs={}):
             })
         if not OSX:
             from xpra.platform.xposix.paths import get_runtime_dir
-            d["XDG_RUNTIME_DIR"] = os.environ.get("XDG_RUNTIME_DIR", get_runtime_dir())
+            rd = get_runtime_dir()
+            if rd and "XDG_RUNTIME_DIR" not in os.environ:
+                d["XDG_RUNTIME_DIR"] = rd
     if actual_username:
         d["USERNAME"] = actual_username
         d["USER"] = actual_username
     #first, expand the substitutions themselves,
     #as they may contain references to other variables:
-    ssub = OrderedDict()
+    ssub = {}
     for k,v in d.items():
         ssub[k] = expanduser(shellsub(str(v), d))
     return os.path.expandvars(expanduser(shellsub(expanduser(s), ssub)))
@@ -558,7 +650,7 @@ def path_permission_info(filename, ftype=None):
 #used by the sound code to get rid of the stupid gst warning below:
 #"** Message: pygobject_register_sinkfunc is deprecated (GstObject)"
 #ideally we would redirect to a buffer so we could still capture and show these messages in debug out
-class HideStdErr(object):
+class HideStdErr:
 
     def __init__(self, *_args):
         self.savedstderr = None
@@ -578,7 +670,7 @@ class HideStdErr(object):
         if self.savedstderr is not None:
             os.dup2(self.savedstderr, 2)
 
-class HideSysArgv(object):
+class HideSysArgv:
 
     def __init__(self, *_args):
         self.savedsysargv = None
@@ -592,7 +684,7 @@ class HideSysArgv(object):
             sys.argv = self.savedsysargv
 
 
-class OSEnvContext(object):
+class OSEnvContext:
 
     def __init__(self):
         self.env = os.environ.copy()
@@ -605,7 +697,7 @@ class OSEnvContext(object):
         return "OSEnvContext"
 
 
-class FDChangeCaptureContext(object):
+class FDChangeCaptureContext:
 
     def __init__(self):
         self.enter_fds = []
@@ -621,7 +713,7 @@ class FDChangeCaptureContext(object):
     def get_lost_fds(self):
         return sorted(tuple(set(self.enter_fds)-set(self.exit_fds)))
 
-class DummyContextManager(object):
+class DummyContextManager:
 
     def __enter__(self):
         pass
@@ -632,7 +724,7 @@ class DummyContextManager(object):
 
 
 #workaround incompatibility between paramiko and gssapi:
-class nomodule_context(object):
+class nomodule_context:
 
     def __init__(self, module_name):
         self.module_name = module_name
@@ -642,16 +734,13 @@ class nomodule_context(object):
     def __exit__(self, *_args):
         if sys.modules.get(self.module_name) is None:
             if self.saved_module is None:
-                try:
-                    del sys.modules[self.module_name]
-                except KeyError:
-                    pass
+                sys.modules.pop(self.module_name, None)
             else:
                 sys.modules[self.module_name] = self.saved_module
     def __repr__(self):
         return "nomodule_context(%s)" % self.module_name
 
-class umask_context(object):
+class umask_context:
 
     def __init__(self, umask):
         self.umask = umask
@@ -676,8 +765,8 @@ def setbinarymode(fd):
         #turn on binary mode:
         try:
             import msvcrt
-            msvcrt.setmode(fd, os.O_BINARY)         #@UndefinedVariable
-        except:
+            msvcrt.setmode(fd, os.O_BINARY)         #@UndefinedVariable pylint: disable=no-member
+        except OSError:
             get_util_logger().error("setting stdin to binary mode failed", exc_info=True)
 
 def find_lib_ldconfig(libname):
@@ -697,10 +786,10 @@ def find_lib_ldconfig(libname):
             ldconfig = t
             break
     import subprocess
-    p = subprocess.Popen([ldconfig, "-p"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
+    p = subprocess.Popen("%s -p" % ldconfig, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
     data = bytestostr(p.communicate()[0])
 
-    libpath = re.search(pattern, data, re.MULTILINE)
+    libpath = re.search(pattern, data, re.MULTILINE)        #@UndefinedVariable
     if libpath:
         libpath = libpath.group(1)
     return libpath
@@ -711,8 +800,10 @@ def find_lib(libname):
     #it found the library in
     assert POSIX
     libpaths = os.environ.get("LD_LIBRARY_PATH", "").split(":")
-    libpaths.append("/usr/lib64")
-    libpaths.append("/usr/lib")
+    if BITS==64 and os.path.exists("/usr/lib64"):
+        libpaths.append("/usr/lib64")
+    else:
+        libpaths.append("/usr/lib")
     for libpath in libpaths:
         if not libpath or not os.path.exists(libpath):
             continue
@@ -724,20 +815,33 @@ def find_lib(libname):
 
 def pollwait(process, timeout=5):
     start = monotonic_time()
+    v = None
     while monotonic_time()-start<timeout:
         v = process.poll()
         if v is not None:
-            return v
+            break
         time.sleep(0.1)
-    return None
+    return v
 
 def which(command):
-    from distutils.spawn import find_executable
     try:
-        return find_executable(command)
-    except Exception:
-        get_util_logger().debug("find_executable(%s)", command, exc_info=True)
+        from distutils.spawn import find_executable
+    except ImportError:
+        path = os.environ.get("PATH", None)
+        if not path:
+            return None
+        paths = path.split(os.pathsep)
+        for p in paths:
+            f = os.path.join(p, command)
+            if os.path.isfile(f):
+                return f
         return None
+    else:
+        try:
+            return find_executable(command)
+        except Exception:
+            get_util_logger().debug("find_executable(%s)", command, exc_info=True)
+            return None
 
 def get_status_output(*args, **kwargs):
     import subprocess
@@ -752,15 +856,16 @@ def get_status_output(*args, **kwargs):
     return p.returncode, stdout.decode("utf-8"), stderr.decode("utf-8")
 
 
-def is_systemd_pid1():
+def is_systemd_pid1() -> bool:
     if not POSIX:
         return False
     d = load_binary_file("/proc/1/cmdline")
     return d and d.find(b"/systemd")>=0
 
 
-def get_ssh_port():
-    #FIXME: how do we find out which port ssh is on?
+def get_ssh_port() -> int:
+    #on Linux we can run "ssh -T | grep port"
+    #but this usually requires root permissions to access /etc/ssh/sshd_config
     if WIN32:
         return 0
     return 22
@@ -776,7 +881,7 @@ def setuidgid(uid, gid):
         try:
             username = getpwuid(uid).pw_name
         except KeyError:
-            raise Exception("uid %i not found" % uid)
+            raise Exception("uid %i not found" % uid) from None
         #set the groups:
         if hasattr(os, "initgroups"):   # python >= 2.7
             os.initgroups(username, gid)
@@ -822,29 +927,7 @@ def get_peercred(sock):
             log.error("Error getting peer credentials: %s", e)
             return None
     elif FREEBSD:
-        #TODO: use getpeereid
+        log.warn("Warning: peercred is not yet implemented for FreeBSD")
+        #use getpeereid
         #then pwd to get the gid?
-        pass
     return None
-
-
-def main():
-    sp = sys.platform
-    log = get_util_logger()
-    log.info("platform_name(%s)=%s", sp, platform_name(sp, ""))
-    if LINUX:
-        log.info("linux_distribution=%s", get_linux_distribution())
-        log.info("Ubuntu=%s", is_Ubuntu())
-        if is_Ubuntu():
-            log.info("Ubuntu version=%s", getUbuntuVersion())
-        log.info("Unity=%s", is_unity())
-        log.info("Fedora=%s", is_Fedora())
-        log.info("systemd=%s", is_systemd_pid1())
-    log.info("get_machine_id()=%s", get_machine_id())
-    log.info("get_user_uuid()=%s", get_user_uuid())
-    log.info("get_hex_uuid()=%s", get_hex_uuid())
-    log.info("get_int_uuid()=%s", get_int_uuid())
-
-
-if __name__ == "__main__":
-    main()
