@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # This file is part of Xpra.
 # Copyright (C) 2014-2019 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
@@ -90,7 +90,8 @@ def find_ppd_file(short_name, filename):
         log("using environment override for %s ppd file: %s", short_name, ev)
         return ev
     paths = []
-    for p in os.environ.get("XDG_DATA_DIRS", "/usr/local/share:/usr/share").split(":"):
+    from xpra.common import DEFAULT_XDG_DATA_DIRS
+    for p in os.environ.get("XDG_DATA_DIRS", DEFAULT_XDG_DATA_DIRS).split(":"):
         if os.path.exists(p) and os.path.isdir(p):
             paths.append(os.path.join(p, "cups", "model"))      #used on Fedora and others
             paths.append(os.path.join(p, "ppd", "cups-pdf"))    #used on Fedora and others
@@ -120,7 +121,6 @@ def get_lpinfo_drv(make_and_model):
         return None
     #use the global child reaper to make sure this doesn't end up as a zombie
     from xpra.child_reaper import getChildReaper
-    from xpra.util import nonl
     cr = getChildReaper()
     cr.add_process(proc, "lpinfo", command, ignore=True, forget=True)
     from xpra.make_thread import start_thread
@@ -148,13 +148,13 @@ def get_lpinfo_drv(make_and_model):
         return None
     try:
         out = out.decode()
-    except:
+    except Exception:
         out = str(out)
-    log("lpinfo out=%s", nonl(out))
-    log("lpinfo err=%s", nonl(err))
+    log("lpinfo out=%r", out)
+    log("lpinfo err=%r", err)
     if err:
         log.warn("Warning: lpinfo command produced some warnings:")
-        log.warn(" '%s'", nonl(err))
+        log.warn(" %r", err)
     for line in out.splitlines():
         if line.startswith("drv://"):
             return line.split(" ")[0]
