@@ -1,6 +1,6 @@
 # This file is part of Xpra.
 # Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
-# Copyright (C) 2011-2018 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2011-2020 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -8,75 +8,122 @@ import os
 import sys
 import struct
 
+from xpra.os_util import bytestostr, hexstr
+from xpra.util import iround, envbool, envint, csv, ellipsizer
+from xpra.os_util import is_unity, is_gnome, is_kde, is_Ubuntu, is_Fedora, is_X11, is_Wayland, saved_env
 from xpra.log import Logger
+
 log = Logger("posix")
 eventlog = Logger("posix", "events")
 screenlog = Logger("posix", "screen")
 dbuslog = Logger("posix", "dbus")
 traylog = Logger("posix", "tray")
-menulog = Logger("posix", "menu")
 mouselog = Logger("posix", "mouse")
 xinputlog = Logger("posix", "xinput")
 
-from xpra.os_util import bytestostr, hexstr
-from xpra.util import iround, envbool, envint, csv
-from xpra.gtk_common.gtk_util import get_xwindow
-from xpra.gtk_common.gobject_compat import is_gtk3
+_X11Window = False
+def X11WindowBindings():
+    global _X11Window
+    if _X11Window is False:
+        _X11Window = None
+        if is_X11():
+            try:
+                from xpra.x11.bindings.window_bindings import X11WindowBindings as _X11WindowBindings #@UnresolvedImport
+                _X11Window = _X11WindowBindings()
+            except Exception as e:
+                log("X11WindowBindings()", exc_info=True)
+                log.error("Error: no X11 bindings")
+                log.error(" %s", e)
+    return _X11Window
+
+def X11RandRBindings():
+    if is_X11():
+        try:
+            from xpra.x11.bindings.randr_bindings import RandRBindings  #@UnresolvedImport
+            return RandRBindings()
+        except Exception as e:
+            log("RandRBindings()", exc_info=True)
+            log.error("Error: no X11 RandR bindings")
+            log.error(" %s", e)
+    return None
+
+X11XI2 = False
+def X11XI2Bindings():
+    global X11XI2
+    if X11XI2 is False:
+        X11XI2 = None
+        if is_X11():
+            try:
+                from xpra.x11.bindings.xi2_bindings import X11XI2Bindings as _X11XI2Bindings       #@UnresolvedImport
+                X11XI2 = _X11XI2Bindings()
+            except Exception:
+                log.error("no XI2 bindings", exc_info=True)
+    return X11XI2
 
-try:
-    from xpra.x11.bindings.window_bindings import X11WindowBindings
-    from xpra.x11.bindings.xi2_bindings import X11XI2Bindings   #@UnresolvedImport
-except Exception as e:
-    log.error("no X11 bindings", exc_info=True)
-    del e
-    X11WindowBindings = None
-    X11XI2Bindings = None
 
 device_bell = None
 GTK_MENUS = envbool("XPRA_GTK_MENUS", False)
 RANDR_DPI = envbool("XPRA_RANDR_DPI", True)
 XSETTINGS_DPI = envbool("XPRA_XSETTINGS_DPI", True)
-USE_NATIVE_TRAY = envbool("XPRA_USE_NATIVE_TRAY", True)
+USE_NATIVE_TRAY = envbool("XPRA_USE_NATIVE_TRAY", is_unity() or (is_Ubuntu() and is_gnome()) or (is_gnome() and not is_Fedora()) or is_kde())
 XINPUT_WHEEL_DIV = envint("XPRA_XINPUT_WHEEL_DIV", 15)
 DBUS_SCREENSAVER = envbool("XPRA_DBUS_SCREENSAVER", False)
 
 
+def gl_check():
+    if not is_X11() and is_Wayland():
+        return "disabled under wayland with GTK3 (buggy)"
+    return None
+
+
 def get_native_system_tray_classes():
-    c = []
-    if USE_NATIVE_TRAY:
-        try:
-            from xpra.platform.xposix.appindicator_tray import AppindicatorTray, can_use_appindicator
-            if can_use_appindicator():
-                c.append(AppindicatorTray)
-        except Exception as e:
-            traylog("cannot load appindicator tray: %s", e)
-    return c
+    return []
 
 def get_wm_name():
-    wm_name = os.environ.get("XDG_CURRENT_DESKTOP", "")
-    try:
-        wm_check = _get_X11_root_property("_NET_SUPPORTING_WM_CHECK", "WINDOW")
-        if wm_check:
-            xid = struct.unpack("=I", wm_check)[0]
-            traylog("_NET_SUPPORTING_WM_CHECK window=%#x", xid)
-            wm_name = _get_X11_window_property(xid, "_NET_WM_NAME", "UTF8_STRING")
-            traylog("_NET_WM_NAME=%s", wm_name)
-            return wm_name.decode("utf-8")
-    except Exception as e:
-        traylog.error("Error accessing window manager information:")
-        traylog.error(" %s", e)
+    wm_name = os.environ.get("XDG_CURRENT_DESKTOP", "") or os.environ.get("XDG_SESSION_DESKTOP") or os.environ.get("DESKTOP_SESSION")
+    if os.environ.get("XDG_SESSION_TYPE")=="wayland" or os.environ.get("GDK_BACKEND")=="wayland":
+        if wm_name:
+            wm_name += " on wayland"
+        else:
+            wm_name = "wayland"
+    elif is_X11():
+        try:
+            wm_check = _get_X11_root_property("_NET_SUPPORTING_WM_CHECK", "WINDOW")
+            if wm_check:
+                xid = struct.unpack(b"@L", wm_check)[0]
+                traylog("_NET_SUPPORTING_WM_CHECK window=%#x", xid)
+                wm_name = _get_X11_window_property(xid, "_NET_WM_NAME", "UTF8_STRING")
+                traylog("_NET_WM_NAME=%s", wm_name)
+                if wm_name:
+                    return wm_name.decode("utf-8")
+        except Exception as e:
+            traylog("get_wm_name()", exc_info=True)
+            traylog.error("Error accessing window manager information:")
+            traylog.error(" %s", e)
     return wm_name
 
+def get_clipboard_native_class():
+    if is_Wayland():
+        return "xpra.gtk_common.gtk_clipboard.GTK_Clipboard"
+    return "xpra.x11.gtk_x11.clipboard.X11Clipboard"
+
 def get_native_tray_classes():
     #could restrict to only DEs that have a broken system tray like "GNOME Shell"?
-    if has_gtk_menu_support():  #and wm_name=="GNOME Shell":
+    c = []
+    if USE_NATIVE_TRAY:
         try:
-            from xpra.platform.xposix.gtkmenu_tray import GTKMenuTray
-            traylog("using GTKMenuTray for '%s' window manager", get_wm_name() or "unknown")
-            return [GTKMenuTray]
-        except Exception as e:
-            traylog("cannot load gtk menu tray: %s", e)
-    return get_native_system_tray_classes()
+            from xpra.platform.xposix.appindicator_tray import AppindicatorTray
+            c.append(AppindicatorTray)
+        except (ImportError, ValueError):
+            traylog("cannot load appindicator tray", exc_info=True)
+            traylog.warn("Warning: appindicator library not found")
+            traylog.warn(" you may want to install libappindicator")
+            traylog.warn(" to enable the system tray.")
+            if saved_env.get("XDG_CURRENT_DESKTOP")=="GNOME":
+                traylog.warn(" With gnome-shell, you may also need some extensions:")
+                traylog.warn(" 'top icons plus' and / or 'appindicator'")
+    traylog("get_native_tray_classes()=%s (USE_NATIVE_TRAY=%s)", c, USE_NATIVE_TRAY)
+    return c
 
 
 def get_native_notifier_classes():
@@ -105,9 +152,8 @@ def _get_X11_window_property(xid, name, req_type):
         from xpra.gtk_common.error import xsync
         from xpra.x11.bindings.window_bindings import PropertyError #@UnresolvedImport
         try:
-            X11Window = X11WindowBindings()
             with xsync:
-                prop = X11Window.XGetWindowProperty(xid, name, req_type)
+                prop = X11WindowBindings().XGetWindowProperty(xid, name, req_type)
             log("_get_X11_window_property(%#x, %s, %s)=%s, len=%s", xid, name, req_type, type(prop), len(prop or []))
             return prop
         except PropertyError as e:
@@ -118,86 +164,37 @@ def _get_X11_window_property(xid, name, req_type):
     return None
 def _get_X11_root_property(name, req_type):
     try:
-        X11Window = X11WindowBindings()
-        root_xid = X11Window.getDefaultRootWindow()
+        root_xid = X11WindowBindings().getDefaultRootWindow()
         return _get_X11_window_property(root_xid, name, req_type)
     except Exception as e:
+        log("_get_X11_root_property(%s, %s)", name, req_type, exc_info=True)
         log.warn("Warning: failed to get X11 root property '%s'", name)
         log.warn(" %s", e)
     return None
 
 
-def _set_gtk_x11_window_menu(add, wid, window, menus, application_action_callback=None, window_action_callback=None):
-    from xpra.x11.dbus.menu import setup_dbus_window_menu
-    from xpra.x11.gtk_x11.prop import prop_set, prop_del
-    window_props = setup_dbus_window_menu(add, wid, menus, application_action_callback, window_action_callback)
-    #window_props may contains X11 window properties we have to clear or set
-    if not window_props:
-        return
-    if not window:
-        #window has already been closed
-        #(but we still want to call setup_dbus_window_menu above to ensure we clear things up!)
-        return
-    menulog("will set/remove the following window properties for wid=%i: %s", wid, window_props)
-    try:
-        from xpra.gtk_common.error import xsync
-        with xsync:
-            for k,v in window_props.items():
-                if v is None:
-                    prop_del(window, k)
-                else:
-                    vtype, value = v
-                    prop_set(window, k, vtype, value)
-    except Exception as e:
-        menulog.error("Error setting menu window properties:")
-        menulog.error(" %s", e)
-
-
-_has_gtk_menu_support = None
-def has_gtk_menu_support():
-    global _has_gtk_menu_support
-    if not GTK_MENUS:
-        _has_gtk_menu_support = False
-    if _has_gtk_menu_support is not None:
-        return _has_gtk_menu_support
-    try:
-        from xpra.gtk_common.gtk_util import get_default_root_window
-        from xpra.x11.dbus.menu import has_gtk_menu_support
-        root = get_default_root_window()
-        _has_gtk_menu_support = has_gtk_menu_support(root)
-        menulog("has_gtk_menu_support(%s)=%s", root, _has_gtk_menu_support)
-    except Exception as e:
-        menulog("cannot enable gtk-x11 menu support: %s", e)
-        _has_gtk_menu_support = False
-    return _has_gtk_menu_support
-
-def get_menu_support_function():
-    if has_gtk_menu_support():
-        return _set_gtk_x11_window_menu
-    return None
-
-
 def _get_xsettings():
-    from xpra.gtk_common.error import xsync
-    try:
-        X11Window = X11WindowBindings()
-        with xsync:
-            selection = "_XSETTINGS_S0"
-            owner = X11Window.XGetSelectionOwner(selection)
-            if not owner:
-                return None
-            XSETTINGS = "_XSETTINGS_SETTINGS"
-            data = X11Window.XGetWindowProperty(owner, XSETTINGS, XSETTINGS)
-            if not data:
-                return None
-            from xpra.x11.xsettings_prop import get_settings
-            return get_settings(X11Window.get_display_name(), data)
-    except Exception as e:
-        log("_get_xsettings()", exc_info=True)
+    from xpra.gtk_common.error import xlog
+    X11Window = X11WindowBindings()
+    if not X11Window:
+        return None
+    with xlog:
+        selection = "_XSETTINGS_S0"
+        owner = X11Window.XGetSelectionOwner(selection)
+        if not owner:
+            return None
+        XSETTINGS = "_XSETTINGS_SETTINGS"
+        data = X11Window.XGetWindowProperty(owner, XSETTINGS, XSETTINGS)
+        if not data:
+            return None
+        from xpra.x11.xsettings_prop import get_settings
+        return get_settings(X11Window.get_display_name(), data)
     return None
 
 def _get_xsettings_dict():
     d = {}
+    if is_Wayland():
+        return d
     v = _get_xsettings()
     if v:
         _, values = v
@@ -207,7 +204,7 @@ def _get_xsettings_dict():
 
 
 def _get_xsettings_dpi():
-    if XSETTINGS_DPI:
+    if XSETTINGS_DPI and is_X11():
         from xpra.x11.xsettings_prop import XSettingsTypeInteger
         d = _get_xsettings_dict()
         for k,div in {
@@ -225,20 +222,18 @@ def _get_xsettings_dpi():
     return -1
 
 def _get_randr_dpi():
-    if RANDR_DPI:
-        try:
-            from xpra.gtk_common.error import xsync
-            from xpra.x11.bindings.randr_bindings import RandRBindings  #@UnresolvedImport
-            with xsync:
-                randr_bindings = RandRBindings()
+    if RANDR_DPI and not is_Wayland():
+        from xpra.gtk_common.error import xlog
+        with xlog:
+            randr_bindings = X11RandRBindings()
+            if randr_bindings and randr_bindings.has_randr():
                 wmm, hmm = randr_bindings.get_screen_size_mm()
-                w, h =  randr_bindings.get_screen_size()
-                dpix = iround(w * 25.4 / wmm)
-                dpiy = iround(h * 25.4 / hmm)
-                screenlog("xdpi=%s, ydpi=%s - size-mm=%ix%i, size=%ix%i", dpix, dpiy, wmm, hmm, w, h)
-                return dpix, dpiy
-        except Exception as e:
-            screenlog.warn("failed to get dpi: %s", e)
+                if wmm>0 and hmm>0:
+                    w, h =  randr_bindings.get_screen_size()
+                    dpix = iround(w * 25.4 / wmm)
+                    dpiy = iround(h * 25.4 / hmm)
+                    screenlog("xdpi=%s, ydpi=%s - size-mm=%ix%i, size=%ix%i", dpix, dpiy, wmm, hmm, w, h)
+                    return dpix, dpiy
     return -1, -1
 
 def get_xdpi():
@@ -255,28 +250,32 @@ def get_ydpi():
 
 
 def get_icc_info():
-    try:
-        data = _get_X11_root_property("_ICC_PROFILE", "CARDINAL")
-        if data:
-            screenlog("_ICC_PROFILE=%s (%s)", type(data), len(data))
-            version = _get_X11_root_property("_ICC_PROFILE_IN_X_VERSION", "CARDINAL")
-            screenlog("get_icc_info() found _ICC_PROFILE_IN_X_VERSION=%s, _ICC_PROFILE=%s", hexstr(version or ""), hexstr(data))
-            icc = {
-                    "source"    : "_ICC_PROFILE",
-                    "data"      : data,
-                    }
-            if version:
-                try:
-                    version = ord(version)
-                except:
-                    pass
-                icc["version"] = version
-            return icc
-    except Exception as e:
-        screenlog.error("Error: cannot access _ICC_PROFILE X11 window property")
-        screenlog.error(" %s", e)
-        screenlog("get_icc_info()", exc_info=True)
-    return {}
+    if not is_Wayland():
+        try:
+            data = _get_X11_root_property("_ICC_PROFILE", "CARDINAL")
+            if data:
+                screenlog("_ICC_PROFILE=%s (%s)", type(data), len(data))
+                version = _get_X11_root_property("_ICC_PROFILE_IN_X_VERSION", "CARDINAL")
+                screenlog("get_icc_info() found _ICC_PROFILE_IN_X_VERSION=%s, _ICC_PROFILE=%s",
+                          hexstr(version or ""), hexstr(data))
+                icc = {
+                        "source"    : "_ICC_PROFILE",
+                        "data"      : data,
+                        }
+                if version:
+                    try:
+                        version = ord(version)
+                    except TypeError:
+                        pass
+                    icc["version"] = version
+                screenlog("get_icc_info()=%s", icc)
+                return icc
+        except Exception as e:
+            screenlog.error("Error: cannot access _ICC_PROFILE X11 window property")
+            screenlog.error(" %s", e)
+            screenlog("get_icc_info()", exc_info=True)
+    from xpra.platform.gui import default_get_icc_info
+    return default_get_icc_info()
 
 
 def get_antialias_info():
@@ -296,11 +295,11 @@ def get_antialias_info():
             return {"hintnone"      : 0,
                     "hintslight"    : 1000,
                     "hintmedium"    : 1600,
-                    "hintfull"      : 2200}.get(value)
+                    "hintfull"      : 2200}.get(bytestostr(value))
         for prop_name, name, convert in (
-                                         ("Xft/HintStyle",  "hintstyle",    str),
+                                         ("Xft/HintStyle",  "hintstyle",    bytestostr),
                                          ("Xft/HintStyle",  "contrast",     get_contrast),
-                                         ("Xft/RGBA",       "orientation",  lambda x : str(x).upper())
+                                         ("Xft/RGBA",       "orientation",  lambda x : bytestostr(x).upper())
                                          ):
             if prop_name in d:
                 value_type, value = d.get(prop_name)
@@ -316,111 +315,122 @@ def get_antialias_info():
 
 def get_current_desktop():
     v = -1
-    d = None
-    try:
-        d = _get_X11_root_property("_NET_CURRENT_DESKTOP", "CARDINAL")
-        if d:
-            v = struct.unpack("=I", d)[0]
-    except Exception as e:
-        log.warn("failed to get current desktop: %s", e)
-    log("get_current_desktop() %s=%s", hexstr(d or ""), v)
+    if not is_Wayland():
+        d = None
+        try:
+            d = _get_X11_root_property("_NET_CURRENT_DESKTOP", "CARDINAL")
+            if d:
+                v = struct.unpack(b"@L", d)[0]
+        except Exception as e:
+            log.warn("failed to get current desktop: %s", e)
+        log("get_current_desktop() %s=%s", hexstr(d or ""), v)
     return v
 
 def get_workarea():
-    try:
-        d = get_current_desktop()
-        if d<0:
-            return None
-        workarea = _get_X11_root_property("_NET_WORKAREA", "CARDINAL")
-        if not workarea:
-            return None
-        screenlog("get_workarea()=%s, len=%s", type(workarea), len(workarea))
-        #workarea comes as a list of 4 CARDINAL dimensions (x,y,w,h), one for each desktop
-        if len(workarea)<(d+1)*4*4:
-            screenlog.warn("get_workarea() invalid _NET_WORKAREA value")
-        else:
-            cur_workarea = workarea[d*4*4:(d+1)*4*4]
-            v = struct.unpack("=IIII", cur_workarea)
-            screenlog("get_workarea() %s=%s", hexstr(cur_workarea), v)
-            return v
-    except Exception as e:
-        screenlog.warn("failed to get workarea: %s", e)
+    if not is_Wayland():
+        try:
+            d = get_current_desktop()
+            if d<0:
+                return None
+            workarea = _get_X11_root_property("_NET_WORKAREA", "CARDINAL")
+            if not workarea:
+                return None
+            screenlog("get_workarea() _NET_WORKAREA=%s (%s), len=%s",
+                      ellipsizer(workarea), type(workarea), len(workarea))
+            #workarea comes as a list of 4 CARDINAL dimensions (x,y,w,h), one for each desktop
+            sizeof_long = struct.calcsize(b"@L")
+            if len(workarea)<(d+1)*4*sizeof_long:
+                screenlog.warn("get_workarea() invalid _NET_WORKAREA value")
+            else:
+                cur_workarea = workarea[d*4*sizeof_long:(d+1)*4*sizeof_long]
+                v = struct.unpack(b"@LLLL", cur_workarea)
+                screenlog("get_workarea() %s=%s", hexstr(cur_workarea), v)
+                return v
+        except Exception as e:
+            screenlog("get_workarea()", exc_info=True)
+            screenlog.warn("Warning: failed to query workarea: %s", e)
     return None
 
 
 def get_number_of_desktops():
-    v = 1
-    d = None
-    try:
-        d = _get_X11_root_property("_NET_NUMBER_OF_DESKTOPS", "CARDINAL")
-        if d:
-            v = struct.unpack("=I", d)[0]
-    except Exception as e:
-        screenlog.warn("failed to get number of desktop: %s", e)
-    v = max(1, v)
-    screenlog("get_number_of_desktops() %s=%s", hexstr(d or ""), v)
+    v = 0
+    if not is_Wayland():
+        d = None
+        try:
+            d = _get_X11_root_property("_NET_NUMBER_OF_DESKTOPS", "CARDINAL")
+            if d:
+                v = struct.unpack(b"@L", d)[0]
+        except Exception as e:
+            screenlog.warn("failed to get number of desktop: %s", e)
+        v = max(1, v)
+        screenlog("get_number_of_desktops() %s=%s", hexstr(d or ""), v)
     return v
 
 def get_desktop_names():
-    v = ["Main"]
-    d = None
-    try:
-        d = _get_X11_root_property("_NET_DESKTOP_NAMES", "UTF8_STRING")
-        if d:
-            v = d.split(b"\0")
-            if len(v)>1 and v[-1]=="":
-                v = v[:-1]
-    except Exception as e:
-        screenlog.warn("failed to get desktop names: %s", e)
-    screenlog("get_desktop_names() %s=%s", hexstr(d or ""), v)
+    v = []
+    if not is_Wayland():
+        v = ["Main"]
+        d = None
+        try:
+            d = _get_X11_root_property("_NET_DESKTOP_NAMES", "UTF8_STRING")
+            if d:
+                v = d.split(b"\0")
+                if len(v)>1 and v[-1]==b"":
+                    v = v[:-1]
+                return [x.decode("utf8") for x in v]
+        except Exception as e:
+            screenlog.warn("failed to get desktop names: %s", e)
+        screenlog("get_desktop_names() %s=%s", hexstr(d or ""), v)
     return v
 
 
 def get_vrefresh():
-    try:
-        from xpra.x11.bindings.randr_bindings import RandRBindings      #@UnresolvedImport
-        randr = RandRBindings()
-        v = randr.get_vrefresh()
-    except Exception as e:
-        log("get_vrefresh()", exc_info=True)
-        log.warn("Warning: failed to query the display vertical refresh rate:")
-        log.warn(" %s", e)
-        v = -1
-    screenlog("get_vrefresh()=%s", v)
+    v = -1
+    if not is_Wayland():
+        try:
+            from xpra.x11.bindings.randr_bindings import RandRBindings      #@UnresolvedImport
+            randr = RandRBindings()
+            if randr.has_randr():
+                v = randr.get_vrefresh()
+        except Exception as e:
+            log("get_vrefresh()", exc_info=True)
+            log.warn("Warning: failed to query the display vertical refresh rate:")
+            log.warn(" %s", e)
+        screenlog("get_vrefresh()=%s", v)
     return v
 
 
 def _get_xresources():
-    try:
-        from xpra.x11.gtk_x11.prop import prop_get
-        from xpra.gtk_common.gtk_util import get_default_root_window
-        root = get_default_root_window()
-        v = prop_get(root, "RESOURCE_MANAGER", "latin1", ignore_errors=True)
-        log("RESOURCE_MANAGER=%s", v)
-        if v is None:
-            return None
-        value = v.decode("utf-8")
-        #parse the resources into a dict:
-        values={}
-        options = value.split("\n")
-        for option in options:
-            if not option:
-                continue
-            parts = option.split(":\t", 1)
-            if len(parts)!=2:
-                log("skipped invalid option: '%s'", option)
-                continue
-            values[parts[0]] = parts[1]
-        return values
-    except Exception as e:
-        log("_get_xresources error: %s", e)
+    if not is_Wayland():
+        try:
+            from xpra.x11.gtk_x11.prop import prop_get
+            from xpra.gtk_common.gtk_util import get_default_root_window
+            root = get_default_root_window()
+            value = prop_get(root, "RESOURCE_MANAGER", "latin1", ignore_errors=True)
+            log("RESOURCE_MANAGER=%s", value)
+            if value is None:
+                return None
+            #parse the resources into a dict:
+            values={}
+            options = value.split("\n")
+            for option in options:
+                if not option:
+                    continue
+                parts = option.split(":\t", 1)
+                if len(parts)!=2:
+                    log("skipped invalid option: '%s'", option)
+                    continue
+                values[parts[0]] = parts[1]
+            return values
+        except Exception as e:
+            log("_get_xresources error: %s", e)
     return None
 
 def get_cursor_size():
     d = _get_xresources() or {}
     try:
         return int(d.get("Xcursor.size", 0))
-    except:
+    except ValueError:
         return -1
 
 
@@ -449,44 +459,50 @@ def get_window_frame_sizes():
 
 
 def system_bell(window, device, percent, _pitch, _duration, bell_class, bell_id, bell_name):
-    from xpra.gtk_common.error import XError
+    if not is_X11():
+        return False
     global device_bell
     if device_bell is False:
         #failed already
         return False
+    from xpra.gtk_common.error import XError
     def x11_bell():
         global device_bell
         if device_bell is None:
             #try to load it:
             from xpra.x11.bindings.keyboard_bindings import X11KeyboardBindings       #@UnresolvedImport
             device_bell = X11KeyboardBindings().device_bell
-        device_bell(get_xwindow(window), device, bell_class, bell_id, percent, bell_name)
+        device_bell(window.get_xid(), device, bell_class, bell_id, percent, bell_name)
     try:
-        from xpra.gtk_common.error import xsync
-        with xsync:
+        from xpra.gtk_common.error import xlog
+        with xlog:
             x11_bell()
         return  True
     except XError as e:
-        log.error("error using device_bell: %s, switching native X11 bell support off", e)
+        log("x11_bell()", exc_info=True)
+        log.error("Error using device_bell: %s", e)
+        log.error(" switching native X11 bell support off")
         device_bell = False
         return False
 
 
 def _send_client_message(window, message_type, *values):
     try:
-        from xpra.x11.gtk_x11.gdk_display_source import init_display_source
-        init_display_source()
+        from xpra.x11.gtk_x11.gdk_display_source import init_gdk_display_source
+        init_gdk_display_source()
         from xpra.x11.bindings.window_bindings import constants #@UnresolvedImport
         X11Window = X11WindowBindings()
         root_xid = X11Window.getDefaultRootWindow()
         if window:
-            xid = get_xwindow(window)
+            xid = window.get_xid()
         else:
             xid = root_xid
         SubstructureNotifyMask = constants["SubstructureNotifyMask"]
         SubstructureRedirectMask = constants["SubstructureRedirectMask"]
         event_mask = SubstructureNotifyMask | SubstructureRedirectMask
-        X11Window.sendClientMessage(root_xid, xid, False, event_mask, message_type, *values)
+        from xpra.gtk_common.error import xsync
+        with xsync:
+            X11Window.sendClientMessage(root_xid, xid, False, event_mask, message_type, *values)
     except Exception as e:
         log.warn("failed to send client message '%s' with values=%s: %s", message_type, values, e)
 
@@ -494,7 +510,7 @@ def show_desktop(b):
     _send_client_message(None, "_NET_SHOWING_DESKTOP", int(bool(b)))
 
 def set_fullscreen_monitors(window, fsm, source_indication=0):
-    if type(fsm) not in (tuple, list):
+    if not isinstance(fsm, (tuple, list)):
         log.warn("invalid type for fullscreen-monitors: %s", type(fsm))
         return
     if len(fsm)!=4:
@@ -547,23 +563,23 @@ def get_info():
     return i
 
 
-class XI2_Window(object):
+class XI2_Window:
     def __init__(self, window):
         log("XI2_Window(%s)", window)
         self.XI2 = X11XI2Bindings()
         self.X11Window = X11WindowBindings()
         self.window = window
-        self.xid = get_xwindow(window.get_window())
+        self.xid = window.get_window().get_xid()
         self.windows = ()
         self.motion_valuators = {}
         window.connect("configure-event", self.configured)
         self.configured()
         #replace event handlers with XI2 version:
-        self.do_motion_notify_event = window.do_motion_notify_event
-        window.do_motion_notify_event = self.noop
-        window.do_button_press_event = self.noop
-        window.do_button_release_event = self.noop
-        window.do_scroll_event = self.noop
+        self._do_motion_notify_event = window._do_motion_notify_event
+        window._do_motion_notify_event = self.noop
+        window._do_button_press_event = self.noop
+        window._do_button_release_event = self.noop
+        window._do_scroll_event = self.noop
         window.connect("destroy", self.cleanup)
 
     def noop(self, *args):
@@ -576,10 +592,10 @@ class XI2_Window(object):
         self.window = None
 
     def configured(self, *_args):
-        from xpra.gtk_common.error import xsync
-        with xsync:
+        from xpra.gtk_common.error import xlog
+        with xlog:
             self.windows = self.get_parent_windows(self.xid)
-        for window in self.windows:
+        for window in (self.windows or ()):
             self.XI2.connect(window, "XI_Motion", self.do_xi_motion)
             self.XI2.connect(window, "XI_ButtonPress", self.do_xi_button)
             self.XI2.connect(window, "XI_ButtonRelease", self.do_xi_button)
@@ -630,18 +646,19 @@ class XI2_Window(object):
         if window.moveresize_event:
             xinputlog("do_xi_motion(%s, %s) handling as a moveresize event on window %s", event, device, window)
             window.motion_moveresize(event)
-            self.do_motion_notify_event(event)
+            self._do_motion_notify_event(event)
             return
         client = window._client
         if client.readonly:
             return
-        pointer, modifiers, buttons = window._pointer_modifiers(event)
+        pointer, relative_pointer, modifiers, buttons = window._pointer_modifiers(event)
         wid = self.window.get_mouse_event_wid(*pointer)
-        #log("server_input_devices=%s, server_precise_wheel=%s", client.server_input_devices, client.server_precise_wheel)
+        #log("server_input_devices=%s, server_precise_wheel=%s",
+        #    client.server_input_devices, client.server_precise_wheel)
         valuators = event.valuators
         unused_valuators = valuators.copy()
         dx, dy = 0, 0
-        if (valuators and device and device.get("enabled") and 
+        if (valuators and device and device.get("enabled") and
             client.server_input_devices=="uinput" and client.server_precise_wheel):
             XIModeRelative = 0
             classes = device.get("classes")
@@ -680,14 +697,18 @@ class XI2_Window(object):
             mv.update(event.valuators)
         #send plain motion first, if any:
         if unused_valuators:
-            xinputlog("do_xi_motion(%s, %s) wid=%s / focus=%s / window wid=%i, device=%s, pointer=%s, modifiers=%s, buttons=%s", event, device, wid, window._client._focused, window._id, event.device, pointer, modifiers, buttons)
-            packet = ["pointer-position", wid, pointer, modifiers, buttons] + self.get_pointer_extra_args(event)
+            xinputlog("do_xi_motion(%s, %s) wid=%s / focus=%s / window wid=%i, device=%s, pointer=%s, modifiers=%s, buttons=%s",
+                      event, device, wid, window._client._focused, window._id, event.device, pointer, modifiers, buttons)
+            pdata = pointer
+            if client.server_pointer_relative:
+                pdata = list(pointer)+list(relative_pointer)
+            packet = ["pointer-position", wid, pdata, modifiers, buttons] + self.get_pointer_extra_args(event)
             client.send_mouse_position(packet)
         #now see if we have anything to send as a wheel event:
         if dx!=0 or dy!=0:
             xinputlog("do_xi_motion(%s, %s) wheel deltas: dx=%i, dy=%i", event, device, dx, dy)
             #normalize (xinput is always using 15 degrees?)
-            client.wheel_event(wid, float(dx)/XINPUT_WHEEL_DIV, float(dy)/XINPUT_WHEEL_DIV, event.device)
+            client.wheel_event(wid, dx/XINPUT_WHEEL_DIV, dy/XINPUT_WHEEL_DIV, event.device)
 
     def get_pointer_extra_args(self, event):
         def intscaled(f):
@@ -703,12 +724,12 @@ class XI2_Window(object):
         args = [event.device]
         for x in ("x", "y", "x_root", "y_root"):
             args.append(intscaled(getattr(event, x)))
-        for v in [event.valuators, raw_valuators]:
+        for v in (event.valuators, raw_valuators):
             args.append(dictscaled(v))
         return args
 
 
-class ClientExtras(object):
+class ClientExtras:
     def __init__(self, client, _opts):
         self.client = client
         self._xsettings_watcher = None
@@ -723,7 +744,8 @@ class ClientExtras(object):
         if client.xsettings_enabled:
             self.setup_xprops()
         self.xi_setup_failures = 0
-        if client.input_devices in ("xi", "auto"):
+        input_devices = getattr(client, "input_devices", None)
+        if input_devices in ("xi", "auto"):
             #this would trigger warnings with our temporary opengl windows:
             #only enable it after we have connected:
             self.client.after_handshake(self.setup_xi)
@@ -736,13 +758,11 @@ class ClientExtras(object):
         if self.x11_filter:
             return
         try:
-            if is_gtk3():
-                from xpra.x11.gtk3.gdk_bindings import init_x11_filter  #@UnresolvedImport, @UnusedImport
-            else:
-                from xpra.x11.gtk2.gdk_bindings import init_x11_filter  #@UnresolvedImport, @Reimport
+            from xpra.x11.gtk_x11.gdk_bindings import init_x11_filter  #@UnresolvedImport, @UnusedImport
             self.x11_filter = init_x11_filter()
             log("x11_filter=%s", self.x11_filter)
-        except Exception:
+        except Exception as e:
+            log("init_x11_filter()", exc_info=True)
             log.error("Error: failed to initialize X11 GDK filter:")
             log.error(" %s", e)
             self.x11_filter = None
@@ -750,11 +770,8 @@ class ClientExtras(object):
     def cleanup(self):
         log("cleanup() xsettings_watcher=%s, root_props_watcher=%s", self._xsettings_watcher, self._root_props_watcher)
         if self.x11_filter:
-            if is_gtk3():
-                from xpra.x11.gtk3.gdk_bindings import cleanup_x11_filter   #@UnresolvedImport, @UnusedImport
-            else:
-                from xpra.x11.gtk2.gdk_bindings import cleanup_x11_filter   #@UnresolvedImport, @Reimport
             self.x11_filter = None
+            from xpra.x11.gtk_x11.gdk_bindings import cleanup_x11_filter   #@UnresolvedImport, @UnusedImport
             cleanup_x11_filter()
         if self._xsettings_watcher:
             self._xsettings_watcher.cleanup()
@@ -764,7 +781,8 @@ class ClientExtras(object):
             self._root_props_watcher = None
         if self.system_bus:
             bus = self.system_bus
-            log("cleanup() system bus=%s, matches: %s", bus, (self.upower_resuming_match, self.upower_sleeping_match, self.login1_match))
+            log("cleanup() system bus=%s, matches: %s",
+                bus, (self.upower_resuming_match, self.upower_sleeping_match, self.login1_match))
             self.system_bus = None
             if self.upower_resuming_match:
                 bus._clean_up_signal_match(self.upower_resuming_match)
@@ -873,16 +891,14 @@ class ClientExtras(object):
 
     def setup_xprops(self):
         #wait for handshake to complete:
-        self.client.after_handshake(self.do_setup_xprops)
+        if not is_Wayland():
+            self.client.after_handshake(self.do_setup_xprops)
 
     def do_setup_xprops(self, *args):
         log("do_setup_xprops(%s)", args)
         ROOT_PROPS = ["RESOURCE_MANAGER", "_NET_WORKAREA", "_NET_CURRENT_DESKTOP"]
         try:
             self.init_x11_filter()
-            if is_gtk3():
-                #xsettings still need porting
-                return
             from xpra.gtk_common.gtk_util import get_default_root_window
             from xpra.x11.xsettings import XSettingsWatcher
             from xpra.x11.xroot_props import XRootPropWatcher
@@ -898,7 +914,9 @@ class ClientExtras(object):
                 self._root_props_watcher.do_notify("RESOURCE_MANAGER")
         except ImportError as e:
             log("do_setup_xprops%s", args, exc_info=True)
-            log.error("failed to load X11 properties/settings bindings: %s - root window properties will not be propagated", e)
+            log.error("Error: failed to load X11 properties/settings bindings:")
+            log.error(" %s", e)
+            log.error(" root window properties will not be propagated")
 
 
     def do_xi_devices_changed(self, event):
@@ -913,9 +931,9 @@ class ClientExtras(object):
 
     def do_setup_xi(self):
         if self.client.server_input_devices not in ("xi", "uinput"):
-            xinputlog.info("server does not support xi input devices")
+            xinputlog("server does not support xi input devices")
             if self.client.server_input_devices:
-                log.info(" server uses: %s", self.client.server_input_devices)
+                log(" server uses: %s", self.client.server_input_devices)
             return False
         try:
             from xpra.gtk_common.error import xsync, XError
@@ -929,7 +947,8 @@ class ClientExtras(object):
                     XI2.select_xi2_events()
             except XError:
                 self.xi_setup_failures += 1
-                xinputlog("select_xi2_events() failed, attempt %i", self.xi_setup_failures,exc_info=True)
+                xinputlog("select_xi2_events() failed, attempt %i",
+                          self.xi_setup_failures, exc_info=True)
                 return self.xi_setup_failures<10    #try again
             with xsync:
                 XI2.gdk_inject()
@@ -996,12 +1015,12 @@ class ClientExtras(object):
 
 def main():
     try:
-        from xpra.x11.gtk_x11.gdk_display_source import init_display_source
-        init_display_source()
-    except:
+        from xpra.x11.gtk_x11.gdk_display_source import init_gdk_display_source
+        init_gdk_display_source()
+    except ImportError:
         pass
-    from xpra.platform.gui import main
-    main()
+    from xpra.platform.gui import main as gui_main
+    gui_main()
 
 
 if __name__ == "__main__":
