@@ -1,21 +1,18 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # This file is part of Xpra.
-# Copyright (C) 2010-2018 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import sys
 import os
 
-#this is here so we can expose the python "platform" module
-#before we import xpra.platform
-import platform as python_platform
-assert python_platform
-from xpra.util import csv
+from xpra.util import csv, sorted_nicely
 from xpra.os_util import (
-    WIN32, OSX, PYTHON2, POSIX,
+    WIN32, OSX, POSIX,
     osexpand, getuid, getgid, get_username_for_uid,
-    is_CentOS, is_RedHat, is_Fedora, is_Debian, is_Ubuntu, getUbuntuVersion,
+    is_Debian, is_Ubuntu, is_arm,
+    which,
     )
 
 def warn(msg):
@@ -33,7 +30,7 @@ class InitInfo(Exception):
 class InitExit(Exception):
     def __init__(self, status, msg):
         self.status = status
-        Exception.__init__(self, msg)
+        super().__init__(msg)
 
 
 DEBUG_CONFIG_PROPERTIES = os.environ.get("XPRA_DEBUG_CONFIG_PROPERTIES", "").split()
@@ -43,10 +40,10 @@ DEFAULT_NET_WM_NAME = os.environ.get("XPRA_NET_WM_NAME", "Xpra")
 
 if POSIX:
     DEFAULT_POSTSCRIPT_PRINTER = os.environ.get("XPRA_POSTSCRIPT_PRINTER", "drv:///sample.drv/generic.ppd")
-else:
+else: # pragma: no cover
     DEFAULT_POSTSCRIPT_PRINTER = ""
 DEFAULT_PULSEAUDIO = None   #auto
-if OSX or WIN32:
+if OSX or WIN32: # pragma: no cover
     DEFAULT_PULSEAUDIO = False
 
 
@@ -63,8 +60,11 @@ def has_sound_support():
 
 
 def get_xorg_bin():
+    xorg = os.environ.get("XPRA_XORG_BIN")
+    if xorg:
+        return xorg
     # Detect Xorg Binary
-    if os.uname()[4].startswith("arm") and is_Debian() and os.path.exists("/usr/bin/Xorg"):
+    if is_arm() and is_Debian() and os.path.exists("/usr/bin/Xorg"):
         #Raspbian breaks if we use a different binary..
         return "/usr/bin/Xorg"
     for p in (
@@ -77,7 +77,7 @@ def get_xorg_bin():
         if os.path.exists(p):
             return p
     #look for it in $PATH:
-    for x in os.environ.get("PATH").split(os.pathsep):
+    for x in os.environ.get("PATH").split(os.pathsep): # pragma: no cover
         xorg = os.path.join(x, "Xorg")
         if os.path.isfile(xorg):
             return xorg
@@ -89,11 +89,11 @@ def get_Xdummy_confdir():
     xrd = get_runtime_dir()
     if xrd:
         base = "${XDG_RUNTIME_DIR}/xpra"
-    else:
+    else:   # pragma: no cover
         base = "${HOME}/.xpra"
     return base+"/xorg.conf.d/$PID"
 
-def get_Xdummy_command(xorg_cmd="Xorg", log_dir="${XPRA_LOG_DIR}", xorg_conf="/etc/xpra/xorg.conf"):
+def get_Xdummy_command(xorg_cmd="Xorg", log_dir="${XPRA_LOG_DIR}", xorg_conf="${XORG_CONFIG_PREFIX}/etc/xpra/xorg.conf"):
     cmd = [xorg_cmd]    #ie: ["Xorg"] or ["xpra_Xdummy"] or ["./install/bin/xpra_Xdummy"]
     cmd += [
           "-noreset", "-novtswitch",
@@ -105,32 +105,42 @@ def get_Xdummy_command(xorg_cmd="Xorg", log_dir="${XPRA_LOG_DIR}", xorg_conf="/e
           "-logfile", "%s/Xorg.${DISPLAY}.log" % log_dir,
           #must be specified with some Xorg versions (ie: arch linux)
           #this directory can store xorg config files, it does not need to be created:
-          "-configdir", get_Xdummy_confdir(),
-          "-config", xorg_conf
+          "-configdir", '"%s"' % get_Xdummy_confdir(),
+          "-config", '"%s"' % xorg_conf,
           ]
     return cmd
 
-def get_Xvfb_command():
+def get_Xvfb_command(width=8192, height=4096, dpi=96):
     cmd = ["Xvfb",
            "+extension", "GLX",
            "+extension", "Composite",
-           "-screen", "0", "5760x2560x24+32",
+           "-screen", "0", "%ix%ix24+32" % (width, height),
            #better than leaving to vfb after a resize?
-           "-dpi", "96",
            "-nolisten", "tcp",
            "-noreset",
            "-auth", "$XAUTHORITY"
            ]
+    if dpi>0:
+        cmd += ["-dpi", str(dpi)]
     return cmd
 
-def detect_xvfb_command(conf_dir="/etc/xpra/", bin_dir=None, Xdummy_ENABLED=None, Xdummy_wrapper_ENABLED=None):
+def detect_xvfb_command(conf_dir="/etc/xpra/", bin_dir=None,
+                        Xdummy_ENABLED=None, Xdummy_wrapper_ENABLED=None, warn_fn=warn):
     #returns the xvfb command to use
-    if WIN32:
+    if WIN32:   # pragma: no cover
         return ""
-    if OSX:
+    if OSX:     # pragma: no cover
         return get_Xvfb_command()
-    if sys.platform.find("bsd")>=0 and Xdummy_ENABLED is None:
-        warn("Warning: sorry, no support for Xdummy on %s" % sys.platform)
+    if sys.platform.find("bsd")>=0 and Xdummy_ENABLED is None:  # pragma: no cover
+        if warn_fn:
+            warn_fn("Warning: sorry, no support for Xdummy on %s" % sys.platform)
+        return get_Xvfb_command()
+    if is_arm():
+        #arm struggles to launch Xdummy, so use Xvfb:
+        return get_Xvfb_command()
+    if is_Ubuntu() or is_Debian():
+        #These distros do weird things and this can cause the real X11 server to crash
+        #see ticket #2834
         return get_Xvfb_command()
 
     xorg_bin = get_xorg_bin()
@@ -139,7 +149,8 @@ def detect_xvfb_command(conf_dir="/etc/xpra/", bin_dir=None, Xdummy_ENABLED=None
             #honour what was specified:
             use_wrapper = Xdummy_wrapper_ENABLED
         elif not xorg_bin:
-            warn("Warning: Xorg binary not found, assuming the wrapper is needed!")
+            if warn_fn:
+                warn_fn("Warning: Xorg binary not found, assuming the wrapper is needed!")
             use_wrapper = True
         else:
             #auto-detect
@@ -147,13 +158,14 @@ def detect_xvfb_command(conf_dir="/etc/xpra/", bin_dir=None, Xdummy_ENABLED=None
             xorg_stat = os.stat(xorg_bin)
             if (xorg_stat.st_mode & stat.S_ISUID)!=0:
                 if (xorg_stat.st_mode & stat.S_IROTH)==0:
-                    warn("%s is suid and not readable, Xdummy support unavailable" % xorg_bin)
+                    if warn_fn:
+                        warn_fn("%s is suid and not readable, Xdummy support unavailable" % xorg_bin)
                     return get_Xvfb_command()
                 debug("%s is suid and readable, using the xpra_Xdummy wrapper" % xorg_bin)
                 use_wrapper = True
             else:
                 use_wrapper = False
-        xorg_conf = os.path.join(conf_dir, "xorg.conf")
+        xorg_conf = "${XORG_CONFIG_PREFIX}"+os.path.join(conf_dir, "xorg.conf")
         if use_wrapper:
             xorg_cmd = "xpra_Xdummy"
         else:
@@ -169,53 +181,76 @@ def detect_xvfb_command(conf_dir="/etc/xpra/", bin_dir=None, Xdummy_ENABLED=None
     if Xdummy_ENABLED is True:
         return Xorg_suid_check()
     debug("Xdummy support unspecified, will try to detect")
-
-    if is_Ubuntu():
-        rnum = getUbuntuVersion()
-        if rnum<(16, 10):
-            debug("Warning: Ubuntu Xenial breaks Xorg/Xdummy usage - using Xvfb fallback")
-            return get_Xvfb_command()
     return Xorg_suid_check()
 
 
-def OpenGL_safety_check():
+def xvfb_cmd_str(xvfb, wrap=False):
+    xvfb_str = ""
+    while xvfb:
+        s = ""
+        while xvfb:
+            item = xvfb[0]
+            l = len(item)
+            if (item.startswith("-") or item.startswith("+")) and len(xvfb)>1:
+                l += len(xvfb[1])
+            if s and len(s)+l>55:
+                break
+            v = xvfb.pop(0)
+            if not s:
+                s += v
+            else:
+                s += " "+v
+        if xvfb_str:
+            if wrap:
+                xvfb_str += " \\\n    "
+            else:
+                xvfb_str += " "
+        xvfb_str += s
+    return xvfb_str
+
+
+def OpenGL_safety_check() -> str:
+    return is_VirtualBox()
+
+def is_VirtualBox() -> str:
     #try to detect VirtualBox:
     #based on the code found here:
     #http://spth.virii.lu/eof2/articles/WarGame/vboxdetect.html
     #because it used to cause hard VM crashes when we probe the GL driver!
-    try:
-        from ctypes import cdll
-        if cdll.LoadLibrary("VBoxHook.dll"):
-            return "VirtualBox is present (VBoxHook.dll)"
-    except (ImportError, OSError):
-        pass
-    try:
+    if WIN32:   # pragma: no cover
         try:
-            f = None
-            f = open("\\\\.\\VBoxMiniRdrDN", "r")
-            return True, "VirtualBox is present (VBoxMiniRdrDN)"
-        finally:
-            if f:
-                f.close()
-    except Exception as e:
-        import errno
-        if e.args[0]==errno.EACCES:
-            return "VirtualBox is present (VBoxMiniRdrDN)"
+            from ctypes import cdll
+            if cdll.LoadLibrary("VBoxHook.dll"):
+                return "VirtualBox is present (VBoxHook.dll)"
+        except (ImportError, OSError):
+            pass
+        try:
+            try:
+                f = None
+                f = open("\\\\.\\VBoxMiniRdrDN", "r")
+                return "VirtualBox is present (VBoxMiniRdrDN)"
+            finally:
+                if f:
+                    f.close()
+        except Exception as e:
+            import errno
+            if e.args[0]==errno.EACCES:
+                return "VirtualBox is present (VBoxMiniRdrDN)"
     return None
 
 
 def get_build_info():
     info = []
     try:
-        from xpra.src_info import REVISION, LOCAL_MODIFICATIONS             #@UnresolvedImport
+        from xpra.src_info import REVISION, LOCAL_MODIFICATIONS, BRANCH, COMMIT #@UnresolvedImport
+        info.append("revision %s" % REVISION)
+        if COMMIT:
+            info.append("commit %s from %s branch" % (COMMIT, BRANCH))
         try:
             mods = int(LOCAL_MODIFICATIONS)
+            info.append("with %s local changes" % (mods))
         except ValueError:
-            mods = 0
-        if mods==0:
-            info.append("revision %s" % REVISION)
-        else:
-            info.append("revision %s with %s local changes" % (REVISION, LOCAL_MODIFICATIONS))
+            pass
     except Exception as e:
         warn("Error: could not find the source information: %s" % e)
     try:
@@ -274,11 +309,7 @@ def read_config(conf_file):
     if not os.path.isfile(conf_file):
         debug("read_config(%s) is not a file!", conf_file)
         return d
-    if PYTHON2:
-        mode = "rU"
-    else:
-        mode = "r"
-    with open(conf_file, mode) as f:
+    with open(conf_file, "r") as f:
         lines = []
         no = 0
         for line in f:
@@ -514,6 +545,7 @@ OPTION_TYPES = {
                     "speaker"           : str,
                     "sound-source"      : str,
                     "html"              : str,
+                    "http-scripts"      : str,
                     "socket-permissions": str,
                     "exec-wrapper"      : str,
                     "dbus-launch"       : str,
@@ -525,6 +557,7 @@ OPTION_TYPES = {
                     "open-url"          : str,
                     "file-transfer"     : str,
                     "printing"          : str,
+                    "headerbar"         : str,
                     "challenge-handlers": list,
                     #ssl options:
                     "ssl"               : str,
@@ -549,7 +582,7 @@ OPTION_TYPES = {
                     "min-speed"         : int,
                     "compression_level" : int,
                     "dpi"               : int,
-                    "file-size-limit"   : int,
+                    "file-size-limit"   : str,
                     "idle-timeout"      : int,
                     "server-idle-timeout" : int,
                     "sync-xvfb"         : int,
@@ -564,9 +597,9 @@ OPTION_TYPES = {
                     "daemon"            : bool,
                     "start-via-proxy"   : bool,
                     "attach"            : bool,
-                    "use-display"       : bool,
+                    "use-display"       : str,
                     "fake-xinerama"     : str,
-                    "resize_display"    : bool,
+                    "resize-display"    : str,
                     "tray"              : bool,
                     "pulseaudio"        : bool,
                     "dbus-proxy"        : bool,
@@ -576,7 +609,7 @@ OPTION_TYPES = {
                     "cursors"           : bool,
                     "bell"              : bool,
                     "notifications"     : bool,
-                    "xsettings"         : bool,
+                    "xsettings"         : str,
                     "system-tray"       : bool,
                     "sharing"           : bool,
                     "lock"              : bool,
@@ -598,9 +631,11 @@ OPTION_TYPES = {
                     "modal-windows"     : bool,
                     "bandwidth-detection" : bool,
                     "ssh-upgrade"       : bool,
+                    "splash"            : bool,
                     #arrays of strings:
                     "pulseaudio-configure-commands" : list,
                     "socket-dirs"       : list,
+                    "client-socket-dirs" : list,
                     "remote-xpra"       : list,
                     "encodings"         : list,
                     "proxy-video-encoders" : list,
@@ -612,6 +647,8 @@ OPTION_TYPES = {
                     "compressors"       : list,
                     "packet-encoders"   : list,
                     "key-shortcut"      : list,
+                    "source"            : list,
+                    "source-start"      : list,
                     "start"             : list,
                     "start-child"       : list,
                     "start-after-connect"       : list,
@@ -654,7 +691,6 @@ NON_COMMAND_LINE_OPTIONS = [
     "pdf-printer",
     "postscript-printer",
     "add-printer-options",
-    "file-size-limit",
     ]
 
 START_COMMAND_OPTIONS = [
@@ -665,13 +701,11 @@ START_COMMAND_OPTIONS = [
     ]
 BIND_OPTIONS = ["bind", "bind-tcp", "bind-udp", "bind-ssl", "bind-ws", "bind-wss", "bind-vsock", "bind-rfb"]
 
-#keep track of the options added since v1,
+#keep track of the options added since v3,
 #so we can generate command lines that work with older supported versions:
-OPTIONS_ADDED_SINCE_V1 = [
-    "attach", "open-files", "open-url", "pixel-depth",
-    "uid", "gid", "chdir", "min-port", "rfb-upgrade", "bandwidth-limit",
-    "forward-xdg-open", "modal-windows", "bandwidth-detection",
-    "bind-ssh", "ssh-auth", "ssh-upgrade",
+OPTIONS_ADDED_SINCE_V3 = [
+    "source", "source-start", "headerbar",
+    "splash", "http-scripts", "client-socket-dirs",
     ]
 OPTIONS_COMPAT_NAMES = {
     "--compression_level=" : "-z"
@@ -683,10 +717,11 @@ CLIENT_OPTIONS = ["title", "username", "password", "session-name",
                   "remote-clipboard", "local-clipboard",
                   "tcp-encryption", "tcp-encryption-keyfile", "encryption",  "encryption-keyfile",
                   "systemd-run", "systemd-run-args",
-                  "socket-dir", "socket-dirs",
+                  "socket-dir", "socket-dirs", "client-socket-dirs",
                   "border", "window-close", "min-size", "max-size", "desktop-scaling",
                   "file-transfer", "file-size-limit", "download-path",
                   "open-command", "open-files", "printing", "open-url",
+                  "headerbar",
                   "challenge-handlers",
                   "dbus-proxy",
                   "remote-logging",
@@ -712,7 +747,9 @@ CLIENT_OPTIONS = ["title", "username", "password", "session-name",
                   "env"]
 
 CLIENT_ONLY_OPTIONS = ["username", "swap-keys", "dock-icon",
-                       "tray", "delay-tray", "tray-icon"]
+                       "tray", "delay-tray", "tray-icon",
+                       "attach",
+                       ]
 
 #options that clients can pass to the proxy
 #and which will be forwarded to the new proxy instance process:
@@ -730,19 +767,21 @@ PROXY_START_OVERRIDABLE_OPTIONS = [
     "input-method",
     "microphone", "speaker", "sound-source", "pulseaudio",
     "idle-timeout", "server-idle-timeout",
-    "use-display", "fake-xinerama", "resize_display", "dpi", "pixel-depth",
+    "use-display",
+    "fake-xinerama", "resize-display", "dpi", "pixel-depth",
     "readonly", "keyboard-sync", "cursors", "bell", "notifications", "xsettings",
-    "system-tray", "sharing", "lock", "windows", "webcam", "html",
+    "system-tray", "sharing", "lock", "windows", "webcam", "html", "http-scripts",
     "terminate-children", "exit-with-children", "exit-with-client",
     "av-sync", "global-menus",
     "forward-xdg-open", "modal-windows", "bandwidth-detection",
     "ssh-upgrade",
+    "splash",
     "printing", "file-transfer", "open-command", "open-files", "open-url", "start-new-commands",
     "mmap", "mmap-group", "mdns",
     "auth", "vsock-auth", "tcp-auth", "udp-auth", "ws-auth", "wss-auth", "ssl-auth", "ssh-auth", "rfb-auth",
     "bind", "bind-vsock", "bind-tcp", "bind-udp", "bind-ssl", "bind-ws", "bind-wss", "bind-ssh", "bind-rfb",
     "rfb-upgrade", "bandwidth-limit",
-    "start", "start-child",
+    "start", "start-child", "source", "source-start",
     "start-after-connect", "start-child-after-connect",
     "start-on-connect", "start-child-on-connect",
     "start-on-last-client-exit", "start-child-on-last-client-exit",
@@ -763,10 +802,7 @@ def get_default_key_shortcuts():
                (True,   "#+F2:show_start_new_command"),
                (True,   "#+F3:show_bug_report"),
                (True,   "#+F4:quit"),
-               (True,   "#+F5:increase_quality"),
-               (True,   "#+F6:decrease_quality"),
-               (True,   "#+F7:increase_speed"),
-               (True,   "#+F8:decrease_speed"),
+               (True,   "#+F5:show_window_info"),
                (True,   "#+F10:magic_key"),
                (True,   "#+F11:show_session_info"),
                (True,   "#+F12:toggle_debug"),
@@ -781,21 +817,19 @@ def get_default_key_shortcuts():
                (True,   "#+bar:scalereset"),
                (True,   "#+question:scalingoff"),
                (OSX,    "#+degree:scalereset"),
+               (OSX,    "meta+grave:void"),
+               (OSX,    "meta+shift+asciitilde:void"),
                )
                  if e]
 
 def get_default_systemd_run():
     if WIN32 or OSX:
         return "no"
-    #don't use systemd-run on CentOS / RedHat
-    #(it causes failures with "Failed to create bus connection: No such file or directory")
-    if is_CentOS() or is_RedHat():
-        return "no"
-    if is_Fedora():
-        return "no"
     #systemd-run was previously broken in Fedora 26:
     #https://github.com/systemd/systemd/issues/3388
     #but with newer kernels, it is working again..
+    #now that we test it before using it,
+    #it should be safe to leave it on auto:
     return "auto"
 
 def get_default_pulseaudio_command():
@@ -811,8 +845,6 @@ def get_default_pulseaudio_command():
         "--load=module-x11-publish",
         "--log-level=2", "--log-target=stderr",
         ]
-    #we don't enable memfd on Ubuntu 16.04,
-    #we just don't disable it (because the option does not exist!):
     from xpra.util import envbool
     MEMFD = envbool("XPRA_PULSEAUDIO_MEMFD", False)
     if not MEMFD:
@@ -831,7 +863,7 @@ def get_defaults():
         OPEN_COMMAND, DEFAULT_PULSEAUDIO_CONFIGURE_COMMANDS,
         DEFAULT_ENV, CAN_DAEMONIZE, SYSTEM_PROXY_SOCKET,
         )
-    from xpra.platform.paths import get_download_dir, get_remote_run_xpra_scripts, get_socket_dirs
+    from xpra.platform.paths import get_download_dir, get_remote_run_xpra_scripts, get_socket_dirs, get_client_socket_dirs
     try:
         from xpra.platform.info import get_username
         username = get_username()
@@ -858,25 +890,24 @@ def get_defaults():
     for conf_dir in conf_dirs:
         if conf_dir and os.path.exists(conf_dir):
             break
-    xvfb = detect_xvfb_command(conf_dir, bin_dir)
-    if WIN32:
-        bind_dirs = ["Main"]
-    else:
-        bind_dirs = ["auto"]
+    xvfb = detect_xvfb_command(conf_dir, bin_dir, warn_fn=None)
+    xvfb_str = xvfb_cmd_str(xvfb)
 
     ssl_protocol = "TLSv1_2"
-    if sys.version_info<(2, 7, 9):
-        ssl_protocol = "SSLv23"
 
-    if POSIX and not OSX:
-        from xpra.x11.fakeXinerama import find_libfakeXinerama
-        fake_xinerama = find_libfakeXinerama()
+    if POSIX and not OSX and not (is_Debian() or is_Ubuntu()):
+        #try to find it quickly:
+        try:
+            import glob
+            fake_xinerama = sorted_nicely(glob.glob("/usr/lib*/libfakeXinerama.so.*"))[-1]
+        except Exception:
+            fake_xinerama = "auto"
     else:
         fake_xinerama = "no"
 
     GLOBAL_DEFAULTS = {
                     "encoding"          : "auto",
-                    "title"             : "@title@ on @client-machine@",
+                    "title"             : "@title@ on @hostinfo@",
                     "username"          : username,
                     "password"          : "",
                     "wm-name"           : DEFAULT_NET_WM_NAME,
@@ -906,7 +937,7 @@ def get_defaults():
                     "systemd-run"       : get_default_systemd_run(),
                     "systemd-run-args"  : "",
                     "system-proxy-socket" : SYSTEM_PROXY_SOCKET,
-                    "xvfb"              : " ".join(xvfb),
+                    "xvfb"              : xvfb_str,
                     "chdir"             : "",
                     "socket-dir"        : "",
                     "log-dir"           : "auto",
@@ -915,7 +946,7 @@ def get_defaults():
                     "window-close"      : "auto",
                     "min-size"          : "",
                     "max-size"          : "",
-                    "desktop-scaling"   : "auto",
+                    "desktop-scaling"   : "on",
                     "display"           : "",
                     "tcp-proxy"         : "",
                     "download-path"     : get_download_dir(),
@@ -923,24 +954,26 @@ def get_defaults():
                     "remote-logging"    : "both",
                     "lpadmin"           : "/usr/sbin/lpadmin",
                     "lpinfo"            : "/usr/sbin/lpinfo",
-                    "add-printer-options" : ["-E", "-o printer-is-shared=false", "-u allow:$USER"],
+                    "add-printer-options" : ["-u allow:$USER", "-E", "-o printer-is-shared=false"],
                     "pdf-printer"       : "",
                     "postscript-printer": DEFAULT_POSTSCRIPT_PRINTER,
                     "debug"             : "",
                     "input-method"      : "none",
                     "sound-source"      : "",
                     "html"              : "auto",
+                    "http-scripts"      : "all",
                     "socket-permissions": "600",
                     "exec-wrapper"      : "",
                     "dbus-launch"       : "dbus-launch --sh-syntax --close-stderr",
-                    "webcam"            : ["auto", "no"][OSX],
-                    "mousewheel"        : "on",
+                    "webcam"            : ["auto", "no"][OSX or WIN32],
+                    "mousewheel"        : ["on", "invert-x"][OSX],
                     "input-devices"     : "auto",
                     "shortcut-modifiers": "auto",
                     "open-files"        : "auto",
                     "open-url"          : "auto",
                     "file-transfer"     : "auto",
                     "printing"          : "yes",
+                    "headerbar"         : ["auto", "no"][OSX or WIN32],
                     "challenge-handlers": ["all"],
                     #ssl options:
                     "ssl"               : "auto",
@@ -953,8 +986,8 @@ def get_defaults():
                     "ssl-client-verify-mode"   : "optional",
                     "ssl-server-verify-mode"   : "required",
                     "ssl-verify-flags"  : "X509_STRICT",
-                    "ssl-check-hostname": False,
-                    "ssl-server-hostname": "localhost",
+                    "ssl-check-hostname": True,
+                    "ssl-server-hostname": "",
                     "ssl-options"       : "ALL,NO_COMPRESSION",
                     "quality"           : 0,
                     "min-quality"       : 30,
@@ -962,7 +995,7 @@ def get_defaults():
                     "min-speed"         : 30,
                     "compression_level" : 1,
                     "dpi"               : 0,
-                    "file-size-limit"   : 100,
+                    "file-size-limit"   : "100M",
                     "idle-timeout"      : 0,
                     "server-idle-timeout" : 0,
                     "sync-xvfb"         : 0,
@@ -975,15 +1008,15 @@ def get_defaults():
                     "daemon"            : CAN_DAEMONIZE,
                     "start-via-proxy"   : False,
                     "attach"            : None,
-                    "use-display"       : False,
+                    "use-display"       : "auto",
                     "fake-xinerama"     : fake_xinerama,
-                    "resize-display"    : not OSX and not WIN32,
+                    "resize-display"    : ["no", "yes"][not OSX and not WIN32],
                     "tray"              : True,
                     "pulseaudio"        : DEFAULT_PULSEAUDIO,
                     "dbus-proxy"        : not OSX and not WIN32,
                     "mmap"              : "yes",
                     "mmap-group"        : "auto",
-                    "speaker"           : ["disabled", "on"][has_sound_support()],
+                    "speaker"           : ["disabled", "on"][has_sound_support() and not is_arm()],
                     "microphone"        : ["disabled", "off"][has_sound_support()],
                     "video-scaling"     : "auto",
                     "readonly"          : False,
@@ -993,7 +1026,7 @@ def get_defaults():
                     "cursors"           : True,
                     "bell"              : True,
                     "notifications"     : True,
-                    "xsettings"         : not OSX and not WIN32,
+                    "xsettings"         : ["auto", "no"][int(OSX or WIN32)],
                     "system-tray"       : True,
                     "sharing"           : None,
                     "lock"              : None,
@@ -1013,11 +1046,13 @@ def get_defaults():
                     "desktop-fullscreen": False,
                     "global-menus"      : True,
                     "forward-xdg-open"  : True,
-                    "modal-windows"     : True,
+                    "modal-windows"     : False,
                     "bandwidth-detection" : True,
                     "ssh-upgrade"       : True,
+                    "splash"            : None,
                     "pulseaudio-configure-commands"  : [" ".join(x) for x in DEFAULT_PULSEAUDIO_CONFIGURE_COMMANDS],
                     "socket-dirs"       : get_socket_dirs(),
+                    "client-socket-dirs" : get_client_socket_dirs(),
                     "remote-xpra"       : get_remote_run_xpra_scripts(),
                     "encodings"         : ["all"],
                     "proxy-video-encoders" : [],
@@ -1029,7 +1064,7 @@ def get_defaults():
                     "compressors"       : ["all"],
                     "packet-encoders"   : ["all"],
                     "key-shortcut"      : get_default_key_shortcuts(),
-                    "bind"              : bind_dirs,
+                    "bind"              : ["auto"],
                     "bind-vsock"        : [],
                     "bind-tcp"          : [],
                     "bind-udp"          : [],
@@ -1048,6 +1083,8 @@ def get_defaults():
                     "ssh-auth"          : [],
                     "rfb-auth"          : [],
                     "password-file"     : [],
+                    "source"            : [],
+                    "source-start"      : [],
                     "start"             : [],
                     "start-child"       : [],
                     "start-after-connect"       : [],
@@ -1115,7 +1152,7 @@ def parse_number(numtype, k, v, auto=0):
         return auto
     try:
         return numtype(v)
-    except ValueError as e:
+    except (ValueError, TypeError) as e:
         warn("Warning: cannot parse value '%s' for '%s' as a type %s: %s" % (v, k, numtype, e))
         return None
 
@@ -1125,6 +1162,8 @@ def print_number(i, auto_value=0):
     return str(i)
 
 def parse_with_unit(numtype, v, subunit="bps", min_value=250000):
+    if isinstance(v, int):
+        return v
     #special case for bandwidth-limit, which can be specified using units:
     try:
         v = str(v).lower().strip()
@@ -1151,10 +1190,13 @@ def parse_with_unit(numtype, v, subunit="bps", min_value=250000):
             assert f>=min_value, "value is too low"
         return int(f)
     except Exception as e:
-        raise InitException("invalid value for %s '%s': %s" % (numtype, v, e))
+        raise InitException("invalid value for %s '%s': %s" % (numtype, v, e)) from None
 
 
-def validate_config(d={}, discard=NO_FILE_OPTIONS, extras_types={}, extras_validation={}):
+def validate_config(d=None, discard=NO_FILE_OPTIONS, extras_types=None, extras_validation=None):
+    return do_validate_config(d or {}, discard, extras_types or {}, extras_validation or {})
+
+def do_validate_config(d:dict, discard, extras_types:dict, extras_validation:dict):
     """
         Validates all the options given in a dict with fields as keys and
         strings or arrays of strings as values.
@@ -1212,16 +1254,20 @@ def validate_config(d={}, discard=NO_FILE_OPTIONS, extras_types={}, extras_valid
     return nd
 
 
-def make_defaults_struct(extras_defaults={}, extras_types={}, extras_validation={}, username="", uid=getuid(), gid=getgid()):
+def make_defaults_struct(extras_defaults=None, extras_types=None, extras_validation=None, username="", uid=getuid(), gid=getgid()):
+    return do_make_defaults_struct(extras_defaults or {}, extras_types or {}, extras_validation or {}, username, uid, gid)
+
+def do_make_defaults_struct(extras_defaults:dict, extras_types:dict, extras_validation:dict, username:str, uid:int, gid:int):
     #populate config with default values:
     if not username and uid:
         username = get_username_for_uid(uid)
     defaults = read_xpra_defaults(username, uid, gid)
     return dict_to_validated_config(defaults, extras_defaults, extras_types, extras_validation)
 
-def dict_to_validated_config(d={}, extras_defaults={}, extras_types={}, extras_validation={}):
+def dict_to_validated_config(d:dict, extras_defaults=None, extras_types=None, extras_validation=None):
     options = get_defaults().copy()
-    options.update(extras_defaults)
+    if extras_defaults:
+        options.update(extras_defaults)
     #parse config:
     validated = validate_config(d, extras_types=extras_types, extras_validation=extras_validation)
     options.update(validated)
@@ -1237,7 +1283,7 @@ def dict_to_config(options):
     return config
 
 
-class XpraConfig(object):
+class XpraConfig:
     def __repr__(self):
         return "XpraConfig(%s)" % self.__dict__
 
@@ -1303,11 +1349,14 @@ def fixup_video_all_or_none(options):
     options.proxy_video_encoders = getlist(pvestr, "proxy video encoders", "HARDWARE_ENCODER_OPTIONS")
 
 def fixup_socketdirs(options):
-    if isinstance(options.socket_dirs, str):
-        options.socket_dirs = options.socket_dirs.split(os.path.pathsep)
-    else:
-        assert isinstance(options.socket_dirs, (list, tuple))
-        options.socket_dirs = [v for x in options.socket_dirs for v in x.split(os.path.pathsep)]
+    for option_name in ("socket_dirs", "client_socket_dirs"):
+        value = getattr(options, option_name)
+        if isinstance(value, str):
+            value = value.split(os.path.pathsep)
+        else:
+            assert isinstance(options.socket_dirs, (list, tuple))
+            value = [v for x in value for v in x.split(os.path.pathsep)]
+        setattr(options, option_name, value)
 
 def fixup_pings(options):
     #pings used to be a boolean, True mapped to "5"
@@ -1326,7 +1375,7 @@ def fixup_pings(options):
 
 def fixup_encodings(options):
     try:
-        from xpra.codecs.codec_constants import PREFERED_ENCODING_ORDER
+        from xpra.codecs.codec_constants import PREFERRED_ENCODING_ORDER
     except ImportError:
         return
     RENAME = {"jpg" : "jpeg"}
@@ -1335,7 +1384,7 @@ def fixup_encodings(options):
     estr = _csvstr(options.encodings)
     if estr=="all":
         #replace with an actual list
-        options.encodings = list(PREFERED_ENCODING_ORDER)
+        options.encodings = list(PREFERRED_ENCODING_ORDER)
         return
     encodings = [RENAME.get(x, x) for x in _nodupes(estr)]
     if "rgb" in encodings:
@@ -1405,6 +1454,19 @@ def fixup_clipboard(options):
         options.clipboard_direction = "disabled"
 
 def abs_paths(options):
+    ew = options.exec_wrapper
+    if ew:
+        import shlex
+        ewp = shlex.split(ew)
+        if ewp and not os.path.isabs(ewp[0]):
+            abscmd = which(ewp[0])
+            if abscmd:
+                ewp[0] = abscmd
+                try:
+                    options.exec_wrapper = shlex.join(ewp)
+                except AttributeError:
+                    #(shlex.join requires Python 3.8)
+                    options.exec_wrapper = " ".join(shlex.quote(x) for x in ewp)
     #convert to absolute paths before we daemonize
     for k in ("clipboard-filter-file",
               "tcp-encryption-keyfile", "encryption-keyfile",
@@ -1420,9 +1482,11 @@ def abs_paths(options):
                 continue
             setattr(options, f, os.path.abspath(v))
 
-def fixup_options(options):
+
+def fixup_options(options, skip_encodings=False):
+    if not skip_encodings:
+        fixup_encodings(options)
     fixup_pings(options)
-    fixup_encodings(options)
     fixup_compression(options)
     fixup_packetencoding(options)
     fixup_video_all_or_none(options)
