@@ -9,7 +9,7 @@ import threading
 
 from xpra.net.bytestreams import untilConcludes
 from xpra.util import repr_ellipsized, envint, envbool
-from xpra.os_util import hexstr, force_quit
+from xpra.os_util import hexstr, force_quit, POSIX
 from xpra.log import Logger
 
 log = Logger("proxy")
@@ -47,6 +47,9 @@ class XpraProxy:
         self._exit_code = 0
         signal.signal(signal.SIGINT, self.signal_quit)
         signal.signal(signal.SIGTERM, self.signal_quit)
+        if POSIX:
+            signal.signal(signal.SIGHUP, self.signal_quit)
+            signal.signal(signal.SIGPIPE, self.signal_quit)
 
     def start_threads(self):
         self._to_client.start()
@@ -86,8 +89,8 @@ class XpraProxy:
                     buf = buf[written:]
                     log("%s: written %s bytes", log_name, written)
             log("%s copy loop ended", log_name)
-        except Exception as e:
-            log("%s: %s", log_name, e)
+        except Exception:
+            log("%s", log_name, exc_info=True)
         finally:
             self.quit()
 
