@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2010-2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -12,18 +12,27 @@
 import sys
 import os.path
 import atexit
+import datetime
 import traceback
 
-from xpra.scripts.main import info, warn, no_gtk, validate_encryption, parse_env, configure_env
-from xpra.scripts.config import InitException, FALSE_OPTIONS
+from xpra.scripts.main import (
+    info, warn,
+    no_gtk, bypass_no_gtk,
+    validate_encryption, parse_env, configure_env,
+    stat_X11_display, get_xpra_sessions,
+    )
+from xpra.scripts.config import InitException, FALSE_OPTIONS, parse_bool
+from xpra.common import CLOBBER_USE_DISPLAY, CLOBBER_UPGRADE
+from xpra.exit_codes import EXIT_VFB_ERROR, EXIT_OK
 from xpra.os_util import (
-    SIGNAMES, POSIX, WIN32, OSX, PYTHON3,
+    SIGNAMES, POSIX, WIN32, OSX,
     FDChangeCaptureContext,
     force_quit,
+    get_rand_chars,
     get_username_for_uid, get_home_for_uid, get_shell_for_uid, getuid, setuidgid,
     get_hex_uuid, get_status_output, strtobytes, bytestostr, get_util_logger, osexpand,
     )
-from xpra.util import envbool, unsetenv
+from xpra.util import envbool, unsetenv, noerr
 from xpra.platform.dotxpra import DotXpra
 
 
@@ -89,10 +98,11 @@ def _get_str(prop_name):
 
 
 def save_xvfb_pid(pid):
-    _save_int("_XPRA_SERVER_PID", pid)
+    _save_int("XPRA_XVFB_PID", pid)
 
 def get_xvfb_pid():
-    return _get_int("_XPRA_SERVER_PID")
+    #lookup the legacy name too:
+    return _get_int("XPRA_XVFB_PID") or _get_int("_XPRA_SERVER_PID")
 
 
 def save_uinput_id(uuid):
@@ -106,7 +116,7 @@ def validate_pixel_depth(pixel_depth, starting_desktop=False):
     try:
         pixel_depth = int(pixel_depth)
     except ValueError:
-        raise InitException("invalid value '%s' for pixel depth, must be a number" % pixel_depth)
+        raise InitException("invalid value '%s' for pixel depth, must be a number" % pixel_depth) from None
     if pixel_depth==0:
         pixel_depth = 24
     if pixel_depth not in (8, 16, 24, 30):
@@ -124,6 +134,9 @@ def display_name_check(display_name):
     n = display_name[1:].split(".")[0]    #ie: ":0.0" -> "0"
     try:
         dno = int(n)
+    except (ValueError, TypeError):
+        raise InitException("invalid display number %r" % n) from None
+    else:
         if 0<=dno<10:
             warn("WARNING: low display number: %s" % dno)
             warn(" You are attempting to run the xpra server")
@@ -131,8 +144,6 @@ def display_name_check(display_name):
             warn(" This is generally not what you want.")
             warn(" You should probably use a higher display number")
             warn(" just to avoid any confusion and this warning message.")
-    except IOError:
-        pass
 
 
 def print_DE_warnings():
@@ -265,16 +276,17 @@ def show_encoding_help(opts):
     set_default_level(logging.WARN)
     logging.root.setLevel(logging.WARN)
     for x in get_all_loggers():
-        x.logger.setLevel(logging.WARN)
+        if x.logger.getEffectiveLevel()==logging.INFO:
+            x.logger.setLevel(logging.WARN)
     from xpra.server.server_base import ServerBase
     sb = ServerBase()
     sb.init(opts)
-    from xpra.codecs.codec_constants import PREFERED_ENCODING_ORDER, HELP_ORDER
+    from xpra.codecs.codec_constants import PREFERRED_ENCODING_ORDER, HELP_ORDER
     if "help" in opts.encodings:
-        sb.allowed_encodings = PREFERED_ENCODING_ORDER
-    from xpra.codecs.video_helper import getVideoHelper
-    getVideoHelper().init()
-    sb.init_encodings()
+        sb.allowed_encodings = PREFERRED_ENCODING_ORDER
+    from xpra.server.mixins.encoding_server import EncodingServer
+    assert isinstance(sb, EncodingServer)
+    EncodingServer.threaded_setup(sb)
     from xpra.codecs.loader import encoding_help
     for e in (x for x in HELP_ORDER if x in sb.encodings):
         print(" * %s" % encoding_help(e))
@@ -307,18 +319,19 @@ def set_server_features(opts):
     server_features.mmap            = b(opts.mmap)
     server_features.input_devices   = not opts.readonly and impcheck("keyboard")
     server_features.commands        = impcheck("server.control_command")
-    server_features.dbus            = opts.dbus_proxy and impcheck("dbus")
+    server_features.dbus            = opts.dbus_proxy and impcheck("dbus", "server.dbus")
     server_features.encoding        = impcheck("codecs")
     server_features.logging         = b(opts.remote_logging)
     #server_features.network_state   = ??
+    server_features.shell           = envbool("XPRA_SHELL", True)
     server_features.display         = opts.windows
     server_features.windows         = opts.windows and impcheck("codecs")
     server_features.rfb             = b(opts.rfb_upgrade) and impcheck("server.rfb")
 
 
-def make_desktop_server(clobber):
+def make_desktop_server():
     from xpra.x11.desktop_server import XpraDesktopServer
-    return XpraDesktopServer(clobber)
+    return XpraDesktopServer()
 
 def make_server(clobber):
     from xpra.x11.server import XpraServer
@@ -333,24 +346,37 @@ def make_proxy_server():
     return ProxyServer()
 
 
-def run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=None):
-    #add finally hook to ensure we will run the cleanups
-    #even if we exit because of an exception:
-    try:
-        return do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display)
-    finally:
-        run_cleanups()
-        import gc
-        gc.collect()
+def verify_display(xvfb=None, display_name=None, shadowing=False, log_errors=True, timeout=None):
+    #check that we can access the X11 display:
+    from xpra.x11.vfb_util import verify_display_ready, VFB_WAIT
+    if timeout is None:
+        timeout = VFB_WAIT
+    if not verify_display_ready(xvfb, display_name, shadowing, log_errors, timeout):
+        return 1
+    from xpra.log import Logger
+    log = Logger("screen", "x11")
+    log("X11 display is ready")
+    no_gtk()
+    from xpra.x11.gtk_x11.gdk_display_source import verify_gdk_display
+    display = verify_gdk_display(display_name)
+    if not display:
+        return 1
+    log("GDK can access the display")
+    return 0
 
-
-def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=None):
+def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=None, progress_cb=None):
     assert mode in (
         "start", "start-desktop",
         "upgrade", "upgrade-desktop",
         "shadow", "proxy",
         )
 
+    def _progress(i, msg):
+        if progress_cb:
+            progress_cb(i, msg)
+    progress = _progress
+
+    progress(10, "initializing environment")
     try:
         cwd = os.getcwd()
     except OSError:
@@ -367,46 +393,54 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
         if k.startswith("DBUS_"):
             del os.environ[k]
 
+    use_display = parse_bool("use-display", opts.use_display)
     starting  = mode == "start"
     starting_desktop = mode == "start-desktop"
     upgrading = mode == "upgrade"
     upgrading_desktop = mode == "upgrade-desktop"
     shadowing = mode == "shadow"
     proxying  = mode == "proxy"
-    clobber   = int(upgrading or upgrading_desktop) | int(opts.use_display)*2
-    start_vfb = not (shadowing or proxying or clobber)
 
-    if not proxying and PYTHON3 and POSIX and not OSX:
+    if not proxying and POSIX and not OSX:
         #we don't support wayland servers,
         #so make sure GDK will use the X11 backend:
+        from xpra.os_util import saved_env
+        saved_env["GDK_BACKEND"] = "x11"
         os.environ["GDK_BACKEND"] = "x11"
 
+    has_child_arg = (
+            opts.start_child or
+            opts.start_child_on_connect or
+            opts.start_child_after_connect or
+            opts.start_child_on_last_client_exit
+            )
     if proxying or upgrading or upgrading_desktop:
         #when proxying or upgrading, don't exec any plain start commands:
-        opts.start = opts.start_child = []
+        opts.start = []
+        opts.start_child = []
     elif opts.exit_with_children:
-        assert opts.start_child, "exit-with-children was specified but start-child is missing!"
+        assert has_child_arg, "exit-with-children was specified but start-child* is missing!"
     elif opts.start_child:
         warn("Warning: the 'start-child' option is used,")
         warn(" but 'exit-with-children' is not enabled,")
         warn(" use 'start' instead")
 
-    if opts.bind_rfb and (proxying or starting):
-        get_util_logger().warn("Warning: bind-rfb sockets cannot be used with '%s' mode" % mode)
-        opts.bind_rfb = ""
-
-    if not shadowing and not starting_desktop:
+    if not (shadowing or starting_desktop or upgrading_desktop):
         opts.rfb_upgrade = 0
+        if opts.bind_rfb:
+            get_util_logger().warn("Warning: bind-rfb sockets cannot be used with '%s' mode" % mode)
+            opts.bind_rfb = []
 
-    if upgrading or upgrading_desktop or shadowing:
+    if (upgrading or upgrading_desktop or shadowing) and opts.pulseaudio is None:
         #there should already be one running
+        #so change None ('auto') to False
         opts.pulseaudio = False
 
     #get the display name:
     if shadowing and not extra_args:
         if WIN32 or OSX:
             #just a virtual name for the only display available:
-            display_name = ":0"
+            display_name = "Main"
         else:
             from xpra.scripts.main import guess_X11_display
             dotxpra = DotXpra(opts.socket_dir, opts.socket_dirs)
@@ -418,7 +452,7 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
             error_cb("too many extra arguments (%i): only expected a display number" % len(extra_args))
         if len(extra_args) == 1:
             display_name = extra_args[0]
-            if not shadowing and not proxying and not upgrading and not opts.use_display:
+            if not shadowing and not upgrading and not use_display:
                 display_name_check(display_name)
         else:
             if proxying:
@@ -435,7 +469,7 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
                         break
                 if not display_name:
                     error_cb("you must specify a free virtual display name to use with the proxy server")
-            elif opts.use_display:
+            elif use_display:
                 #only use automatic guess for xpra displays and not X11 displays:
                 display_name = guess_xpra_display(opts.socket_dir, opts.socket_dirs)
             else:
@@ -443,8 +477,33 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
                 # Use the temporary magic value 'S' as marker:
                 display_name = 'S' + str(os.getpid())
 
+    if upgrading or upgrading_desktop:
+        assert display_name, "no display found to upgrade"
+        #try to stop the existing server if it exists:
+        dotxpra = DotXpra(opts.socket_dir, opts.socket_dirs)
+        sessions = get_xpra_sessions(dotxpra, ignore_state=(DotXpra.UNKNOWN, DotXpra.DEAD), matching_display=display_name, query=True)
+        session = sessions.get(display_name)
+        if session:
+            from xpra.platform.paths import get_xpra_command
+            cmd = get_xpra_command()+["exit"]
+            socket_path = session.get("socket-path")
+            if socket_path:
+                cmd += ["socket://%s" % socket_path]
+            else:
+                cmd += [display_name]
+            env = os.environ.copy()
+            #don't wait too long:
+            env["XPRA_CONNECT_TIMEOUT"] = "5"
+            from subprocess import Popen  #pylint: disable=import-outside-toplevel
+            try:
+                p = Popen(cmd, env=env)
+                p.wait()
+            except OSError as e:
+                noerr(sys.stderr.write, "Error: failed to 'exit' the server to upgrade\n")
+                noerr(sys.stderr.write, " %s\n" % e)
+
     if not (shadowing or proxying or upgrading or upgrading_desktop) and \
-    opts.exit_with_children and not opts.start_child:
+    opts.exit_with_children and not has_child_arg:
         error_cb("--exit-with-children specified without any children to spawn; exiting immediately")
 
     atexit.register(run_cleanups)
@@ -456,6 +515,7 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
         write_runner_shell_scripts,
         find_log_dir,
         create_input_devices,
+        source_env,
         )
     script = None
     if POSIX and getuid()!=0:
@@ -465,9 +525,6 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
     gid = int(opts.gid)
     username = get_username_for_uid(uid)
     home = get_home_for_uid(uid)
-    xauth_data = None
-    if start_vfb:
-        xauth_data = get_hex_uuid()
     ROOT = POSIX and getuid()==0
 
     protected_fds = []
@@ -493,6 +550,12 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
             stderr.write(" %s\n" % e)
             del e
 
+    clobber = int(upgrading or upgrading_desktop)*CLOBBER_UPGRADE | int(use_display or 0)*CLOBBER_USE_DISPLAY
+    start_vfb = not (shadowing or proxying or clobber)
+    xauth_data = None
+    if start_vfb:
+        xauth_data = get_hex_uuid()
+
     # if pam is present, try to create a new session:
     pam = None
     PAM_OPEN = POSIX and envbool("XPRA_PAM_OPEN", ROOT and uid!=0)
@@ -559,7 +622,9 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
         # environment:
         write_runner_shell_scripts(script)
 
-    if start_vfb or opts.daemon:
+    extra_expand = {"TIMESTAMP" : datetime.datetime.now().strftime("%Y%m%d-%H%M%S")}
+    log_to_file = opts.daemon or os.environ.get("XPRA_LOG_TO_FILE", "")=="1"
+    if start_vfb or log_to_file:
         #we will probably need a log dir
         #either for the vfb, or for our own log file
         log_dir = opts.log_dir or ""
@@ -572,32 +637,46 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
         if "XPRA_LOG_DIR" not in os.environ:
             os.environ["XPRA_LOG_DIR"] = log_dir
 
-        if opts.daemon:
-            from xpra.server.server_util import select_log_file, open_log_file, redirect_std_to_log
-            log_filename0 = osexpand(select_log_file(log_dir, opts.log_file, display_name), username, uid, gid)
-            logfd = open_log_file(log_filename0)
-            if ROOT and (uid>0 or gid>0):
-                try:
-                    os.fchown(logfd, uid, gid)
-                except (OSError, IOError):
-                    pass
-            stdout, stderr = redirect_std_to_log(logfd, *protected_fds)
+    if log_to_file:
+        from xpra.server.server_util import select_log_file, open_log_file, redirect_std_to_log
+        log_filename0 = osexpand(select_log_file(log_dir, opts.log_file, display_name),
+                                 username, uid, gid, extra_expand)
+        if os.path.exists(log_filename0) and not display_name.startswith("S"):
+            #don't overwrite the log file just yet,
+            #as we may still fail to start
+            log_filename0 += ".new"
+        logfd = open_log_file(log_filename0)
+        if POSIX and ROOT and (uid>0 or gid>0):
             try:
-                stderr.write("Entering daemon mode; "
-                         + "any further errors will be reported to:\n"
-                         + ("  %s\n" % log_filename0))
-            except IOError:
-                #we tried our best, logging another error won't help
-                pass
+                os.fchown(logfd, uid, gid)
+            except OSError as e:
+                noerr(stderr.write, "failed to chown the log file '%s'\n" % log_filename0)
+                noerr(stderr.flush)
+        stdout, stderr = redirect_std_to_log(logfd, *protected_fds)
+        noerr(stderr.write, "Entering daemon mode; "
+                     + "any further errors will be reported to:\n"
+                     + ("  %s\n" % log_filename0))
+        noerr(stderr.flush)
+        os.environ["XPRA_SERVER_LOG"] = log_filename0
+    else:
+        #server log does not exist:
+        os.environ.pop("XPRA_SERVER_LOG", None)
 
     #warn early about this:
     if (starting or starting_desktop) and desktop_display and opts.notifications and not opts.dbus_launch:
         print_DE_warnings()
 
+    if start_vfb and opts.xvfb.find("Xephyr")>=0 and opts.sync_xvfb<=0:
+        warn("Warning: using Xephyr as vfb")
+        warn(" you should also enable the sync-xvfb option")
+        warn(" to keep the Xephyr window updated")
+
+    progress(10, "creating sockets")
     from xpra.net.socket_util import get_network_logger, setup_local_sockets, create_sockets
     sockets = create_sockets(opts, error_cb)
 
     sanitize_env()
+    os.environ.update(source_env(opts.source))
     if POSIX:
         if xrd:
             os.environ["XDG_RUNTIME_DIR"] = xrd
@@ -610,11 +689,8 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
         os.environ["DISPLAY"] = display_name
         if POSIX:
             os.environ["CKCON_X11_DISPLAY"] = display_name
-    else:
-        try:
-            del os.environ["DISPLAY"]
-        except KeyError:
-            pass
+    elif not start_vfb or opts.xvfb.find("Xephyr")<0:
+        os.environ.pop("DISPLAY", None)
     os.environ.update(protected_env)
     from xpra.log import Logger
     log = Logger("server")
@@ -628,19 +704,82 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
     xvfb = None
     xvfb_pid = None
     uinput_uuid = None
+    xauthority = None
+    if start_vfb or clobber:
+        #XAUTHORITY
+        from xpra.x11.vfb_util import (
+            start_Xvfb, check_xvfb_process, parse_resolution,
+            get_xauthority_path,
+            xauth_add,
+            )
+        xauthority = get_xauthority_path(display_name, username, uid, gid)
+        os.environ["XAUTHORITY"] = xauthority
+        if not os.path.exists(xauthority):
+            log("creating XAUTHORITY file '%s'", xauthority)
+            try:
+                with open(xauthority, "a") as f:
+                    if getuid()==0 and (uid!=0 or gid!=0):
+                        os.fchown(f.fileno(), uid, gid)
+            except Exception as e:
+                #trying to continue anyway!
+                log.error("Error trying to create XAUTHORITY file %s:", xauthority)
+                log.error(" %s", e)
+        else:
+            log("found existing XAUTHORITY file '%s'", xauthority)
+        #resolve use-display='auto':
+        if use_display is None or upgrading or upgrading_desktop:
+            #figure out if we have to start the vfb or not:
+            if not display_name:
+                if upgrading or upgrading_desktop:
+                    error_cb("no displays found to upgrade")
+                use_display = False
+            else:
+                progress(20, "connecting to the display")
+                display_no = int(display_name[1:])
+                stat = stat_X11_display(display_no)
+                log("stat_X11_display(%i)=%s", display_no, stat)
+                if not stat:
+                    if upgrading or upgrading_desktop:
+                        error_cb("cannot access display '%s'" % (display_name,))
+                    #no X11 socket to connect to, so we have to start one:
+                    start_vfb = True
+                elif verify_display(None, display_name, log_errors=False, timeout=1)==0:
+                    #accessed OK:
+                    start_vfb = False
+                    #we have already loaded gtk in 'verify_display':
+                    bypass_no_gtk()
+                else:
+                    #verify failed but we can stat the X11 server socket...
+                    #perhaps we need to re-add an xauth entry
+                    if not xauth_data:
+                        xauth_data = get_hex_uuid()
+                        if pam:
+                            pam.set_items({"XAUTHDATA" : xauth_data})
+                    xauth_add(xauthority, display_name, xauth_data, uid, gid)
+                    if verify_display(None, display_name, log_errors=False, timeout=1)!=0:
+                        warn("display %s is not accessible" % (display_name,))
+                    else:
+                        #now OK!
+                        start_vfb = False
     if start_vfb:
+        progress(20, "starting a virtual display")
         assert not proxying and xauth_data
         pixel_depth = validate_pixel_depth(opts.pixel_depth, starting_desktop)
-        from xpra.x11.vfb_util import start_Xvfb, check_xvfb_process
         from xpra.server.server_util import has_uinput
         uinput_uuid = None
         if has_uinput() and opts.input_devices.lower() in ("uinput", "auto") and not shadowing:
-            from xpra.os_util import get_rand_chars
             uinput_uuid = get_rand_chars(UINPUT_UUID_LEN)
-        xvfb, display_name, cleanups = start_Xvfb(opts.xvfb, pixel_depth, display_name, cwd,
-                                                  uid, gid, username, xauth_data, uinput_uuid)
+        vfb_geom = ""
+        try:
+            vfb_geom = parse_resolution(opts.resize_display)
+        except Exception:
+            pass
+
+        xvfb, display_name, cleanups = start_Xvfb(opts.xvfb, vfb_geom, pixel_depth, display_name, cwd,
+                                                  uid, gid, username, uinput_uuid)
         for f in cleanups:
             add_cleanup(f)
+        xauth_add(xauthority, display_name, xauth_data, uid, gid)
         xvfb_pid = xvfb.pid
         #always update as we may now have the "real" display name:
         os.environ["DISPLAY"] = display_name
@@ -649,19 +788,10 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
         if display_name!=odisplay_name and pam:
             pam.set_items({"XDISPLAY" : display_name})
 
-        def check_xvfb():
-            return check_xvfb_process(xvfb)
-    else:
-        if POSIX and clobber:
-            #if we're meant to be using a private XAUTHORITY file,
-            #make sure to point to it:
-            from xpra.x11.vfb_util import get_xauthority_path
-            xauthority = get_xauthority_path(display_name, username, uid, gid)
-            if os.path.exists(xauthority):
-                log("found XAUTHORITY=%s", xauthority)
-                os.environ["XAUTHORITY"] = xauthority
-        def check_xvfb():
+    def check_xvfb(timeout=0):
+        if xvfb is None:
             return True
+        return check_xvfb_process(xvfb, timeout=timeout, command=opts.xvfb)
 
     if POSIX and not OSX and displayfd>0:
         from xpra.platform.displayfd import write_displayfd
@@ -677,16 +807,23 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
             log.error(" %s", str(e) or type(e))
             del e
 
+    if not check_xvfb(1):
+        noerr(stderr.write, "vfb failed to start, exiting\n")
+        return EXIT_VFB_ERROR
+
+    if WIN32 and os.environ.get("XPRA_LOG_FILENAME"):
+        os.environ["XPRA_SERVER_LOG"] = os.environ["XPRA_LOG_FILENAME"]
     if opts.daemon:
-        def noerr(fn, *args):
-            try:
-                fn(*args)
-            except Exception:
-                pass
-        log_filename1 = osexpand(select_log_file(log_dir, opts.log_file, display_name), username, uid, gid)
+        log_filename1 = osexpand(select_log_file(log_dir, opts.log_file, display_name),
+                                 username, uid, gid, extra_expand)
         if log_filename0 != log_filename1:
             # we now have the correct log filename, so use it:
-            os.rename(log_filename0, log_filename1)
+            try:
+                os.rename(log_filename0, log_filename1)
+            except (OSError, IOError):
+                pass
+            else:
+                os.environ["XPRA_SERVER_LOG"] = log_filename1
             if odisplay_name!=display_name:
                 #this may be used by scripts, let's try not to change it:
                 noerr(stderr.write, "Actual display used: %s\n" % display_name)
@@ -699,8 +836,8 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
     del stderr
 
     if not check_xvfb():
-        #xvfb problem: exit now
-        return  1
+        noerr(stderr.write, "vfb failed to start, exiting\n")
+        return EXIT_VFB_ERROR
 
     #create devices for vfb if needed:
     devices = {}
@@ -752,65 +889,80 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
         os.chdir(opts.chdir)
 
     dbus_pid, dbus_env = 0, {}
-    if not shadowing and POSIX and not OSX and not clobber:
-        no_gtk()
-        assert starting or starting_desktop or proxying
-        try:
-            from xpra.server.dbus.dbus_start import start_dbus
-        except ImportError as e:
-            log("dbus components are not installed: %s", e)
-        else:
-            dbus_pid, dbus_env = start_dbus(opts.dbus_launch)
-            if dbus_env:
-                os.environ.update(dbus_env)
+    if not shadowing and POSIX and not OSX:
+        dbuslog = Logger("dbus")
+        if not start_vfb:
+            #try to retrieve it:
+            from xpra.scripts.main import exec_wminfo
+            wminfo = exec_wminfo(display_name)
+            try:
+                dbus_pid = int(wminfo.get("dbus-pid", 0))
+            except ValueError:
+                pass
+            dbus_address = wminfo.get("dbus-address")
+            if dbus_pid and os.path.exists("/proc") and not os.path.exists("/proc/%s" % dbus_pid):
+                dbuslog("dbus pid %s is no longer valid", dbus_pid)
+                dbus_pid = 0
+            if dbus_pid and dbus_address:
+                dbus_env["DBUS_SESSION_BUS_PID"] = str(dbus_pid)
+                dbus_env["DBUS_SESSION_BUS_ADDRESS"] = dbus_address
+                dbus_window = wminfo.get("dbus-window")
+                if dbus_window:
+                    dbus_env["DBUS_SESSION_BUS_WINDOW_ID"] = dbus_window
+                dbuslog("retrieved dbus environment: %s", dbus_env)
+        if not dbus_env:
+            no_gtk()
+            assert starting or starting_desktop or proxying
+            try:
+                from xpra.server.dbus.dbus_start import start_dbus
+            except ImportError as e:
+                dbuslog("dbus components are not installed: %s", e)
+            else:
+                dbus_pid, dbus_env = start_dbus(opts.dbus_launch)
+                if dbus_env:
+                    dbuslog("started new dbus instance: %s", dbus_env)
+                    os.environ.update(dbus_env)
 
-    display = None
     if not proxying:
         if POSIX and not OSX:
             no_gtk()
             if starting or starting_desktop or shadowing:
-                #check that we can access the X11 display:
-                from xpra.x11.vfb_util import verify_display_ready
-                if not verify_display_ready(xvfb, display_name, shadowing):
-                    return 1
-                log("X11 display is ready")
-                no_gtk()
-                from xpra.x11.gtk_x11.gdk_display_source import verify_gdk_display
-                display = verify_gdk_display(display_name)
-                if not display:
-                    return 1
-                log("GDK can access the display")
+                r = verify_display(xvfb, display_name, shadowing)
+                if r:
+                    return r
         #on win32, this ensures that we get the correct screen size to shadow:
         from xpra.platform.gui import init as gui_init
         log("gui_init()")
         gui_init()
 
-    #setup unix domain socket:
-    netlog = get_network_logger()
-    local_sockets = setup_local_sockets(opts.bind,
-                                        opts.socket_dir, opts.socket_dirs,
-                                        display_name, clobber,
-                                        opts.mmap_group, opts.socket_permissions,
-                                        username, uid, gid)
-    netlog("setting up local sockets: %s", local_sockets)
-    sockets += local_sockets
-    if POSIX and (starting or upgrading or starting_desktop or upgrading_desktop):
-        #all unix domain sockets:
-        ud_paths = [sockpath for stype, _, sockpath, _ in local_sockets if stype=="unix-domain"]
-        if ud_paths:
-            #choose one so our xdg-open override script can use to talk back to us:
-            if opts.forward_xdg_open:
-                for x in ("/usr/libexec/xpra", "/usr/lib/xpra"):
-                    xdg_override = os.path.join(x, "xdg-open")
-                    if os.path.exists(xdg_override):
-                        os.environ["PATH"] = x+os.pathsep+os.environ.get("PATH", "")
-                        os.environ["XPRA_SERVER_SOCKET"] = ud_paths[0]
-                        break
-        else:
-            log.warn("Warning: no local server sockets,")
-            if opts.forward_xdg_open:
-                log.warn(" forward-xdg-open cannot be enabled")
-            log.warn(" non-embedded ssh connections will not be available")
+    def init_local_sockets():
+        progress(60, "initializing local sockets")
+        #setup unix domain socket:
+        netlog = get_network_logger()
+        local_sockets = setup_local_sockets(opts.bind,
+                                            opts.socket_dir, opts.socket_dirs,
+                                            display_name, clobber,
+                                            opts.mmap_group, opts.socket_permissions,
+                                            username, uid, gid)
+        netlog("setting up local sockets: %s", local_sockets)
+        sockets.update(local_sockets)
+        if POSIX and (starting or upgrading or starting_desktop or upgrading_desktop):
+            #all unix domain sockets:
+            ud_paths = [sockpath for stype, _, sockpath, _ in local_sockets if stype=="unix-domain"]
+            if ud_paths:
+                #choose one so our xdg-open override script can use to talk back to us:
+                if opts.forward_xdg_open:
+                    for x in ("/usr/libexec/xpra", "/usr/lib/xpra"):
+                        xdg_override = os.path.join(x, "xdg-open")
+                        if os.path.exists(xdg_override):
+                            os.environ["PATH"] = x+os.pathsep+os.environ.get("PATH", "")
+                            os.environ["XPRA_SERVER_SOCKET"] = ud_paths[0]
+                            break
+            else:
+                log.warn("Warning: no local server sockets,")
+                if opts.forward_xdg_open:
+                    log.warn(" forward-xdg-open cannot be enabled")
+                log.warn(" non-embedded ssh connections will not be available")
 
     set_server_features(opts)
 
@@ -825,6 +977,7 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
         if uinput_uuid:
             save_uinput_id(uinput_uuid)
 
+    progress(40, "initializing server")
     if shadowing:
         app = make_shadow_server()
     elif proxying:
@@ -834,21 +987,26 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
             app = make_server(clobber)
         else:
             assert starting_desktop or upgrading_desktop
-            app = make_desktop_server(clobber)
+            app = make_desktop_server()
         app.init_virtual_devices(devices)
 
     try:
         app.exec_cwd = opts.chdir or cwd
         app.display_name = display_name
         app.init(opts)
+        if not app.validate():
+            return 1
+        init_local_sockets()
         app.init_sockets(sockets)
         app.init_dbus(dbus_pid, dbus_env)
-        if not shadowing and (xvfb_pid or clobber):
+        if not shadowing and not proxying:
             app.init_display_pid(xvfb_pid)
+            app.save_pid()
         app.original_desktop_display = desktop_display
         del opts
         if not app.server_ready():
             return 1
+        progress(80, "finalizing")
         app.server_init()
         app.setup()
         app.init_when_ready(_when_ready)
@@ -870,13 +1028,14 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
         return 1
 
     try:
+        progress(100, "running")
         log("running %s", app.run)
         r = app.run()
         log("%s()=%s", app.run, r)
     except KeyboardInterrupt:
         log.info("stopping on KeyboardInterrupt")
         app.cleanup()
-        return 0
+        return EXIT_OK
     except Exception:
         log.error("server error", exc_info=True)
         app.cleanup()
