@@ -1,5 +1,5 @@
 # This file is part of Xpra.
-# Copyright (C) 2016-2017 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2016-2020 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -11,7 +11,7 @@ class Authenticator(SysAuthenticator):
     def __init__(self, username, **kwargs):
         self.value = kwargs.pop("value", None)
         super().__init__(username, **kwargs)
-        self.authenticate = self.authenticate_hmac
+        self.authenticate_check = self.authenticate_hmac
 
     def __repr__(self):
         return "password"
