@@ -1,14 +1,19 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2021 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
+import shlex
 import os.path
+from subprocess import Popen, PIPE
+from threading import Event
+from gi.repository import GLib
 
 from xpra.os_util import pollwait, monotonic_time, bytestostr, osexpand, OSX, POSIX
 from xpra.util import typedict, envbool, csv, engs
+from xpra.make_thread import start_thread
 from xpra.platform import get_username
 from xpra.platform.paths import get_icon_filename
 from xpra.scripts.parsing import sound_option
@@ -17,6 +22,7 @@ from xpra.log import Logger
 
 log = Logger("server")
 soundlog = Logger("sound")
+httplog = Logger("http")
 
 PRIVATE_PULSEAUDIO = envbool("XPRA_PRIVATE_PULSEAUDIO", True)
 
@@ -27,6 +33,8 @@ Mixin for servers that handle audio forwarding.
 class AudioServer(StubServerMixin):
 
     def __init__(self):
+        self.audio_init_done = Event()
+        self.audio_init_done.set()
         self.pulseaudio = False
         self.pulseaudio_command = None
         self.pulseaudio_configure_commands = []
@@ -58,14 +66,25 @@ class AudioServer(StubServerMixin):
         log("AudioServer.init(..) av-sync=%s", self.av_sync)
 
     def setup(self):
+        #the setup code will mostly be waiting for subprocesses to run,
+        #so do it in a separate thread
+        #and just wait for the results where needed:
+        self.audio_init_done.clear()
+        start_thread(self.do_audio_setup, "audio-setup", True)
+        #we don't use threaded_setup() here because it would delay
+        #all the other mixins that use it, for no good reason.
+
+    def do_audio_setup(self):
         self.init_pulseaudio()
         self.init_sound_options()
 
     def cleanup(self):
+        self.audio_init_done.wait(5)
         self.cleanup_pulseaudio()
 
 
     def get_info(self, _proto) -> dict:
+        self.audio_init_done.wait(5)
         info = {}
         if self.pulseaudio is not False:
             info["pulseaudio"] = self.get_pulseaudio_info()
@@ -88,6 +107,105 @@ class AudioServer(StubServerMixin):
         return d
 
 
+    def get_http_scripts(self) -> dict:
+        return {
+            "/audio.mp3" : self.http_audio_mp3_request,
+            }
+
+    def http_audio_mp3_request(self, handler):
+        def err(code=500):
+            handler.send_response(code)
+            return None
+        from xpra.net.http_handler import parse_url
+        args = parse_url(handler)
+        if not args:
+            return err()
+        httplog("http_audio_mp3_request(%s) args=%s", handler, args)
+        uuid = args.get("uuid")
+        if not uuid:
+            httplog.warn("Warning: http-stream audio request, missing uuid")
+            return err()
+        source = None
+        for x in self._server_sources.values():
+            if x.uuid==uuid:
+                source = x
+                break
+        if not source:
+            httplog.warn("Warning: no client matching uuid '%s'", uuid)
+            return err()
+        #don't close the connection when handler.finish() is called,
+        #we will continue to write to this socket as we process more buffers:
+        finish = handler.finish
+        def do_finish():
+            try:
+                finish()
+            except Exception:
+                log("error calling %s", finish, exc_info=True)
+        def noop():
+            pass
+        handler.finish = noop
+        state = {}
+        def new_buffer(_sound_source, data, _metadata, packet_metadata=()):
+            if state.get("failed"):
+                return
+            if not state.get("started"):
+                httplog.warn("buffer received but stream is not started yet")
+                source.stop_sending_sound()
+                err()
+                do_finish()
+                return
+            count = state.get("buffers", 0)
+            httplog("new_buffer [%i] for %s sound stream: %i bytes", count, state.get("codec", "?"), len(data))
+            #httplog("buffer %i: %s", count, hexstr(data))
+            state["buffers"] = count+1
+            try:
+                for x in packet_metadata:
+                    handler.wfile.write(x)
+                handler.wfile.write(data)
+                handler.wfile.flush()
+            except Exception as e:
+                state["failed"] = True
+                httplog("failed to send new audio buffer", exc_info=True)
+                httplog.warn("Error: failed to send audio packet:")
+                httplog.warn(" %s", e)
+                source.stop_sending_sound()
+                do_finish()
+        def new_stream(sound_source, codec):
+            codec = bytestostr(codec)
+            httplog("new_stream: %s", codec)
+            sound_source.codec = codec
+            headers = {
+                "Content-type"      : "audio/mpeg",
+                }
+            try:
+                handler.send_response(200)
+                for k,v in headers.items():
+                    handler.send_header(k, v)
+                handler.end_headers()
+            except ValueError:
+                httplog("new_stream error writing headers", exc_info=True)
+                state["failed"] = True
+                source.stop_sending_sound()
+                do_finish()
+            else:
+                state["started"] = True
+                state["buffers"] = 0
+                state["codec"] = codec
+        start = monotonic_time()
+        def timeout_check():
+            self.http_stream_check_timers.pop(start, None)
+            if not state.get("started"):
+                err()
+                source.stop_sending_sound()
+        if source.sound_source:
+            source.stop_sending_sound()
+        def start_sending_sound():
+            source.start_sending_sound("mp3", volume=1.0, new_stream=new_stream,
+                                       new_buffer=new_buffer, skip_client_codec_check=True)
+        GLib.idle_add(start_sending_sound)
+        self.http_stream_check_timers[start] = (self.timeout_add(1000*5, timeout_check), source.stop_sending_sound)
+
+
     def init_pulseaudio(self):
         soundlog("init_pulseaudio() pulseaudio=%s, pulseaudio_command=%s", self.pulseaudio, self.pulseaudio_command)
         if self.pulseaudio is False:
@@ -120,8 +238,8 @@ class AudioServer(StubServerMixin):
         if PRIVATE_PULSEAUDIO and POSIX and not OSX:
             from xpra.platform.xposix.paths import _get_xpra_runtime_dir, get_runtime_dir
             rd = osexpand(get_runtime_dir())
-            if not os.path.exists(rd) or not os.path.isdir(rd):
-                log.warn("Warning: the runtime directory '%s' does not exist,")
+            if not rd or not os.path.exists(rd) or not os.path.isdir(rd):
+                log.warn("Warning: the runtime directory '%s' does not exist,", rd)
                 log.warn(" cannot start a private pulseaudio server")
             else:
                 xpra_rd = _get_xpra_runtime_dir()
@@ -133,7 +251,6 @@ class AudioServer(StubServerMixin):
                 env["XDG_RUNTIME_DIR"] = self.pulseaudio_private_dir
                 self.pulseaudio_private_socket = os.path.join(self.pulseaudio_private_dir, "pulse", "native")
                 os.environ["XPRA_PULSE_SERVER"] = self.pulseaudio_private_socket
-        import shlex
         cmd = shlex.split(self.pulseaudio_command)
         cmd = list(osexpand(x) for x in cmd)
         #find the absolute path to the command:
@@ -151,6 +268,7 @@ class AudioServer(StubServerMixin):
                     soundlog.info(msg)
                 else:
                     soundlog.warn(msg)
+                self.clean_pulseaudio_private_dir()
                 return
             cmd[0] = pa_cmd
         started_at = monotonic_time()
@@ -172,15 +290,15 @@ class AudioServer(StubServerMixin):
             else:
                 soundlog.warn("Warning: the pulseaudio server process has terminated after %i seconds", int(elapsed))
             self.pulseaudio_proc = None
-        import subprocess
         try:
             soundlog("pulseaudio cmd=%s", " ".join(cmd))
             soundlog("pulseaudio env=%s", env)
-            self.pulseaudio_proc = subprocess.Popen(cmd, env=env)
+            self.pulseaudio_proc = Popen(cmd, env=env)
         except Exception as e:
             soundlog("Popen(%s)", cmd, exc_info=True)
             soundlog.error("Error: failed to start pulseaudio:")
             soundlog.error(" %s", e)
+            self.clean_pulseaudio_private_dir()
             return
         self.add_process(self.pulseaudio_proc, "pulseaudio", cmd, ignore=True, callback=pulseaudio_ended)
         if self.pulseaudio_proc:
@@ -194,11 +312,12 @@ class AudioServer(StubServerMixin):
                 if p is None or p.poll() is not None:
                     return
                 for i, x in enumerate(self.pulseaudio_configure_commands):
-                    proc = subprocess.Popen(x, env=env, shell=True)
+                    proc = Popen(x, env=env, shell=True)
                     self.add_process(proc, "pulseaudio-configure-command-%i" % i, x, ignore=True)
             self.timeout_add(2*1000, configure_pulse)
 
     def cleanup_pulseaudio(self):
+        self.audio_init_done.wait(5)
         proc = self.pulseaudio_proc
         if not proc:
             return
@@ -208,9 +327,8 @@ class AudioServer(StubServerMixin):
             soundlog.info("stopping pulseaudio with pid %s", proc.pid)
             try:
                 #first we try pactl (required on Ubuntu):
-                import subprocess
                 cmd = ["pactl", "exit"]
-                proc = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+                proc = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
                 self.add_process(proc, "pactl exit", cmd, True)
                 r = pollwait(proc)
                 #warning: pactl will return 0 whether it succeeds or not...
@@ -234,6 +352,9 @@ class AudioServer(StubServerMixin):
             now = monotonic_time()
             while (monotonic_time()-now)<1 and os.path.exists(self.pulseaudio_private_socket):
                 time.sleep(0.1)
+        self.clean_pulseaudio_private_dir()
+
+    def clean_pulseaudio_private_dir(self):
         if self.pulseaudio_private_dir:
             if os.path.exists(self.pulseaudio_private_socket):
                 log.warn("Warning: the pulseaudio private socket file still exists:")
@@ -305,23 +426,33 @@ class AudioServer(StubServerMixin):
             self.supports_speaker = False
         if not self.microphone_codecs:
             self.supports_microphone = False
+        #query_pulseaudio_properties may access X11,
+        #do this from the main thread:
         if bool(self.sound_properties):
-            try:
-                from xpra.sound.pulseaudio.pulseaudio_util import set_icon_path, get_info as get_pa_info
-                pa_info = get_pa_info()
-                soundlog("pulseaudio info=%s", pa_info)
-                self.sound_properties.update(pa_info)
-                set_icon_path(get_icon_filename("xpra.png"))
-            except ImportError as e:
-                if POSIX and not OSX:
-                    log.warn("Warning: failed to set pulseaudio tagging icon:")
-                    log.warn(" %s", e)
+            GLib.idle_add(self.query_pulseaudio_properties)
+        GLib.idle_add(self.log_sound_properties)
+        self.audio_init_done.set()
+
+    def query_pulseaudio_properties(self):
+        try:
+            from xpra.sound.pulseaudio.pulseaudio_util import set_icon_path, get_info as get_pa_info
+            pa_info = get_pa_info()
+            soundlog("pulseaudio info=%s", pa_info)
+            self.sound_properties.update(pa_info)
+            set_icon_path(get_icon_filename("xpra.png"))
+        except ImportError as e:
+            if POSIX and not OSX:
+                log.warn("Warning: failed to set pulseaudio tagging icon:")
+                log.warn(" %s", e)
+
+    def log_sound_properties(self):
         soundlog("init_sound_options speaker: supported=%s, encoders=%s",
                  self.supports_speaker, csv(self.speaker_codecs))
         soundlog("init_sound_options microphone: supported=%s, decoders=%s",
                  self.supports_microphone, csv(self.microphone_codecs))
         soundlog("init_sound_options sound properties=%s", self.sound_properties)
 
+
     def get_pulseaudio_info(self) -> dict:
         info = {
             "command"               : self.pulseaudio_command,
