@@ -1,24 +1,25 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2010-2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 #pylint: disable-msg=E1101
 
 import os.path
 
-from xpra.os_util import thread, OSX, POSIX
-from xpra.util import repr_ellipsized
+from xpra.os_util import OSX, POSIX
+from xpra.util import ellipsizer
+from xpra.make_thread import start_thread
 from xpra.server.mixins.stub_server_mixin import StubServerMixin
 from xpra.log import Logger
 
 log = Logger("notify")
 
 
-"""
-Mixin for servers that forward notifications.
-"""
 class NotificationForwarder(StubServerMixin):
+    """
+    Mixin for servers that forward notifications.
+    """
 
     def __init__(self):
         self.notifications_forwarder = None
@@ -34,24 +35,27 @@ class NotificationForwarder(StubServerMixin):
         nf = self.notifications_forwarder
         if nf:
             self.notifications_forwarder = None
-            thread.start_new_thread(nf.release, ())     #@UndefinedVariable
+            start_thread(nf.release, "notifier-release", daemon=True)
 
-    def get_info(self, _source=None):
+    def get_info(self, _source=None) -> dict:
         if not self.notifications_forwarder:
             return {}
         return {"notifications" : self.notifications_forwarder.get_info()}
 
-    def get_server_features(self, _source=None):
+    def get_server_features(self, _source=None) -> dict:
         return {
             "notifications"                : self.notifications,
-            "notifications.close"          : self.notifications,
-            "notifications.actions"        : self.notifications,
+            "notifications.close"          : self.notifications,    #added in v2.3
+            "notifications.actions"        : self.notifications,    #added in v2.3
             }
 
     def parse_hello(self, ss, _caps, send_ui):
         log("parse_hello(%s, {...}, %s) notifications_forwarder=%s", ss, send_ui, self.notifications_forwarder)
         if send_ui and self.notifications_forwarder:
-            client_notification_actions = dict((s.uuid,s.send_notifications_actions) for s in self._server_sources.values())
+            client_notification_actions = dict(
+                (s.uuid, getattr(s, "send_notifications_actions", False))
+                for s in self._server_sources.values()
+                )
             log("client_notification_actions=%s", client_notification_actions)
             self.notifications_forwarder.support_actions = any(v for v in client_notification_actions.values())
 
@@ -111,7 +115,7 @@ class NotificationForwarder(StubServerMixin):
                 app_icon = ""
             log("notify_callback%s icon=%s",
                 (dbus_id, nid, app_name, replaces_nid, app_icon,
-                 summary, body, actions, hints, expire_timeout), repr_ellipsized(str(icon)))
+                 summary, body, actions, hints, expire_timeout), ellipsizer(icon))
             for ss in self._server_sources.values():
                 ss.notify(dbus_id, nid, app_name, replaces_nid, app_icon,
                           summary, body, actions, hints, expire_timeout, icon)
