@@ -7,6 +7,7 @@
 
 from xpra.util import typedict
 from xpra.server.mixins.stub_server_mixin import StubServerMixin
+from xpra.server.source.windows_mixin import WindowsMixin
 from xpra.log import Logger
 
 log = Logger("window")
@@ -15,11 +16,14 @@ metalog = Logger("metadata")
 geomlog = Logger("geometry")
 eventslog = Logger("events")
 
+def noop(*_args):
+    pass
+
 
-"""
-Mixin for servers that forward windows.
-"""
 class WindowServer(StubServerMixin):
+    """
+    Mixin for servers that forward windows.
+    """
 
     def __init__(self):
         # Window id 0 is reserved for "not a window"
@@ -39,7 +43,7 @@ class WindowServer(StubServerMixin):
                 w, h = pv
                 assert w>=0 and h>0 and w<32768 and h<32768
                 return w, h
-            except:
+            except Exception:
                 return default_value
         self.window_min_size = parse_window_size(opts.min_size, (0, 0))
         self.window_max_size = parse_window_size(opts.max_size, (2**15-1, 2**15-1))
@@ -62,18 +66,13 @@ class WindowServer(StubServerMixin):
         self._focus(None, 0, [])
 
 
-    def get_server_features(self, _source):
+    def get_server_features(self, _source) -> dict:
         return {
-            "window_refresh_config" : True,
-            #legacy flags:
-            "window_unmap"          : True,
-            "suspend-resume"        : True,
-            #newer flags:
-            "window.configure.skip-geometry"    : True,
-            "window-filters"        : True,
+            "window_refresh_config" : True,     #v4 clients assume this is available
+            "window-filters"        : True,     #v4 clients assume this is available
             }
 
-    def get_info(self, _proto):
+    def get_info(self, _proto) -> dict:
         return {
             "state" : {
                 "windows" : sum(int(window.is_managed()) for window in tuple(self._id_to_window.values())),
@@ -81,7 +80,7 @@ class WindowServer(StubServerMixin):
             "filters" : tuple((uuid,repr(f)) for uuid, f in self.window_filters),
             }
 
-    def get_ui_info(self, _proto, _client_uuids=None, wids=None, *_args):
+    def get_ui_info(self, _proto, _client_uuids=None, wids=None, *_args) -> dict:
         """ info that must be collected from the UI thread
             (ie: things that query the display)
         """
@@ -100,6 +99,8 @@ class WindowServer(StubServerMixin):
         minw, minh = self.window_min_size
         maxw, maxh = self.window_max_size
         for ss in tuple(self._server_sources.values()):
+            if not isinstance(ss, WindowsMixin):
+                continue
             cminw, cminh = ss.window_min_size
             cmaxw, cmaxh = ss.window_max_size
             minw = max(minw, cminw)
@@ -122,12 +123,15 @@ class WindowServer(StubServerMixin):
 
 
     def send_initial_data(self, ss, caps, send_ui, share_count):
-        if send_ui:
-            self.send_initial_windows(ss, share_count>0)
-            self.send_initial_cursors(ss, share_count>0)
+        if not send_ui:
+            return
+        if not isinstance(ss, WindowsMixin):
+            return
+        self.send_initial_windows(ss, share_count>0)
+        self.send_initial_cursors(ss, share_count>0)
 
 
-    def is_shown(self, _window):
+    def is_shown(self, _window) -> bool:
         return True
 
 
@@ -172,24 +176,26 @@ class WindowServer(StubServerMixin):
 
     def _update_metadata(self, window, pspec):
         metalog("updating metadata on %s: %s", window, pspec)
-        wid = self._window_to_id[window]
+        wid = self._window_to_id.get(window)
+        if not wid:
+            return  #window is already gone
         for ss in self._server_sources.values():
-            ss.window_metadata(wid, window, pspec.name)
+            if isinstance(ss, WindowsMixin):
+                ss.window_metadata(wid, window, pspec.name)
 
 
     def _remove_window(self, window):
         wid = self._window_to_id[window]
         log("remove_window: %s - %s", wid, window)
         for ss in self._server_sources.values():
-            ss.lost_window(wid, window)
+            if isinstance(ss, WindowsMixin):
+                ss.lost_window(wid, window)
         del self._window_to_id[window]
         del self._id_to_window[wid]
         for ss in self._server_sources.values():
-            ss.remove_window(wid, window)
-        try:
-            del self.client_properties[wid]
-        except KeyError:
-            pass
+            if isinstance(ss, WindowsMixin):
+                ss.remove_window(wid, window)
+        self.client_properties.pop(wid, None)
         return wid
 
     def _add_new_window_common(self, window):
@@ -205,7 +211,9 @@ class WindowServer(StubServerMixin):
 
     def _do_send_new_window_packet(self, ptype, window, geometry):
         wid = self._window_to_id[window]
-        for ss in self._server_sources.values():
+        for ss in tuple(self._server_sources.values()):
+            if not isinstance(ss, WindowsMixin):
+                continue
             wprops = self.client_properties.get(wid, {}).get(ss.uuid)
             x, y, w, h = geometry
             #adjust if the transient-for window is not mapped in the same place by the client we send to:
@@ -235,7 +243,7 @@ class WindowServer(StubServerMixin):
             message = packet[6]
         else:
             message = ""
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if ss:
             ss.client_ack_damage(packet_sequence, wid, width, height, decode_time, message)
 
@@ -243,10 +251,11 @@ class WindowServer(StubServerMixin):
         ww, wh = window.get_dimensions()
         self.refresh_window_area(window, 0, 0, ww, wh)
 
-    def refresh_window_area(self, window, x, y, width, height, options={}):
+    def refresh_window_area(self, window, x, y, width, height, options=None):
         wid = self._window_to_id[window]
         for ss in tuple(self._server_sources.values()):
-            ss.damage(wid, window, x, y, width, height, options)
+            damage = getattr(ss, "damage", noop)
+            damage(wid, window, x, y, width, height, options)
 
     def _process_buffer_refresh(self, proto, packet):
         """ can be used for requesting a refresh, or tuning batch config, or both """
@@ -279,7 +288,7 @@ class WindowServer(StubServerMixin):
 
 
     def update_batch_config(self, proto, wid_windows, batch_props, client_properties):
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if ss is None:
             return
         for wid, window in wid_windows.items():
@@ -288,8 +297,8 @@ class WindowServer(StubServerMixin):
             self._set_client_properties(proto, wid, window, client_properties)
             ss.update_batch(wid, window, batch_props)
 
-    def _refresh_windows(self, proto, wid_windows, opts={}):
-        ss = self._server_sources.get(proto)
+    def _refresh_windows(self, proto, wid_windows, opts):
+        ss = self.get_server_source(proto)
         if ss is None:
             return
         for wid, window in wid_windows.items():
@@ -301,7 +310,7 @@ class WindowServer(StubServerMixin):
             ss.refresh(wid, window, opts)
 
     def _idle_refresh_all_windows(self, proto):
-        self.idle_add(self._refresh_windows, proto, self._id_to_window)
+        self.idle_add(self._refresh_windows, proto, self._id_to_window, {})
 
 
     def get_window_position(self, _window):
@@ -311,7 +320,7 @@ class WindowServer(StubServerMixin):
     def _window_mapped_at(self, proto, wid, window, coords=None):
         #record where a window is mapped by a client
         #(in order to support multiple clients and different offsets)
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if not ss:
             return
         if coords:
@@ -334,7 +343,7 @@ class WindowServer(StubServerMixin):
     def _process_configure_window(self, proto, packet):
         log.info("_process_configure_window(%s, %s)", proto, packet)
 
-    def _get_window_dict(self, wids):
+    def _get_window_dict(self, wids) -> dict:
         wd = {}
         for wid in wids:
             window = self._id_to_window.get(wid)
@@ -346,7 +355,7 @@ class WindowServer(StubServerMixin):
         eventslog("suspend(%s)", packet[1:])
         ui = bool(packet[1])
         wd = self._get_window_dict(packet[2])
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if ss:
             ss.suspend(ui, wd)
 
@@ -354,7 +363,7 @@ class WindowServer(StubServerMixin):
         eventslog("resume(%s)", packet[1:])
         ui = bool(packet[1])
         wd = self._get_window_dict(packet[2])
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if ss:
             ss.resume(ui, wd)
 
@@ -378,7 +387,7 @@ class WindowServer(StubServerMixin):
             modifiers = packet[2]
         else:
             modifiers = None
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if ss:
             self._focus(ss, wid, modifiers)
             #if the client focused one of our windows, count this as a user event:
@@ -388,7 +397,7 @@ class WindowServer(StubServerMixin):
     def _focus(self, _server_source, wid, modifiers):
         focuslog("_focus(%s,%s)", wid, modifiers)
 
-    def get_focus(self):
+    def get_focus(self) -> int:
         #can be overriden by subclasses that do manage focus
         #(ie: not shadow servers which only have a single window)
         #default: no focus
@@ -396,14 +405,14 @@ class WindowServer(StubServerMixin):
 
 
     def init_packet_handlers(self):
-        self._authenticated_ui_packet_handlers.update({
-            "map-window":                           self._process_map_window,
-            "unmap-window":                         self._process_unmap_window,
-            "configure-window":                     self._process_configure_window,
-            "close-window":                         self._process_close_window,
-            "focus":                                self._process_focus,
-            "damage-sequence":                      self._process_damage_sequence,
-            "buffer-refresh":                       self._process_buffer_refresh,
-            "suspend":                              self._process_suspend,
-            "resume":                               self._process_resume,
+        self.add_packet_handlers({
+            "map-window" :          self._process_map_window,
+            "unmap-window" :        self._process_unmap_window,
+            "configure-window" :    self._process_configure_window,
+            "close-window" :        self._process_close_window,
+            "focus" :               self._process_focus,
+            "damage-sequence" :     self._process_damage_sequence,
+            "buffer-refresh" :      self._process_buffer_refresh,
+            "suspend" :             self._process_suspend,
+            "resume" :              self._process_resume,
             })
