@@ -1,5 +1,5 @@
 # This file is part of Xpra.
-# Copyright (C) 2017-2018 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2017-2020 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -7,57 +7,140 @@ import sys
 import os.path
 
 from xpra.util import envbool
-from xpra.os_util import OSX, POSIX, shellsub, getuid, getgid, get_util_logger, osexpand, umask_context
+from xpra.os_util import (
+    OSX, POSIX,
+    which,
+    shellsub, getuid,
+    get_util_logger,
+    osexpand, umask_context,
+    )
 from xpra.platform.dotxpra import norm_makepath
 from xpra.scripts.config import InitException
 
 
-def add_cleanup(fn):
-    from xpra.scripts.server import add_cleanup
-    add_cleanup(fn)
+def source_env(source=()) -> dict:
+    log = get_util_logger()
+    env = {}
+    for f in source:
+        e = env_from_sourcing(f)
+        log("source_env %s=%s", f, e)
+        env.update(e)
+    return env
+
+
+# credit: https://stackoverflow.com/a/47080959/428751
+# returns a dictionary of the environment variables resulting from sourcing a file
+def env_from_sourcing(file_to_source_path, include_unexported_variables=False):
+    import json
+    import subprocess
+    from xpra.log import Logger
+    log = Logger("exec")
+    filename = file_to_source_path
+    if not os.path.exists(file_to_source_path):
+        filename = which(file_to_source_path)
+        if not filename:
+            log.error("Error: cannot find file '%s' to source", file_to_source_path)
+            return {}
+    if not os.path.isabs(filename):
+        filename = os.path.abspath(filename)
+    #figure out if this is a script to source,
+    #or if we're meant to execute it directly
+    try:
+        with open(filename, 'rb') as f:
+            first_line = f.readline()
+    except OSError as e:
+        log.error("Error: failed to read from '%s'", filename)
+        log.error(" %s", e)
+        first_line = b""
+    if first_line.startswith(b"\x7fELF") or b"\x00" in first_line:
+        cmd = [filename]
+        def decode(out):
+            env = {}
+            for line in out.splitlines():
+                parts = line.split("=", 1)
+                if len(parts)==2:
+                    env[parts[0]] = parts[1]
+            log("decode(%r)=%s", out, env)
+            return env
+    else:
+        source = '%ssource %s' % ("set -a && " if include_unexported_variables else "", filename)
+        dump = 'python%i.%i -c "import os, json;print(json.dumps(dict(os.environ)))"' % (sys.version_info.major, sys.version_info.minor)
+        cmd = ['/bin/bash', '-c', '%s && %s' % (source, dump)]
+        def decode(out):
+            try:
+                env = json.loads(out)
+                log("json loads(%r)=%s", out, env)
+                return env
+            except json.decoder.JSONDecodeError:
+                log.error("Error decoding json output from sourcing script '%s'",
+                          file_to_source_path, exc_info=True)
+                return {}
+    try:
+        log("env_from_sourcing%s cmd=%s", (filename, include_unexported_variables), cmd)
+        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)
+        out = proc.communicate()[0]
+        if proc.returncode!=0:
+            log.error("Error %i running source script '%s'", proc.returncode, filename)
+    except OSError as e:
+        log("env_from_sourcing%s", (filename, include_unexported_variables), exc_info=True)
+        log(" stdout=%r (%s)", out, type(out))
+        log.error("Error running source script '%s'", proc.returncode, file_to_source_path)
+        log.error(" %s", e)
+        return {}
+    log("stdout(%s)=%r", filename, out)
+    if not out:
+        return {}
+    try:
+        out = out.decode()
+    except UnicodeDecodeError:
+        log.error("Error decoding output from '%s'", filename, exc_info=True)
+        return {}
+    return decode(out)
+
 
 def sh_quotemeta(s):
-    return "'" + s.replace("'", "'\\''") + "'"
+    return b"'" + s.replace(b"'", b"'\\''") + b"'"
 
 def xpra_runner_shell_script(xpra_file, starting_dir, socket_dir):
     script = []
-    script.append("#!/bin/sh\n")
-    for var, value in os.environ.items():
+    script.append(b"#!/bin/sh\n")
+    for var, value in os.environb.items():
         # these aren't used by xpra, and some should not be exposed
         # as they are either irrelevant or simply do not match
         # the new environment used by xpra
         # TODO: use a whitelist
-        if var in ["XDG_SESSION_COOKIE", "LS_COLORS", "DISPLAY"]:
+        if var in (b"XDG_SESSION_COOKIE", b"LS_COLORS", b"DISPLAY"):
             continue
         #XPRA_SOCKET_DIR is a special case, it is handled below
-        if var=="XPRA_SOCKET_DIR":
+        if var==b"XPRA_SOCKET_DIR":
             continue
-        if var.startswith("BASH_FUNC"):
+        if var.startswith(b"BASH_FUNC"):
             #some versions of bash will apparently generate functions
             #that cannot be reloaded using this script
             continue
         # :-separated envvars that people might change while their server is
         # going:
-        if var in ("PATH", "LD_LIBRARY_PATH", "PYTHONPATH"):
+        if var in (b"PATH", b"LD_LIBRARY_PATH", b"PYTHONPATH"):
             #prevent those paths from accumulating the same values multiple times,
             #only keep the first one:
-            pval = value.split(os.pathsep)      #ie: ["/usr/bin", "/usr/local/bin", "/usr/bin"]
+            pathsep = os.pathsep.encode()
+            pval = value.split(pathsep)      #ie: ["/usr/bin", "/usr/local/bin", "/usr/bin"]
             seen = set()
-            value = os.pathsep.join(x for x in pval if not (x in seen or seen.add(x)))
-            script.append("%s=%s:\"$%s\"; export %s\n"
+            value = pathsep.join(x for x in pval if not (x in seen or seen.add(x)))
+            script.append(b"%s=%s:\"$%s\"; export %s\n"
                           % (var, sh_quotemeta(value), var, var))
         else:
-            script.append("%s=%s; export %s\n"
+            script.append(b"%s=%s; export %s\n"
                           % (var, sh_quotemeta(value), var))
     #XPRA_SOCKET_DIR is a special case, we want to honour it
     #when it is specified, but the client may override it:
     if socket_dir:
-        script.append('if [ -z "${XPRA_SOCKET_DIR}" ]; then\n');
-        script.append('    XPRA_SOCKET_DIR=%s; export XPRA_SOCKET_DIR\n' % sh_quotemeta(os.path.expanduser(socket_dir)))
-        script.append('fi\n');
+        script.append(b'if [ -z "${XPRA_SOCKET_DIR}" ]; then\n')
+        script.append(b'    XPRA_SOCKET_DIR="%s"; export XPRA_SOCKET_DIR\n' % sh_quotemeta(os.path.expanduser(socket_dir).encode()))
+        script.append(b'fi\n')
     # We ignore failures in cd'ing, b/c it's entirely possible that we were
     # started from some temporary directory and all paths are absolute.
-    script.append("cd %s\n" % sh_quotemeta(starting_dir))
+    script.append(b"cd %s\n" % sh_quotemeta(starting_dir.encode()))
     if OSX:
         #OSX contortions:
         #The executable is the python interpreter,
@@ -66,8 +149,8 @@ def xpra_runner_shell_script(xpra_file, starting_dir, socket_dir):
         bini = sexec.rfind("Resources/bin/")
         if bini>0:
             sexec = os.path.join(sexec[:bini], "Resources", "MacOS", "Xpra")
-        script.append("_XPRA_SCRIPT=%s\n" % (sh_quotemeta(sexec),))
-        script.append("""
+        script.append(b"_XPRA_SCRIPT=%s\n" % (sh_quotemeta(sexec.encode()),))
+        script.append(b"""
 if which "$_XPRA_SCRIPT" > /dev/null; then
     # Happypath:
     exec "$_XPRA_SCRIPT" "$@"
@@ -77,9 +160,9 @@ else
 fi
 """)
     else:
-        script.append("_XPRA_PYTHON=%s\n" % (sh_quotemeta(sys.executable),))
-        script.append("_XPRA_SCRIPT=%s\n" % (sh_quotemeta(xpra_file),))
-        script.append("""
+        script.append(b"_XPRA_PYTHON=%s\n" % (sh_quotemeta(sys.executable.encode()),))
+        script.append(b"_XPRA_SCRIPT=%s\n" % (sh_quotemeta(xpra_file.encode()),))
+        script.append(b"""
 if which "$_XPRA_PYTHON" > /dev/null && [ -e "$_XPRA_SCRIPT" ]; then
     # Happypath:
     exec "$_XPRA_PYTHON" "$_XPRA_SCRIPT" "$@"
@@ -92,9 +175,10 @@ END
     exec xpra "$@"
 fi
 """)
-    return "".join(script)
+    return b"".join(script)
 
 def write_runner_shell_scripts(contents, overwrite=True):
+    assert POSIX
     # This used to be given a display-specific name, but now we give it a
     # single fixed name and if multiple servers are started then the last one
     # will clobber the rest.  This isn't great, but the tradeoff is that it
@@ -104,14 +188,15 @@ def write_runner_shell_scripts(contents, overwrite=True):
     # people run into problems or autodiscovery turns out to be less useful
     # than expected.
     log = get_util_logger()
+    MODE = 0o700
     from xpra.platform.paths import get_script_bin_dirs
     for d in get_script_bin_dirs():
         scriptdir = osexpand(d)
         if not os.path.exists(scriptdir):
             try:
-                os.mkdir(scriptdir, 0o700)
+                os.mkdir(scriptdir, MODE)
             except Exception as e:
-                log("os.mkdir(%s, 0o700)", scriptdir, exc_info=True)
+                log("os.mkdir(%s, %s)", scriptdir, oct(MODE), exc_info=True)
                 log.warn("Warning: failed to create script directory '%s':", scriptdir)
                 log.warn(" %s", e)
                 if scriptdir.startswith("/var/run/user") or scriptdir.startswith("/run/user"):
@@ -124,9 +209,9 @@ def write_runner_shell_scripts(contents, overwrite=True):
         # environment:
         try:
             with umask_context(0o022):
-                h = os.open(scriptpath, os.O_WRONLY|os.O_CREAT|os.O_TRUNC, 0o700)
+                h = os.open(scriptpath, os.O_WRONLY|os.O_CREAT|os.O_TRUNC, MODE)
                 try:
-                    os.write(h, contents.encode())
+                    os.write(h, contents)
                 finally:
                     os.close(h)
         except Exception as e:
@@ -162,12 +247,12 @@ def open_log_file(logpath):
     if os.path.exists(logpath):
         try:
             os.rename(logpath, logpath + ".old")
-        except:
+        except OSError:
             pass
     try:
         return os.open(logpath, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, 0o644)
     except OSError as e:
-        raise InitException("cannot open log file '%s': %s" % (logpath, e))
+        raise InitException("cannot open log file '%s': %s" % (logpath, e)) from None
 
 def select_log_file(log_dir, log_file, display_name):
     """ returns the log file path we should be using given the parameters,
@@ -216,48 +301,47 @@ def redirect_std_to_log(logfd, *noclose_fds):
 def daemonize():
     os.chdir("/")
     if os.fork():
-        os._exit(0)
+        os._exit(0)     #pylint: disable=protected-access
     os.setsid()
     if os.fork():
-        os._exit(0)
+        os._exit(0)     #pylint: disable=protected-access
 
 
-def write_pidfile(pidfile, uid, gid):
+def write_pidfile(pidfile):
     log = get_util_logger()
     pidstr = str(os.getpid())
+    inode = 0
     try:
         with open(pidfile, "w") as f:
             os.fchmod(f.fileno(), 0o600)
             f.write("%s\n" % pidstr)
             try:
                 inode = os.fstat(f.fileno()).st_ino
-            except:
-                inode = -1
-            if POSIX and uid!=getuid() or gid!=getgid():
-                try:
-                    os.fchown(f.fileno(), uid, gid)
-                except:
-                    pass
+            except OSError:
+                inode = 0
         log.info("wrote pid %s to '%s'", pidstr, pidfile)
-        def cleanuppidfile():
-            #verify this is the right file!
-            log("cleanuppidfile: inode=%i", inode)
-            if inode>0:
-                try:
-                    i = os.stat(pidfile).st_ino
-                    log("cleanuppidfile: current inode=%i", i)
-                    if i!=inode:
-                        return
-                except:
-                    pass
-            try:
-                os.unlink(pidfile)
-            except:
-                pass
-        add_cleanup(cleanuppidfile)
     except Exception as e:
         log.error("Error: failed to write pid %i to pidfile '%s':", os.getpid(), pidfile)
         log.error(" %s", e)
+    return inode
+
+def rm_pidfile(pidfile, inode):
+    #verify this is the right file!
+    log = get_util_logger()
+    log("cleanuppidfile(%s, %s)", pidfile, inode)
+    if inode>0:
+        try:
+            i = os.stat(pidfile).st_ino
+            log("cleanuppidfile: current inode=%i", i)
+            if i!=inode:
+                return 0
+        except OSError:
+            pass
+    try:
+        os.unlink(pidfile)
+    except OSError:
+        log("rm_pidfile(%s, %s)", pidfile, inode, exc_info=True)
+    return 0
 
 
 def get_uinput_device_path(device):
@@ -270,17 +354,16 @@ def get_uinput_device_path(device):
         import ctypes
         l = 16
         buf = ctypes.create_string_buffer(l)
-        _IOC_READ = 2
         #this magic value was calculated using the C macros:
         l = fcntl.ioctl(fd, 2148554028, buf)
-        if l>0 and l<16:
-            virt_dev_path = buf.raw[:l].rstrip("\0")
+        if 0<l<16:
+            virt_dev_path = buf.raw[:l].rstrip(b"\0")
             log("UI_GET_SYSNAME(%s)=%s", fd, virt_dev_path)
             uevent_path = b"/sys/devices/virtual/input/%s" % virt_dev_path
-            event_dirs = [x for x in os.listdir(uevent_path) if x.startswith("event")]
+            event_dirs = [x for x in os.listdir(uevent_path) if x.startswith(b"event")]
             log("event dirs(%s)=%s", uevent_path, event_dirs)
             for d in event_dirs:
-                uevent_filename = os.path.join(uevent_path, d, "uevent")
+                uevent_filename = os.path.join(uevent_path, d, b"uevent")
                 uevent_conf = open(uevent_filename, "rb").read()
                 for line in uevent_conf.splitlines():
                     if line.find(b"=")>0:
@@ -291,6 +374,7 @@ def get_uinput_device_path(device):
                             log("found device path: %s" % dev_path)
                             return dev_path
     except Exception as e:
+        log("get_uinput_device_path(%s)", device, exc_info=True)
         log.error("Error: cannot query uinput device path:")
         log.error(" %s", e)
     return None
@@ -352,7 +436,7 @@ def create_uinput_device(uuid, uid, events, name):
 
 def create_uinput_pointer_device(uuid, uid):
     if not envbool("XPRA_UINPUT_POINTER", True):
-        return
+        return None
     from uinput import (
         REL_X, REL_Y, REL_WHEEL,                    #@UnresolvedImport
         BTN_LEFT, BTN_RIGHT, BTN_MIDDLE, BTN_SIDE,  #@UnresolvedImport
@@ -370,7 +454,7 @@ def create_uinput_pointer_device(uuid, uid):
 
 def create_uinput_touchpad_device(uuid, uid):
     if not envbool("XPRA_UINPUT_TOUCHPAD", False):
-        return
+        return None
     from uinput import (
         BTN_TOUCH, ABS_X, ABS_Y, ABS_PRESSURE,      #@UnresolvedImport
         )
