@@ -14,7 +14,7 @@ from xpra.server.window.batch_config import DamageBatchConfig
 from xpra.server.server_core import ClientException
 from xpra.codecs.video_helper import getVideoHelper
 from xpra.codecs.codec_constants import video_spec
-from xpra.net.compression import use_lz4, use_lzo, use_brotli
+from xpra.net.compression import use
 from xpra.os_util import monotonic_time, strtobytes
 from xpra.server.background_worker import add_work_item
 from xpra.util import csv, typedict, envint
@@ -25,6 +25,7 @@ proxylog = Logger("proxy")
 statslog = Logger("stats")
 
 MIN_PIXEL_RECALCULATE = envint("XPRA_MIN_PIXEL_RECALCULATE", 2000)
+DEFAULT_VREFRESH = envint("XPRA_DEFAULT_VREFRESH", 100)
 
 
 """
@@ -38,22 +39,17 @@ class EncodingsMixin(StubSourceMixin):
         return bool(caps.strtupleget("encodings"))
 
 
-    def __init__(self):
-        self.server_core_encodings = []
-        self.server_encodings = []
-        self.default_encoding = None
-        self.scaling_control = None
+    def __init__(self, *_args):
+        pass
 
-        self.default_quality = 40       #default encoding quality for lossy encodings
-        self.default_min_quality = 10   #default minimum encoding quality
-        self.default_speed = 40         #encoding speed (only used by x264)
-        self.default_min_speed = 10     #default minimum encoding speed
+    def init_state(self):
+        self.wants_encodings = False
+        self.wants_features = False
 
         #contains default values, some of which may be supplied by the client:
         self.default_batch_config = DamageBatchConfig()
         self.global_batch_config = self.default_batch_config.clone()      #global batch config
 
-        self.vrefresh = -1
         self.supports_transparency = False
         self.encoding = None                        #the default encoding for all windows
         self.encodings = ()                         #all the encodings supported by the client
@@ -67,8 +63,8 @@ class EncodingsMixin(StubSourceMixin):
         self.auto_refresh_delay = 0
 
         self.zlib = True
-        self.lz4 = use_lz4
-        self.lzo = use_lzo
+        self.lz4 = use("lz4")
+        self.lzo = use("lzo")
 
         #for managing the recalculate_delays work:
         self.calculate_window_pixels = {}
@@ -80,6 +76,7 @@ class EncodingsMixin(StubSourceMixin):
         #new encoders, so we must make a deep copy to preserve the original
         #which may be used by other clients (other ServerSource instances)
         self.video_helper = getVideoHelper().clone()
+        self.cuda_device_context = None
 
 
     def init_from(self, _protocol, server):
@@ -100,7 +97,10 @@ class EncodingsMixin(StubSourceMixin):
         self.queue_encode(None)
         #this should be a noop since we inherit an initialized helper:
         self.video_helper.cleanup()
-
+        cdd = self.cuda_device_context
+        if cdd:
+            self.cuda_device_context = None
+            cdd.free()
 
     def all_window_sources(self):
         #we can't assume that the window mixin is loaded:
@@ -151,10 +151,7 @@ class EncodingsMixin(StubSourceMixin):
         for wid in wids:
             #this is safe because we only add to this set from other threads:
             self.calculate_window_ids.remove(wid)
-            try:
-                del self.calculate_window_pixels[wid]
-            except KeyError:
-                pass
+            self.calculate_window_pixels.pop(wid, None)
             ws = self.window_sources.get(wid)
             if ws is None:
                 continue
@@ -254,28 +251,23 @@ class EncodingsMixin(StubSourceMixin):
 
         #general features:
         self.zlib = c.boolget("zlib", True)
-        self.lz4 = c.boolget("lz4", False) and use_lz4
-        self.lzo = c.boolget("lzo", False) and use_lzo
-        self.brotli = c.boolget("brotli", False) and use_brotli
+        self.lz4 = c.boolget("lz4", False) and use("lz4")
+        self.lzo = c.boolget("lzo", False) and use("lzo")
+        self.brotli = c.boolget("brotli", False) and use("brotli")
         log("compressors: zlib=%s, lz4=%s, lzo=%s, brotli=%s",
             self.zlib, self.lz4, self.lzo, self.brotli)
 
-        self.vrefresh = c.intget("vrefresh", -1)
-
-        #assume 50Hz:
-        ms_per_frame = 1000//50
-        if 30<=self.vrefresh<=500:
-            #looks like a valid vrefresh value, use it:
-            ms_per_frame = 1000//self.vrefresh
-        default_min_delay = max(DamageBatchConfig.MIN_DELAY, ms_per_frame)
+        delay = DamageBatchConfig.START_DELAY
         dbc = self.default_batch_config
         dbc.always      = bool(batch_value("always", DamageBatchConfig.ALWAYS))
-        dbc.min_delay   = batch_value("min_delay", default_min_delay, 0, 1000)
+        dbc.min_delay   = batch_value("min_delay", DamageBatchConfig.MIN_DELAY, 0, 1000)
         dbc.max_delay   = batch_value("max_delay", DamageBatchConfig.MAX_DELAY, 1, 15000)
         dbc.max_events  = batch_value("max_events", DamageBatchConfig.MAX_EVENTS)
         dbc.max_pixels  = batch_value("max_pixels", DamageBatchConfig.MAX_PIXELS)
         dbc.time_unit   = batch_value("time_unit", DamageBatchConfig.TIME_UNIT, 1)
-        dbc.delay       = batch_value("delay", DamageBatchConfig.START_DELAY, 0)
+        self.vrefresh = c.intget("vrefresh", -1)
+        dbc.match_vrefresh(self.vrefresh)
+        dbc.delay       = batch_value("delay", delay, dbc.min_delay)
         log("default batch config: %s", dbc)
 
         #encodings:
@@ -366,7 +358,23 @@ class EncodingsMixin(StubSourceMixin):
         log("default encoding options: %s", self.default_encoding_options)
         self.auto_refresh_delay = c.intget("auto_refresh_delay", 0)
 
+        #are we going to need a cuda context?
+        common_encodings = tuple(x for x in self.encodings if x in self.server_encodings)
+        from xpra.codecs.loader import has_codec
+        if "jpeg" in common_encodings and has_codec("enc_nvjpeg"):
+            try:
+                from xpra.codecs.cuda_common.cuda_context import get_device_context
+                self.cuda_device_context = get_device_context(self.encoding_options)
+                log("cuda_device_context=%s", self.cuda_device_context)
+            except Exception as e:
+                log("failed to get a cuda device contenxt using encoding options %s",
+                    self.encoding_options, exc_info=True)
+                log.error("Error: failed to allocate a CUDA context")
+                log.error(" NVJPEG will not be available")
+
     def print_encoding_info(self):
+        log("print_encoding_info() core-encodings=%s, server-core-encodings=%s",
+            self.core_encodings, self.server_core_encodings)
         others = tuple(x for x in self.core_encodings
                        if x in self.server_core_encodings and x!=self.encoding)
         if self.encoding=="auto":
@@ -465,13 +473,9 @@ class EncodingsMixin(StubSourceMixin):
                 "auto_refresh"      : self.auto_refresh_delay,
                 "lz4"               : self.lz4,
                 "lzo"               : self.lzo,
-                "vertical-refresh"  : self.vrefresh,
                 }
         ieo = dict(self.icons_encoding_options)
-        try:
-            del ieo["default.icons"]
-        except KeyError:
-            pass
+        ieo.pop("default.icons", None)
         #encoding:
         info.update({
                      "encodings"        : {
