@@ -31,15 +31,21 @@ class FilePrintMixin(FileTransferHandler, StubSourceMixin):
         return bool(caps.boolget("file-transfer") or caps.boolget("printing"))
 
 
+    def __init__(self, *_args):
+        pass
+
     def init_state(self):
         self.printers = {}
         self.printers_added = set()
+        #duplicated from clientinfo mixin
+        self.machine_id = ""
 
     def cleanup(self):
         self.remove_printers()
 
     def parse_client_caps(self, c : dict):
         FileTransferHandler.parse_file_transfer_caps(self, c)
+        self.machine_id = c.strget("machine_id")
 
     def get_info(self) -> dict:
         return {
@@ -77,10 +83,7 @@ class FilePrintMixin(FileTransferHandler, StubSourceMixin):
             npd = printers.get(k)
             if cpd==npd:
                 #unchanged: make sure we don't try adding it again:
-                try:
-                    del printers[k]
-                except KeyError:
-                    pass
+                printers.pop(k, None)
                 continue
             if npd is None:
                 log("printer %s no longer exists", k)
@@ -89,10 +92,7 @@ class FilePrintMixin(FileTransferHandler, StubSourceMixin):
                 log(" was %s", cpd)
                 log(" now %s", npd)
             #remove it:
-            try:
-                del self.printers[k]
-            except KeyError:
-                pass
+            self.printers.pop(k, None)
             self.remove_printer(k)
         #expand it here so the xpraforwarder doesn't need to import anything xpra:
         attributes = {"display"         : os.environ.get("DISPLAY"),
@@ -103,7 +103,7 @@ class FilePrintMixin(FileTransferHandler, StubSourceMixin):
         if auth:
             auth_password_file = None
             try:
-                name, authclass, authoptions = auth
+                name, _, authclass, authoptions = auth
                 auth_password_file = authoptions.get("file")
                 log("file for %s / %s: '%s'", name, authclass, password_file)
             except Exception as e:
