@@ -65,6 +65,7 @@ class WindowPerformanceStatistics:
         self.packet_count = 0
 
         self.last_resized = 0
+        self.last_packet_time = 0
 
         #these values are calculated from the values above (see update_averages)
         self.target_latency = self.DEFAULT_TARGET_LATENCY
@@ -255,18 +256,12 @@ class WindowPerformanceStatistics:
                 log.error(" connection may be closed or closing,")
                 log.error(" sequence numbers missing: %s", csv(drop_missing_acks))
                 for sequence in drop_missing_acks:
-                    try:
-                        del self.damage_ack_pending[sequence]
-                    except KeyError:
-                        pass
+                    self.damage_ack_pending.pop(sequence, None)
         return packets_backlog, pixels_backlog, bytes_backlog
 
     def get_acks_pending(self):
         return sum(1 for x in self.damage_ack_pending.values() if x[0]!=0)
 
-    def get_acks_pending(self):
-        return sum(1 for x in self.damage_ack_pending.values() if x[0]!=0)
-
     def get_late_acks(self, latency):
         now = monotonic_time()
         sent_before = now-latency
