@@ -1,24 +1,23 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # This file is part of Xpra.
-# Copyright (C) 2010-2018 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import sys
 import os
 
-from xpra.sound.common import FLAC_OGG, OPUS_OGG, OPUS_MKA, SPEEX_OGG, VORBIS_OGG, VORBIS_MKA, \
-                                AAC_MPEG4, WAV_LZ4, WAV_LZO, \
-                                VORBIS, FLAC, MP3, MP3_MPEG4, OPUS, SPEEX, WAV, WAVPACK, \
-                                MPEG4, MKA, OGG
-
-from xpra.os_util import WIN32, OSX, POSIX, PYTHON3, bytestostr
-from xpra.util import csv, engs, parse_simple_dict, envint, envbool
+from xpra.sound.common import (
+    FLAC_OGG, OPUS_OGG, OPUS_MKA, SPEEX_OGG, VORBIS_OGG, VORBIS_MKA, \
+    AAC_MPEG4, WAV_LZ4, WAV_LZO, \
+    VORBIS, FLAC, MP3, MP3_MPEG4, OPUS, SPEEX, WAV, WAVPACK, MP3_ID3V2, \
+    MPEG4, MKA, OGG,
+    )
+from xpra.os_util import WIN32, OSX, POSIX, bytestostr
+from xpra.util import csv, engs, parse_simple_dict, reverse_dict, envint, envbool
 from xpra.log import Logger
-log = Logger("sound", "gstreamer")
 
-if PYTHON3:
-    unicode = str           #@ReservedAssignment
+log = Logger("sound", "gstreamer")
 
 
 #used on the server (reversed):
@@ -31,6 +30,9 @@ GST_QUEUE_LEAK_DOWNSTREAM     = 2
 GST_QUEUE_LEAK_DEFAULT = GST_QUEUE_LEAK_DOWNSTREAM
 MS_TO_NS = 1000000
 
+GST_FLOW_OK = 0     #Gst.FlowReturn.OK
+
+
 QUEUE_LEAK = envint("XPRA_SOUND_QUEUE_LEAK", GST_QUEUE_LEAK_DEFAULT)
 if QUEUE_LEAK not in (GST_QUEUE_NO_LEAK, GST_QUEUE_LEAK_UPSTREAM, GST_QUEUE_LEAK_DOWNSTREAM):
     log.error("invalid leak option %s", QUEUE_LEAK)
@@ -45,7 +47,7 @@ def get_queue_time(default_value=450, prefix=""):
 ALLOW_SOUND_LOOP = envbool("XPRA_ALLOW_SOUND_LOOP", False)
 USE_DEFAULT_DEVICE = envbool("XPRA_USE_DEFAULT_DEVICE", True)
 IGNORED_INPUT_DEVICES = os.environ.get("XPRA_SOUND_IGNORED_INPUT_DEVICES", "bell.ogg,bell.wav").split(",")
-IGNORED_OUTPUT_DEVICES = os.environ.get("XPRA_SOUND_IGNORED_OUTPUT_DEVICES", "").split(",")
+IGNORED_OUTPUT_DEVICES = os.environ.get("XPRA_SOUND_IGNORED_OUTPUT_DEVICES", "bell-window-system").split(",")
 def force_enabled(codec_name):
     return os.environ.get("XPRA_SOUND_CODEC_ENABLE_%s" % codec_name.upper().replace("+", "_"), "0")=="1"
 
@@ -62,10 +64,8 @@ NAME_TO_SRC_PLUGIN = {
     "direct"        : "directsoundsrc",
     "wasapi"        : "wasapisrc",
     }
+SRC_TO_NAME_PLUGIN = reverse_dict(NAME_TO_SRC_PLUGIN)
 SRC_HAS_DEVICE_NAME = ["alsasrc", "osssrc", "oss4src", "jackaudiosrc", "pulsesrc", "directsoundsrc", "osxaudiosrc"]
-SRC_TO_NAME_PLUGIN = {}
-for k,v in NAME_TO_SRC_PLUGIN.items():
-    SRC_TO_NAME_PLUGIN[v] = k
 PLUGIN_TO_DESCRIPTION = {
     "pulsesrc"      : "Pulseaudio",
     "jacksrc"       : "JACK Audio Connection Kit",
@@ -89,6 +89,7 @@ NAME_TO_INFO_PLUGIN = {
 #we keep multiple options here for the same encoding
 #and will populate the ones that are actually available into the "CODECS" dict
 CODEC_OPTIONS = [
+        (VORBIS_MKA , "vorbisenc",      "matroskamux",  "vorbisdec",                    "matroskademux"),
         (VORBIS_MKA , "vorbisenc",      "webmmux",      "vorbisdec",                    "matroskademux"),
         #those two used to fail silently (older versions of gstreamer?)
         (VORBIS_OGG , "vorbisenc",      "oggmux",       "vorbisparse ! vorbisdec",      "oggdemux"),
@@ -96,14 +97,16 @@ CODEC_OPTIONS = [
         (FLAC       , "flacenc",        None,           "flacparse ! flacdec",          None),
         (FLAC_OGG   , "flacenc",        "oggmux",       "flacparse ! flacdec",          "oggdemux"),
         (MP3        , "lamemp3enc",     None,           "mpegaudioparse ! mad",         None),
+        (MP3_ID3V2  , "lamemp3enc",     "id3v2mux",     "mpegaudioparse ! mpg123audiodec", "id3demux"),
         (MP3        , "lamemp3enc",     None,           "mpegaudioparse ! mpg123audiodec", None),
         (MP3_MPEG4  , "lamemp3enc",     "mp4mux",       "mpegaudioparse ! mad",         "qtdemux"),
         (WAV        , "wavenc",         None,           "wavparse",                     None),
-        (WAV_LZ4    , "wavenc",         None,           "wavparse",                     None,                       "lz4"),
-        (WAV_LZO    , "wavenc",         None,           "wavparse",                     None,                       "lzo"),
+        (WAV_LZ4    , "wavenc",         None,           "wavparse",                     None,               "lz4"),
+        (WAV_LZO    , "wavenc",         None,           "wavparse",                     None,               "lzo"),
         (OPUS_OGG   , "opusenc",        "oggmux",       "opusdec",                      "oggdemux"),
         (OPUS       , "opusenc",        None,           "opusparse ! opusdec",          None),
         #this can cause "could not link opusenc0 to webmmux0"
+        (OPUS_MKA   , "opusenc",        "matroskamux",  "opusdec",                      "matroskademux"),
         (OPUS_MKA   , "opusenc",        "webmmux",      "opusdec",                      "matroskademux"),
         (SPEEX_OGG  , "speexenc",       "oggmux",       "speexdec",                     "oggdemux"),
         (WAVPACK    , "wavpackenc",      None,          "wavpackparse ! wavpackdec",    None),
@@ -114,6 +117,7 @@ CODEC_OPTIONS = [
 MUX_OPTIONS = [
                (OGG,    "oggmux",   "oggdemux"),
                (MKA,    "webmmux",  "matroskademux"),
+               (MKA,    "matroskamux",  "matroskademux"),
                (MPEG4,  "mp4mux",   "qtdemux"),
               ]
 emux = [x for x in os.environ.get("XPRA_MUXER_OPTIONS", "").split(",") if len(x.strip())>0]
@@ -130,7 +134,7 @@ del emux
 #these encoders require an "audioconvert" element:
 ENCODER_NEEDS_AUDIOCONVERT = ("flacenc", "wavpackenc")
 #if this is lightweight enough, maybe we should include it unconditionally?
-SOURCE_NEEDS_AUDIOCONVERT = ("directsoundsrc", "osxaudiosrc", "autoaudiosrc")
+SOURCE_NEEDS_AUDIOCONVERT = ("directsoundsrc", "osxaudiosrc", "autoaudiosrc", "wasapisrc")
 
 CUTTER_NEEDS_RESAMPLE = ("opusenc", )
 #those don't work anyway:
@@ -152,7 +156,7 @@ ENCODER_DEFAULT_OPTIONS_COMMON = {
                                "quality"    : 0,        #"fast"
                                },
             "avenc_aac"     : {
-                               "compliance" : 1,       #allows experimental
+                               #"compliance" : 1,
                                "perfect-timestamp"  : 1,
                                },
             "faac"          : {
@@ -161,13 +165,12 @@ ENCODER_DEFAULT_OPTIONS_COMMON = {
             #"vorbisenc"     : {"perfect-timestamp" : 1},
                            }
 ENCODER_DEFAULT_OPTIONS = {
-                                #FIXME: figure out when it is safe to apply the "bitrate-type" setting:
-                                "opusenc"       : {
-                                                   #only available with 1.6 onwards?
-                                                   #"bitrate-type"   : 2,      #constrained vbr
-                                                   "complexity"     : 0
-                                                   },
-                           }
+    "opusenc"       : {
+        #only available with 1.6 onwards?
+        "bitrate-type"   : 1,      #vbr
+        "complexity"     : 0
+        },
+    }
 #we may want to review this if/when we implement UDP transport:
 MUXER_DEFAULT_OPTIONS = {
             "oggmux"        : {
@@ -179,6 +182,10 @@ MUXER_DEFAULT_OPTIONS = {
                                "streamable"         : 1,
                                #"min-index-interval" : 0,
                                },
+            "matroskamux"   : {
+                               "writing-app"        : "Xpra",
+                               "streamable"         : 1,
+                },
             "mp4mux"        : {
                                "faststart"          : 1,
                                "streamable"         : 1,
@@ -188,7 +195,7 @@ MUXER_DEFAULT_OPTIONS = {
            }
 
 #based on the encoder options above:
-RECORD_PIPELINE_LATENCY = 75
+RECORD_PIPELINE_LATENCY = 25
 ENCODER_LATENCY = {
         VORBIS      : 0,
         VORBIS_OGG  : 0,
@@ -201,7 +208,12 @@ ENCODER_LATENCY = {
         SPEEX       : 0,
        }
 
-CODEC_ORDER = [OPUS, OPUS_OGG, VORBIS_MKA, VORBIS_OGG, VORBIS, MP3, FLAC_OGG, AAC_MPEG4, WAV_LZ4, WAV_LZO, WAV, WAVPACK, SPEEX_OGG, VORBIS, OPUS_MKA, FLAC, MP3_MPEG4]
+CODEC_ORDER = [
+    OPUS, OPUS_OGG, VORBIS_MKA, VORBIS_OGG, VORBIS,
+    MP3, MP3_ID3V2, FLAC_OGG, AAC_MPEG4,
+    WAV_LZ4, WAV_LZO, WAV, WAVPACK,
+    SPEEX_OGG, VORBIS, OPUS_MKA, FLAC, MP3_MPEG4,
+    ]
 
 
 gst = None
@@ -216,45 +228,28 @@ def get_gst_version():
     return gst.version()
 
 
-def import_gst():
+def do_import_gst():
     global gst
     if gst is not None:
         return gst
 
     #hacks to locate gstreamer plugins on win32 and osx:
     if WIN32:
-        frozen = hasattr(sys, "frozen") and sys.frozen in ("windows_exe", "console_exe", True)
+        frozen = getattr(sys, "frozen", None) in ("windows_exe", "console_exe", True)
         log("gstreamer_util: frozen=%s", frozen)
         if frozen:
-            #on win32, we keep separate trees
-            #because GStreamer 0.10 and 1.x were built using different and / or incompatible version of the same libraries:
             from xpra.platform.paths import get_app_dir
-            gi_dir = os.path.join(get_app_dir(), "lib", "girepository-1.0")
             gst_dir = os.path.join(get_app_dir(), "lib", "gstreamer-1.0")   #ie: C:\Program Files\Xpra\lib\gstreamer-1.0
-            gst_bin_dir = os.path.join(get_app_dir(), "bin")                #ie: C:\Program Files\Xpra\bin
-            if not os.path.exists(gst_dir):
-                #fallback to old build locations:
-                gi_dir = os.path.join(get_app_dir(), "girepository-1.0")
-                gst_dir = os.path.join(get_app_dir(), "gstreamer-1.0")
-                gst_bin_dir = os.path.join(gst_dir, "bin")                  #ie: C:\Program Files\Xpra\gstreamer-0.10\bin
-            os.environ["GI_TYPELIB_PATH"] = gi_dir
             os.environ["GST_PLUGIN_PATH"] = gst_dir
-            os.environ["PATH"] = os.pathsep.join(x for x in (gst_bin_dir, os.environ.get("PATH", "")) if x)
-            sys.path.insert(0, gst_bin_dir)
-            scanner = os.path.join(gst_bin_dir, "gst-plugin-scanner.exe")
-            if os.path.exists(scanner):
-                os.environ["GST_PLUGIN_SCANNER"]    = scanner
     elif OSX:
         bundle_contents = os.environ.get("GST_BUNDLE_CONTENTS")
         log("OSX: GST_BUNDLE_CONTENTS=%s", bundle_contents)
         if bundle_contents:
-            os.environ["GST_PLUGIN_PATH"]       = os.path.join(bundle_contents, "Resources", "lib", "gstreamer-1.0")
-            os.environ["GST_PLUGIN_SCANNER"]    = os.path.join(bundle_contents, "Resources", "bin", "gst-plugin-scanner-1.0")
-            #typelib path should have been set in PythonExecWrapper
-            if not os.environ.get("GI_TYPELIB_PATH"):
-                gi_dir = os.path.join(bundle_contents, "Resources", "lib", "girepository-1.0")
-                os.environ["GI_TYPELIB_PATH"]       = gi_dir
-    log("GStreamer 1.x environment: %s", dict((k,v) for k,v in os.environ.items() if (k.startswith("GST") or k.startswith("GI") or k=="PATH")))
+            rsc_dir = os.path.join(bundle_contents, "Resources")
+            os.environ["GST_PLUGIN_PATH"]       = os.path.join(rsc_dir, "lib", "gstreamer-1.0")
+            os.environ["GST_PLUGIN_SCANNER"]    = os.path.join(rsc_dir, "bin", "gst-plugin-scanner-1.0")
+    log("GStreamer 1.x environment: %s",
+        dict((k,v) for k,v in os.environ.items() if (k.startswith("GST") or k.startswith("GI") or k=="PATH")))
     log("GStreamer 1.x sys.path=%s", csv(sys.path))
 
     try:
@@ -271,9 +266,11 @@ def import_gst():
         log.warn(" %s", e)
         return None
     return gst
+import_gst = do_import_gst
 
 def prevent_import():
-    global gst, import_gst
+    global gst
+    global import_gst
     if gst or "gst" in sys.modules or "gi.repository.Gst" in sys.modules:
         raise Exception("cannot prevent the import of the GStreamer bindings, already loaded: %s" % gst)
     def fail_import():
@@ -292,7 +289,7 @@ def normv(v):
 all_plugin_names = []
 def get_all_plugin_names():
     global all_plugin_names, gst
-    if len(all_plugin_names)==0 and gst:
+    if not all_plugin_names and gst:
         registry = gst.Registry.get()
         all_plugin_names = [el.get_name() for el in registry.get_feature_list(gst.ElementFactory)]
         all_plugin_names.sort()
@@ -308,7 +305,7 @@ def has_plugins(*names):
             continue
         snames += [v.strip() for v in x.split("!")]
     missing = [name for name in snames if (name is not None and name not in allp)]
-    if len(missing)>0:
+    if missing:
         log("missing %s from %s", missing, names)
     return len(missing)==0
 
@@ -355,8 +352,8 @@ def init_codecs():
     log("initialized sound codecs:")
     def ci(v):
         return "%-22s" % (v or "")
-    log("  - %s", "".join([ci(v) for v in ("encoder/decoder", "(de)payloader", "stream-compressor")]))
-    for k in [x for x in CODEC_ORDER]:
+    log("  - %s", "".join(ci(v) for v in ("encoder/decoder", "(de)payloader", "stream-compressor")))
+    for k in CODEC_ORDER:
         if k in ENCODERS or k in DECODERS:
             CODECS[k] = True
             log("* %s :", k)
@@ -390,10 +387,10 @@ def validate_encoding(elements):
     if force_enabled(encoding):
         log.info("sound codec %s force enabled", encoding)
         return True
-    elif encoding in (VORBIS_OGG, VORBIS) and get_gst_version()<(1, 12):
+    if encoding in (VORBIS_OGG, VORBIS) and get_gst_version()<(1, 12):
         log("skipping %s - not sure which GStreamer versions support it", encoding)
         return False
-    elif encoding.startswith(OPUS):
+    if encoding.startswith(OPUS):
         if encoding==OPUS_MKA and get_gst_version()<(1, 8):
             #this causes "could not link opusenc0 to webmmux0"
             #(not sure which versions are affected, but 1.8.x is not)
@@ -401,7 +398,7 @@ def validate_encoding(elements):
             return False
     try:
         stream_compressor = elements[5]
-    except:
+    except IndexError:
         stream_compressor = None
     if stream_compressor and not has_stream_compressor(stream_compressor):
         log("skipping %s: missing %s", encoding, stream_compressor)
@@ -412,10 +409,10 @@ def has_stream_compressor(stream_compressor):
     if stream_compressor not in ("lz4", "lzo"):
         log.warn("Warning: invalid stream compressor '%s'", stream_compressor)
         return False
-    from xpra.net.compression import use_lz4, use_lzo
-    if stream_compressor=="lz4" and not use_lz4:
+    from xpra.net.compression import use
+    if stream_compressor=="lz4" and not use("lz4"):
         return False
-    if stream_compressor=="lzo" and not use_lzo:
+    if stream_compressor=="lzo" and not use("lzo"):
         return False
     return True
 
@@ -434,13 +431,14 @@ def get_demuxers():
     return demuxers
 
 def get_stream_compressors():
-    return [x for x in ["lz4", "lzo"] if has_stream_compressor(x)]
+    return [x for x in ("lz4", "lzo") if has_stream_compressor(x)]
 
 def get_encoder_elements(name):
     encoders = get_encoders()
     assert name in encoders, "invalid codec: %s (should be one of: %s)" % (name, encoders.keys())
     encoder, formatter, stream_compressor = encoders.get(name)
-    assert stream_compressor is None or has_stream_compressor(stream_compressor), "stream-compressor %s not found" % stream_compressor
+    if stream_compressor:
+        assert has_stream_compressor(stream_compressor), "stream-compressor %s not found" % stream_compressor
     assert encoder is None or has_plugins(encoder), "encoder %s not found" % encoder
     assert formatter is None or has_plugins(formatter), "formatter %s not found" % formatter
     return encoder, formatter, stream_compressor
@@ -449,7 +447,8 @@ def get_decoder_elements(name):
     decoders = get_decoders()
     assert name in decoders, "invalid codec: %s (should be one of: %s)" % (name, decoders.keys())
     decoder, parser, stream_compressor = decoders.get(name)
-    assert stream_compressor is None or has_stream_compressor(stream_compressor), "stream-compressor %s not found" % stream_compressor
+    if stream_compressor:
+        assert has_stream_compressor(stream_compressor), "stream-compressor %s not found" % stream_compressor
     assert decoder is None or has_plugins(decoder), "decoder %s not found" % decoder
     assert parser is None or has_plugins(parser), "parser %s not found" % parser
     return decoder, parser, stream_compressor
@@ -483,7 +482,7 @@ def plugin_str(plugin, options):
     s = "%s" % plugin
     def qstr(v):
         #only quote strings
-        if type(v) in (str, unicode):
+        if isinstance(v, str):
             return "\"%s\"" % v
         return v
     if options:
@@ -508,7 +507,7 @@ def get_source_plugins():
         sources.append("osxaudiosrc")
     elif WIN32:
         sources.append("directsoundsrc")
-        #sources.append("wasapisrc")
+        sources.append("wasapisrc")
     sources.append("autoaudiosrc")
     if POSIX:
         sources += ["alsasrc",
@@ -547,7 +546,7 @@ def get_sink_plugins():
         SINKS.append("osxaudiosink")
     elif WIN32:
         SINKS.append("directsoundsink")
-        #SINKS.append("wasapisink")
+        SINKS.append("wasapisink")
     SINKS.append("autoaudiosink")
     try:
         from xpra.sound.pulseaudio.pulseaudio_util import has_pa
@@ -559,7 +558,7 @@ def get_sink_plugins():
         SINKS += ["alsasink", "osssink", "oss4sink", "jackaudiosink"]
     return SINKS
 
-def get_default_sink():
+def get_default_sink_plugin():
     sink = os.environ.get("XPRA_SOUND_SINK")
     sinks = get_sink_plugins()
     if sink:
@@ -576,7 +575,7 @@ def get_default_sink():
             else:
                 return "pulsesink"
     except ImportError as e:
-        log("get_default_sink() no pulsesink: %s", e)
+        log("get_default_sink_plugin() no pulsesink: %s", e)
     for sink in sinks:
         if has_plugins(sink):
             return sink
@@ -587,11 +586,13 @@ def get_test_defaults(*_args):
     return  {"wave" : 2, "freq" : 110, "volume" : 0.4}
 
 WARNED_MULTIPLE_DEVICES = False
-def get_pulse_defaults(device_name_match=None, want_monitor_device=True, input_or_output=None, remote=None, env_device_name=None):
+def get_pulse_defaults(device_name_match=None, want_monitor_device=True,
+                       input_or_output=None, remote=None, env_device_name=None):
     try:
         device = get_pulse_device(device_name_match, want_monitor_device, input_or_output, remote, env_device_name)
     except Exception as e:
-        log("get_pulse_defaults%s", (device_name_match, want_monitor_device, input_or_output, remote, env_device_name), exc_info=True)
+        log("get_pulse_defaults%s",
+            (device_name_match, want_monitor_device, input_or_output, remote, env_device_name), exc_info=True)
         log.warn("Warning: failed to identify the pulseaudio default device to use")
         log.warn(" %s", e)
         return {}
@@ -609,13 +610,18 @@ def get_pulse_defaults(device_name_match=None, want_monitor_device=True, input_o
         log("device %s may still be muted: %s", device, e)
     return {"device" : bytestostr(device)}
 
-def get_pulse_device(device_name_match=None, want_monitor_device=True, input_or_output=None, remote=None, env_device_name=None):
+def get_pulse_device(device_name_match=None, want_monitor_device=True,
+                     input_or_output=None, remote=None, env_device_name=None):
     """
         choose the device to use
     """
     log("get_pulse_device%s", (device_name_match, want_monitor_device, input_or_output, remote, env_device_name))
     try:
-        from xpra.sound.pulseaudio.pulseaudio_util import has_pa, get_pa_device_options, get_default_sink, get_pactl_server, get_pulse_id
+        from xpra.sound.pulseaudio.pulseaudio_util import (
+            has_pa, get_pa_device_options,
+            get_default_sink, get_pactl_server,
+            get_pulse_id,
+            )
         if not has_pa():
             log.warn("Warning: pulseaudio is not available!")
             return None
@@ -643,7 +649,7 @@ def get_pulse_device(device_name_match=None, want_monitor_device=True, input_or_
 
     device_type_str = ""
     if input_or_output is not None:
-        device_type_str = ["output", "input"][input_or_output]
+        device_type_str = "input" if input_or_output else "output"
     if want_monitor_device:
         device_type_str += " monitor"
     #def get_pa_device_options(monitors=False, input_or_output=None, ignored_devices=["bell-window-system"])
@@ -666,7 +672,7 @@ def get_pulse_device(device_name_match=None, want_monitor_device=True, input_or_
                 filtered[k] = v
         devices = filtered
 
-    if len(devices)==0:
+    if not devices:
         log.error("Error: sound forwarding is disabled")
         log.error(" could not detect any Pulseaudio %s devices", device_type_str)
         return None
@@ -685,7 +691,9 @@ def get_pulse_device(device_name_match=None, want_monitor_device=True, input_or_
                 filters.append(match)
             match = match.lower()
             log("trying to match '%s' in devices=%s", match, devices)
-            matches = dict((k,v) for k,v in devices.items() if (bytestostr(k).strip().lower().find(match)>=0 or bytestostr(v).strip().lower().find(match)>=0))
+            matches = dict((k,v) for k,v in devices.items()
+                           if (bytestostr(k).strip().lower().find(match)>=0 or
+                               bytestostr(v).strip().lower().find(match)>=0))
             #log("matches(%s, %s)=%s", devices, match, matches)
             if len(matches)==1:
                 log("found name match for '%s': %s", match, tuple(matches.items())[0])
@@ -693,8 +701,8 @@ def get_pulse_device(device_name_match=None, want_monitor_device=True, input_or_
             elif len(matches)>1:
                 log.warn("Warning: Pulseaudio %s device name filter '%s'", device_type_str, match)
                 log.warn(" matched %i devices", len(matches))
-        if filters or len(matches)>0:
-            if len(matches)==0:
+        if filters or matches:
+            if not matches:
                 log.warn("Warning: Pulseaudio %s device name filter%s:", device_type_str, engs(filters))
                 log.warn(" %s", csv("'%s'" % x for x in filters))
                 log.warn(" did not match any of the devices found:")
@@ -722,16 +730,16 @@ def get_pulse_device(device_name_match=None, want_monitor_device=True, input_or_
             if want_monitor_device:
                 dtype = "output monitor"
             elif input_or_output is False:
-                dtype = "audio input"
-            elif input_or_output is True:
                 dtype = "audio output"
-            log.warn("Warning: found %i %s devices:", len(devices), dtype)
+            elif input_or_output is True:
+                dtype = "audio input"
+            log.info("found %i %s devices:", len(devices), dtype)
             for k,v in devices.items():
-                log.warn(" * %s", bytestostr(v))
-                log.warn("   %s", bytestostr(k))
+                log.info(" * %s", bytestostr(v))
+                log.info("   %s", bytestostr(k))
             if not env_device: #used already!
-                log.warn(" to select a specific one,")
-                log.warn(" use the environment variable '%s'", env_device_name)
+                log.info(" to select a specific one,")
+                log.info(" use the environment variable '%s'", env_device_name)
         #default to first one:
         if USE_DEFAULT_DEVICE:
             log.info("using default pulseaudio device")
@@ -743,10 +751,13 @@ def get_pulse_device(device_name_match=None, want_monitor_device=True, input_or_
     return device
 
 def get_pulse_source_defaults(device_name_match=None, want_monitor_device=True, remote=None):
-    return get_pulse_defaults(device_name_match, want_monitor_device, input_or_output=not want_monitor_device, remote=remote, env_device_name=XPRA_PULSE_SOURCE_DEVICE_NAME)
+    return get_pulse_defaults(device_name_match, want_monitor_device,
+                              input_or_output=not want_monitor_device, remote=remote,
+                              env_device_name=XPRA_PULSE_SOURCE_DEVICE_NAME)
 
 def get_pulse_sink_defaults():
-    return get_pulse_defaults(want_monitor_device=False, input_or_output=False, env_device_name=XPRA_PULSE_SINK_DEVICE_NAME)
+    return get_pulse_defaults(want_monitor_device=False, input_or_output=False,
+                              env_device_name=XPRA_PULSE_SINK_DEVICE_NAME)
 
 def get_directsound_source_defaults(device_name_match=None, want_monitor_device=True, remote=None):
     try:
@@ -761,7 +772,7 @@ def get_directsound_source_defaults(device_name_match=None, want_monitor_device=
         if devices:
             for guid, name in devices:
                 if guid:
-                    log("* %-32s %s : %s", name, guid)
+                    log("* %-32s %s", name, guid)
                 else:
                     log("* %s", name)
                 names.append(name)
@@ -849,13 +860,15 @@ def parse_sound_source(all_plugins, sound_source_plugin, device, want_monitor_de
     options_str = (sound_source_plugin+":").split(":",1)[1].rstrip(":")
     simple_str = (plugin).lower().strip()
     if not simple_str:
-        #choose the first one from
-        options = [x for x in get_source_plugins() if x in all_plugins]
-        if not options:
-            log.error("no source plugins available")
-            return None
-        log("parse_sound_source: no plugin specified, using default: %s", options[0])
-        simple_str = options[0]
+        simple_str = get_default_source()
+        if not simple_str:
+            #choose the first one from
+            options = [x for x in get_source_plugins() if x in all_plugins]
+            if not options:
+                log.error("no source plugins available")
+                return None, {}
+            log("parse_sound_source: no plugin specified, using default: %s", options[0])
+            simple_str = options[0]
     for s in ("src", "sound", "audio"):
         if simple_str.endswith(s):
             simple_str = simple_str[:-len(s)]
@@ -893,7 +906,9 @@ def main():
         if v[-1]==0:
             v = v[:-1]
         gst_vinfo = ".".join((str(x) for x in v))
-        print("Loaded Python GStreamer version %s for Python %s.%s" % (gst_vinfo, sys.version_info[0], sys.version_info[1]))
+        print("Loaded Python GStreamer version %s for Python %s.%s" % (
+            gst_vinfo, sys.version_info[0], sys.version_info[1])
+        )
         apn = get_all_plugin_names()
         print("GStreamer plugins found: %s" % csv(apn))
         print("")
@@ -909,7 +924,7 @@ def main():
         print("stream compressors: %s" % csv(get_stream_compressors()))
         print("source plugins:     %s" % csv([x for x in get_source_plugins() if x in apn]))
         print("sink plugins:       %s" % csv([x for x in get_sink_plugins() if x in apn]))
-        print("default sink:       %s" % get_default_sink())
+        print("default sink:       %s" % get_default_sink_plugin())
 
 
 if __name__ == "__main__":
