@@ -10,3 +10,7 @@ gi.require_version('GdkX11', '3.0')
 
 from gi.repository import GdkX11  #pylint: disable=wrong-import-position
 assert GdkX11
+
+def GDKX11Window(*args, **kwargs) -> GdkX11.X11Window:
+    from xpra.gtk_common.gtk_util import new_GDKWindow
+    return new_GDKWindow(GdkX11.X11Window, *args, **kwargs)
