@@ -1,17 +1,18 @@
 # This file is part of Xpra.
 # Copyright (C) 2008, 2009 Nathaniel Smith <njs@pobox.com>
-# Copyright (C) 2011-2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2011-2021 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import os
 import signal
+from socket import gethostname
 from gi.repository import GObject, Gdk, GLib
 
 from xpra.util import envbool, first_time
 from xpra.os_util import bytestostr
 from xpra.x11.common import Unmanageable
-from xpra.gtk_common.gobject_util import one_arg_signal
+from xpra.gtk_common.gobject_util import one_arg_signal, two_arg_signal
 from xpra.gtk_common.error import XError, xsync, xswallow
 from xpra.x11.bindings.window_bindings import X11WindowBindings, constants, SHAPE_KIND #@UnresolvedImport
 from xpra.x11.models.model_stub import WindowModelStub
@@ -37,6 +38,19 @@ FORCE_QUIT = envbool("XPRA_FORCE_QUIT", True)
 XSHAPE = envbool("XPRA_XSHAPE", True)
 FRAME_EXTENTS = envbool("XPRA_FRAME_EXTENTS", True)
 
+# Re-stacking:
+Above = 0
+Below = 1
+TopIf = 2
+BottomIf = 3
+Opposite = 4
+RESTACKING_STR = {
+    Above : "Above",
+    Below : "Below",
+    TopIf : "TopIf",
+    BottomIf : "BottomIf",
+    Opposite : "Opposite",
+    }
 
 # grab stuff:
 NotifyNormal        = constants["NotifyNormal"]
@@ -156,7 +170,7 @@ class CoreX11WindowModel(WindowModelStub):
     __common_signals__ = {
         #signals we emit:
         "unmanaged"                     : one_arg_signal,
-        "raised"                        : one_arg_signal,
+        "restack"                       : two_arg_signal,
         "initiate-moveresize"           : one_arg_signal,
         "grab"                          : one_arg_signal,
         "ungrab"                        : one_arg_signal,
@@ -212,7 +226,7 @@ class CoreX11WindowModel(WindowModelStub):
         self._internal_set_property("client-window", client_window)
 
 
-    def __repr__(self):
+    def __repr__(self):  #pylint: disable=arguments-differ
         try:
             return "%s(%#x)" % (type(self).__name__, self.xid)
         except AttributeError:
@@ -242,7 +256,8 @@ class CoreX11WindowModel(WindowModelStub):
                 self._internal_set_property("geometry", geom[:4])
                 self._read_initial_X11_properties()
         except XError as e:
-            raise Unmanageable(e)
+            log("failed to manage %#x", self.xid, exc_info=True)
+            raise Unmanageable(e) from e
         add_event_receiver(self.client_window, self)
         # Keith Packard says that composite state is undefined following a
         # reparent, so I'm not sure doing this here in the superclass,
@@ -261,7 +276,7 @@ class CoreX11WindowModel(WindowModelStub):
             self._composite = None
             if isinstance(e, Unmanageable):
                 raise
-            raise Unmanageable(e)
+            raise Unmanageable(e) from e
         #compositing is now enabled,
         #from now on we must call setup_failed to clean things up
         self._managed = True
@@ -269,12 +284,13 @@ class CoreX11WindowModel(WindowModelStub):
             with xsync:
                 self.setup()
         except XError as e:
+            log("failed to setup %#x", self.xid, exc_info=True)
             try:
                 with xsync:
                     self.setup_failed(e)
             except Exception as ex:
                 log.error("error in cleanup handler: %s", ex)
-            raise Unmanageable(e)
+            raise Unmanageable(e) from None
         self._setup_done = True
 
     def setup_failed(self, e):
@@ -377,6 +393,7 @@ class CoreX11WindowModel(WindowModelStub):
                 try:
                     handler(self)
                 except XError:
+                    log("handler %s failed", handler, exc_info=True)
                     #these will be caught in call_setup()
                     raise
                 except Exception:
@@ -416,7 +433,7 @@ class CoreX11WindowModel(WindowModelStub):
              "Bounding.extents" : bextents,
              "Clip.extents"     : cextents,
              }
-        for kind, kind_name in SHAPE_KIND.items():
+        for kind, kind_name in SHAPE_KIND.items():  # @UndefinedVariable
             rectangles = X11Window.XShapeGetRectangles(self.xid, kind)
             v[kind_name+".rectangles"] = rectangles
         shapelog("_read_shape()=%s", v)
@@ -648,16 +665,13 @@ class CoreX11WindowModel(WindowModelStub):
 
 
     def do_xpra_shape_event(self, event):
-        shapelog("shape event: %s, kind=%s", event, SHAPE_KIND.get(event.kind, event.kind))
+        shapelog("shape event: %s, kind=%s", event, SHAPE_KIND.get(event.kind, event.kind))  # @UndefinedVariable
         cur_shape = self.get_property("shape")
         if cur_shape and cur_shape.get("serial", 0)>=event.serial:
             shapelog("same or older xshape serial no: %#x (current=%#x)", event.serial, cur_shape.get("serial", 0))
             return
         #remove serial before comparing dicts:
-        try:
-            del cur_shape["serial"]
-        except KeyError:
-            pass
+        cur_shape.pop("serial", None)
         #read new xshape:
         with xswallow:
             #should we pass the x and y offsets here?
@@ -698,7 +712,7 @@ class CoreX11WindowModel(WindowModelStub):
         grablog("focus_in_event(%s) mode=%s, detail=%s",
             event, GRAB_CONSTANTS.get(event.mode), DETAIL_CONSTANTS.get(event.detail, event.detail))
         if event.mode==NotifyNormal and event.detail==NotifyNonlinearVirtual:
-            self.emit("raised", event)
+            self.emit("restack", Above, None)
         else:
             self.may_emit_grab(event)
 
@@ -738,7 +752,7 @@ class CoreX11WindowModel(WindowModelStub):
     ################################
 
     def request_close(self):
-        if b"WM_DELETE_WINDOW" in self.get_property("protocols"):
+        if "WM_DELETE_WINDOW" in self.get_property("protocols"):
             with xswallow:
                 send_wm_delete_window(self.client_window)
         else:
@@ -757,7 +771,6 @@ class CoreX11WindowModel(WindowModelStub):
     def force_quit(self):
         pid = self.get_property("pid")
         machine = self.get_property("client-machine")
-        from socket import gethostname
         localhost = gethostname()
         log("force_quit() pid=%s, machine=%s, localhost=%s", pid, machine, localhost)
         def XKill():
@@ -765,20 +778,22 @@ class CoreX11WindowModel(WindowModelStub):
                 X11Window.XKillClient(self.xid)
         if pid > 0 and machine is not None and machine == localhost:
             if pid==os.getpid():
-                log.warn("force_quit() refusing to kill ourselves!")
+                log.warn("Warning: force_quit is refusing to kill ourselves!")
                 return
             if self._kill_count==0:
                 #first time around: just send a SIGINT and hope for the best
                 try:
                     os.kill(pid, signal.SIGINT)
-                except OSError:
-                    log.warn("failed to kill(SIGINT) client with pid %s", pid)
+                except OSError as e:
+                    log.warn("Warning: failed to kill(SIGINT) client with pid %s", pid)
+                    log.warn(" %s", e)
             else:
                 #the more brutal way: SIGKILL + XKill
                 try:
                     os.kill(pid, signal.SIGKILL)
-                except OSError:
-                    log.warn("failed to kill(SIGKILL) client with pid %s", pid)
+                except OSError as e:
+                    log.warn("Warning: failed to kill(SIGKILL) client with pid %s", pid)
+                    log.warn(" %s", e)
                 XKill()
             self._kill_count += 1
             return
