@@ -1,6 +1,6 @@
 # This file is part of Xpra.
 # Copyright (C) 2008, 2009 Nathaniel Smith <njs@pobox.com>
-# Copyright (C) 2012-2017 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2012-2019 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -11,19 +11,14 @@ Functions for converting to and from X11 properties.
 """
 
 import struct
+from io import BytesIO
 
-from xpra.os_util import hexstr, BytesIOClass, PYTHON3
+from xpra.os_util import hexstr
 from xpra.x11.bindings.window_bindings import constants     #@UnresolvedImport
-from xpra.x11.bindings.window_bindings import X11WindowBindings #@UnresolvedImport
 from xpra.log import Logger
 
 log = Logger("x11", "window")
 
-X11Window = X11WindowBindings()
-
-if PYTHON3:
-    long = int              #@ReservedAssignment
-    unicode = str           #@ReservedAssignment
 
 USPosition      = constants["USPosition"]
 PPosition       = constants["PPosition"]
@@ -50,26 +45,27 @@ def _force_length(name, data, length, noerror_length=None):
         log.warn("Odd-lengthed property %s: wanted %s bytes, got %s: %r"
                  % (name, length, len(data), data))
     # Zero-pad data
-    data += "\0" * length
+    data += b"\0" * length
     return data[:length]
 
 
-class NetWMStrut(object):
+class NetWMStrut:
     def __init__(self, _disp, data):
         # This eats both _NET_WM_STRUT and _NET_WM_STRUT_PARTIAL.  If we are
         # given a _NET_WM_STRUT instead of a _NET_WM_STRUT_PARTIAL, then it
         # will be only length 4 instead of 12, we just don't define the other values
         # and let the client deal with it appropriately
         if len(data)==16:
-            self.left, self.right, self.top, self.bottom = struct.unpack(b"=IIII", data)
+            self.left, self.right, self.top, self.bottom = struct.unpack(b"@LLLL", data)
         else:
             data = _force_length("_NET_WM_STRUT or _NET_WM_STRUT_PARTIAL", data, 4 * 12)
-            (self.left, self.right, self.top, self.bottom,
-             self.left_start_y, self.left_end_y,
-             self.right_start_y, self.right_end_y,
-             self.top_start_x, self.top_end_x,
-             self.bottom_start_x, self.bottom_stop_x,
-             ) = struct.unpack(b"=" + "I" * 12, data)
+            (
+                self.left, self.right, self.top, self.bottom,
+                self.left_start_y, self.left_end_y,
+                self.right_start_y, self.right_end_y,
+                self.top_start_x, self.top_end_x,
+                self.bottom_start_x, self.bottom_stop_x,
+                ) = struct.unpack(b"@" + b"L" * 12, data)
 
     def todict(self):
         return self.__dict__
@@ -78,12 +74,13 @@ class NetWMStrut(object):
         return "NetWMStrut(%s)" % self.todict()
 
 
-class MotifWMHints(object):
+class MotifWMHints:
     def __init__(self, _disp, data):
         #some applications use the wrong size (ie: blender uses 16) so pad it:
-        pdata = _force_length("_MOTIF_WM_HINTS", data, 20, 16)
+        sizeof_long = struct.calcsize(b"@L")
+        pdata = _force_length("_MOTIF_WM_HINTS", data, sizeof_long*5, sizeof_long*4)
         self.flags, self.functions, self.decorations, self.input_mode, self.status = \
-            struct.unpack(b"=IIIiI", pdata)
+            struct.unpack(b"@LLLlL", pdata)
         log("MotifWMHints(%s)=%s", hexstr(data), self)
 
     #found in mwmh.h:
@@ -204,24 +201,33 @@ class MotifWMHints(object):
 
 def _read_image(_disp, stream):
     try:
-        header = stream.read(2 * 4)
+        int_size = struct.calcsize(b"@I")
+        long_size = struct.calcsize(b"@L")
+        header = stream.read(long_size*2)
         if not header:
             return None
-        (width, height) = struct.unpack(b"=II", header)
-        data = stream.read(width * height * 4)
-        if len(data) < width * height * 4:
-            log.warn("Corrupt _NET_WM_ICON")
+        width, height = struct.unpack(b"@LL", header)
+        data = stream.read(width * height * long_size)
+        expected = width * height * long_size
+        if len(data) < expected:
+            log.warn("Warning: corrupt _NET_WM_ICON, execpted %i bytes but got %i",
+                     expected, len(data))
             return None
-    except Exception as e:
-        log.warn("Weird corruption in _NET_WM_ICON: %s", e)
+        if int_size!=long_size:
+            #long to ints (CARD32):
+            longs = struct.unpack(b"@"+b"l"*(width*height), data)
+            data = struct.pack(b"@"+b"i"*(width*height), *longs)
+    except Exception:
+        log.warn("Weird corruption in _NET_WM_ICON", exc_info=True)
         return None
     return width, height, "BGRA", data
 
-# This returns a cairo ImageSurface which contains the largest icon defined in
-# a _NET_WM_ICON property.
+# This returns a list of icons from a _NET_WM_ICON property.
+# each icon is a tuple:
+# (width, height, fmt, data)
 def NetWMIcons(disp, data):
     icons = []
-    stream = BytesIOClass(data)
+    stream = BytesIO(data)
     while True:
         icon = _read_image(disp, stream)
         if icon is None:
@@ -245,27 +251,24 @@ def _from_utf8(_disp, v):
     return v.decode("UTF-8")
 
 
+def _from_long(_disp, v):
+    return struct.unpack(b"@L", v)[0]
+
+def _to_long(_disp, v):
+    return struct.pack(b"@L", v)
+
 
 PROP_TYPES = {
     # Python type, X type Atom, formatbits, serializer, deserializer, list
     # terminator
-    "utf8": (unicode, "UTF8_STRING", 8, _to_utf8, _from_utf8, b"\0"),
+    "utf8": (str, "UTF8_STRING", 8, _to_utf8, _from_utf8, b"\0"),
     # In theory, there should be something clever about COMPOUND_TEXT here.  I
     # am not sufficiently clever to deal with COMPOUNT_TEXT.  Even knowing
     # that Xutf8TextPropertyToTextList exists.
-    "latin1": (unicode, "STRING", 8, _to_latin1, _from_latin1, b"\0"),
-    "state": ((int, long), "WM_STATE", 32,
-            lambda _disp, c: struct.pack(b"=I", c),
-            lambda _disp, d: struct.unpack(b"=I", d)[0],
-            b""),
-    "u32": ((int, long), "CARDINAL", 32,
-            lambda _disp, c: struct.pack(b"=I", c),
-            lambda _disp, d: struct.unpack(b"=I", d)[0],
-            b""),
-    "integer": ((int, long), "INTEGER", 32,
-            lambda _disp, c: struct.pack(b"=I", c),
-            lambda _disp, d: struct.unpack(b"=I", d)[0],
-            b""),
+    "latin1": (str, "STRING", 8, _to_latin1, _from_latin1, b"\0"),
+    "state": (int, "WM_STATE", 32, _to_long, _from_long, b""),
+    "u32": (int, "CARDINAL", 32, _to_long, _from_long, b""),
+    "integer": (int, "INTEGER", 32, _to_long, _from_long, b""),
     "strut": (NetWMStrut, "CARDINAL", 32,
               unsupported, NetWMStrut, None),
     "strut-partial": (NetWMStrut, "CARDINAL", 32,
@@ -274,37 +277,34 @@ PROP_TYPES = {
               unsupported, MotifWMHints, None),
     "icons": (list, "CARDINAL", 32,
               unsupported, NetWMIcons, None),
-    # For uploading ad-hoc instances of the above complex structures to the
-    # server, so we can test reading them out again:
-    "debug-CARDINAL": (str, "CARDINAL", 32,
-                       lambda _disp, c: c,
-                       lambda _disp, d: d,
-                       None),
+    }
+
+PROP_SIZES = {
+    "icons" : 4*1024*1024,
     }
 
 
 def prop_encode(disp, etype, value):
-    if isinstance(etype, list):
+    if isinstance(etype, (list, tuple)):
         return _prop_encode_list(disp, etype[0], value)
     return _prop_encode_scalar(disp, etype, value)
 
 def _prop_encode_scalar(disp, etype, value):
-    (pytype, atom, formatbits, serialize, _, _) = PROP_TYPES[etype]
+    pytype, atom, formatbits, serialize = PROP_TYPES[etype][:4]
     assert isinstance(value, pytype), "value for atom %s is not a %s: %s" % (atom, pytype, type(value))
     return (atom, formatbits, serialize(disp, value))
 
 def _prop_encode_list(disp, etype, value):
     (_, atom, formatbits, _, _, terminator) = PROP_TYPES[etype]
     value = tuple(value)
-    serialized = [_prop_encode_scalar(disp, etype, v)[2] for v in value]
-    no_none = [x for x in serialized if x is not None]
+    serialized = tuple(_prop_encode_scalar(disp, etype, v)[2] for v in value)
     # Strings in X really are null-separated, not null-terminated (ICCCM
     # 2.7.1, see also note in 4.1.2.5)
-    return (atom, formatbits, terminator.join(no_none))
+    return (atom, formatbits, terminator.join(x for x in serialized if x is not None))
 
 
 def prop_decode(disp, etype, data):
-    if isinstance(etype, list):
+    if isinstance(etype, (list, tuple)):
         return _prop_decode_list(disp, etype[0], data)
     return _prop_decode_scalar(disp, etype, data)
 
@@ -320,10 +320,12 @@ def _prop_decode_list(disp, etype, data):
         datums = data.split(terminator)
     else:
         datums = []
-        nbytes = formatbits // 8
+        if formatbits==32:
+            nbytes = struct.calcsize("@L")
+        else:
+            nbytes = formatbits // 8
         while data:
             datums.append(data[:nbytes])
             data = data[nbytes:]
-    props = [_prop_decode_scalar(disp, etype, datum) for datum in datums]
-    #assert None not in props
+    props = (_prop_decode_scalar(disp, etype, datum) for datum in datums)
     return [x for x in props if x is not None]
