@@ -8,7 +8,7 @@
 
 from gi.repository import Gdk
 
-from xpra.util import csv, nonl, envbool, repr_ellipsized
+from xpra.util import csv, envbool
 from xpra.os_util import bytestostr
 from xpra.gtk_common.keymap import get_gtk_keymap
 from xpra.gtk_common.gtk_util import get_default_root_window
@@ -48,7 +48,6 @@ class KeyboardConfig(KeyboardConfigBase):
     def __init__(self):
         KeyboardConfigBase.__init__(self)
         self.xkbmap_raw = False
-        self.xkbmap_print = None
         self.xkbmap_query = None
         self.xkbmap_query_struct = None
         self.xkbmap_mod_meanings = {}
@@ -118,7 +117,7 @@ class KeyboardConfig(KeyboardConfigBase):
             for mod, mod_name in self.xkbmap_mod_meanings.items():
                 modinfo[mod] = mod_name
         info["x11_keycode"] = self.xkbmap_x11_keycodes
-        for x in ("print", "layout", "variant", "mod_managed", "mod_pointermissing", "raw", "layout_groups"):
+        for x in ("layout", "variant", "mod_managed", "mod_pointermissing", "raw", "layout_groups"):
             v = getattr(self, "xkbmap_%s" % x)
             if v:
                 info[x] = v
@@ -146,7 +145,7 @@ class KeyboardConfig(KeyboardConfigBase):
                 setattr(self, prop, nv)
                 modded[prop] = nv
         #plain strings:
-        for x in ("print", "query"):
+        for x in ("query", ):
             parse_option(x, props.strget)
         #lists:
         parse_option("keycodes", props.tupleget)
@@ -174,7 +173,7 @@ class KeyboardConfig(KeyboardConfigBase):
         def hashadd(v):
             m.update(("/%s" % str(v)).encode("utf8"))
         m.update(KeyboardConfigBase.get_hash(self))
-        for x in (self.xkbmap_print, self.xkbmap_query, self.xkbmap_raw, \
+        for x in (self.xkbmap_query, self.xkbmap_raw, \
                   self.xkbmap_mod_meanings, self.xkbmap_mod_pointermissing, \
                   self.xkbmap_keycodes, self.xkbmap_x11_keycodes):
             hashadd(x)
@@ -299,6 +298,9 @@ class KeyboardConfig(KeyboardConfigBase):
 
     def set_layout(self, layout, variant, options):
         log("set_layout(%s, %s, %s)", layout, variant, options)
+        if layout=="en":
+            log.warn("Warning: invalid keyboard layout name '%s', using 'us' instead", layout)
+            layout = "us"
         if layout!=self.xkbmap_layout or variant!=self.xkbmap_variant or options!=self.xkbmap_options:
             self.xkbmap_layout = layout
             self.xkbmap_variant = variant
@@ -310,9 +312,9 @@ class KeyboardConfig(KeyboardConfigBase):
     def set_keymap(self, translate_only=False):
         if not self.enabled:
             return
-        log("set_keymap(%s) layout=%s, variant=%s, options=%s, print=%s, query=%s",
+        log("set_keymap(%s) layout=%r, variant=%r, options=%r, query=%r",
             translate_only, self.xkbmap_layout, self.xkbmap_variant, self.xkbmap_options,
-            nonl(self.xkbmap_print), nonl(self.xkbmap_query))
+            self.xkbmap_query)
         if translate_only:
             self.keycode_translation = set_keycode_translation(self.xkbmap_x11_keycodes, self.xkbmap_keycodes)
             self.add_gtk_keynames()
@@ -325,8 +327,8 @@ class KeyboardConfig(KeyboardConfigBase):
         with xlog:
             clean_keyboard_state()
             do_set_keymap(self.xkbmap_layout, self.xkbmap_variant, self.xkbmap_options,
-                          self.xkbmap_print, self.xkbmap_query, self.xkbmap_query_struct)
-        log("set_keymap: xkbmap_print=%s, xkbmap_query=%s", nonl(self.xkbmap_print), nonl(self.xkbmap_query))
+                          self.xkbmap_query, self.xkbmap_query_struct)
+        log("set_keymap: xkbmap_query=%r", self.xkbmap_query)
         with xlog:
             #first clear all existing modifiers:
             clean_keyboard_state()
@@ -432,119 +434,132 @@ class KeyboardConfig(KeyboardConfigBase):
         self.keycode_mappings = get_keycode_mappings()
 
 
-    def do_get_keycode(self, client_keycode, keyname, pressed, modifiers, keystr, group):
+    def kmlog(self, keyname, msg, *args):
+        """ logs at info level is we are debugging this keyname """
+        if keyname in DEBUG_KEYSYMS:
+            l = log.info
+        else:
+            l = log
+        l(msg, *args)
+
+    def do_get_keycode(self, client_keycode, keyname, pressed, modifiers, keyval, keystr, group):
         if not self.enabled:
-            log("ignoring keycode since keyboard is turned off")
+            self.kmlog(keyname, "ignoring keycode since keyboard is turned off")
             return -1, group
         if keyname=="0xffffff":
             return -1, group
         if self.xkbmap_raw:
             return client_keycode, group
-        def kmlog(msg, *args):
-            if keyname in DEBUG_KEYSYMS:
-                l = log.info
-            else:
-                l = log
-            l(msg, *args)
-        def klog(msg, *args):
-            kmlog("do_get_keycode%s"+msg, (client_keycode, keyname, pressed, modifiers, group), *args)
-        keycode = None
-        rgroup = group
         if self.xkbmap_query:
             keycode = self.keycode_translation.get((client_keycode, keyname)) or client_keycode
-            klog("=%s (native keymap)", keycode)
-        else:
-            """
-            from man xmodmap:
-            The list of keysyms is assigned to the indicated keycode (which may be specified in decimal,
-            hex or octal and can be determined by running the xev program).
-            Up to eight keysyms may be attached to a key, however the last four are not used in any major
-            X server implementation.
-            The first keysym is used when no modifier key is pressed in conjunction with this key,
-            the second with Shift, the third when the Mode_switch key is used with this key and
-            the fourth when both the Mode_switch and Shift keys are used.
-            """
-            #non-native: try harder to find matching keysym
-            #first, try to honour shift state:
-            lock = ("lock" in modifiers) and (SHIFT_LOCK or (bool(keystr) and keystr.isalpha()))
-            shift = ("shift" in modifiers) ^ lock
-            mode = 0
-            numlock = 0
-            numlock_modifier = None
-            for mod, keynames in self.keynames_for_mod.items():
-                if "Num_Lock" in keynames:
-                    numlock_modifier = mod
-                    break
-            for mod in modifiers:
-                names = self.keynames_for_mod.get(mod, [])
-                if "Num_Lock" in names:
-                    numlock = 1
-                for name in names:
-                    if name in ("ISO_Level3_Shift", "Mode_switch"):
-                        mode = 1
-                        break
-            levels = []
-            #try to preserve the mode (harder to toggle):
-            for m in (int(bool(mode)), int(not mode)):
-                #try to preserve shift state:
-                for s in (int(bool(shift)), int(not shift)):
-                    #group is comparatively easier to toggle (one function call):
-                    for g in (int(bool(group)), int(not group)):
-                        level = int(g)*4 + int(m)*2 + int(s)*1
-                        levels.append(level)
-            kmlog("will try levels: %s", levels)
-            for level in levels:
-                keycode = self.keycode_translation.get((keyname, level))
-                if keycode:
-                    keysyms = self.keycode_mappings.get(keycode)
-                    klog("=%i (level=%i, shift=%s, mode=%i, keysyms=%s)", keycode, level, shift, mode, keysyms)
-                    if self.xkbmap_raw:
-                        break
-                    level0 = levels[0]
-                    uq_keysyms = set(keysyms)
-                    if len(uq_keysyms)<=1 or (len(keysyms)>level0 and keysyms[level0]==""):
-                        #if the keysym we would match for this keycode is 'NoSymbol',
-                        #then we can probably ignore it ('NoSymbol' shows up as "")
-                        #same if there's only one actual keysym for this keycode
-                        kmlog("not toggling any modifiers state for keysyms=%s", keysyms)
-                        break
-                    def toggle_modifier(mod):
-                        keynames = self.keynames_for_mod.get(mod)
-                        if keyname in keynames:
-                            kmlog("not toggling '%s' since '%s' should deal with it", mod, keyname)
-                            #the keycode we're returning is for this modifier,
-                            #assume that this will end up doing what is needed
-                            return
-                        if mod in modifiers:
-                            kmlog("removing '%s' from modifiers", mod)
-                            modifiers.remove(mod)
-                        else:
-                            kmlog("adding '%s' to modifiers", mod)
-                            modifiers.append(mod)
-                    #keypad overrules shift state (see #2702):
-                    if keyname.startswith("KP_"):
-                        if numlock_modifier and not numlock:
-                            toggle_modifier(numlock_modifier)
-                    elif (level & 1) ^ shift:
-                        #shift state does not match
-                        toggle_modifier("shift")
-                    if int(bool(level & 2)) ^ mode:
-                        #try to set / unset mode:
-                        for mod, keynames in self.keynames_for_mod.items():
-                            if "ISO_Level3_Shift" in keynames or "Mode_switch" in keynames:
-                                #found mode switch modified
-                                toggle_modifier(mod)
-                                break
-                    rgroup = level//4
-                    if rgroup!=group:
-                        kmlog("switching group from %i to %i", group, rgroup)
-                    break
-            #this should not find anything new?:
-            if keycode is None:
-                keycode = self.keycode_translation.get(keyname, -1)
-                klog("=%i, %i (keyname translation)", keycode, rgroup)
-        return keycode, rgroup
+            self.kmlog(keyname, "do_get_keycode (%i, %s)=%s (native keymap)", client_keycode, keyname, keycode)
+            return keycode, group
+        return self.find_matching_keycode(client_keycode, keyname, pressed, modifiers, keyval, keystr, group)
 
+    def find_matching_keycode(self, client_keycode, keyname, pressed, modifiers, keyval, keystr, group):
+        """
+        from man xmodmap:
+        The list of keysyms is assigned to the indicated keycode (which may be specified in decimal,
+        hex or octal and can be determined by running the xev program).
+        Up to eight keysyms may be attached to a key, however the last four are not used in any major
+        X server implementation.
+        The first keysym is used when no modifier key is pressed in conjunction with this key,
+        the second with Shift, the third when the Mode_switch key is used with this key and
+        the fourth when both the Mode_switch and Shift keys are used.
+        """
+        keycode = None
+        rgroup = group
+        def klog(msg, *args):
+            self.kmlog(keyname, "do_get_keycode%s"+msg, (client_keycode, keyname, pressed, modifiers, keyval, keystr, group), *args)
+        def kmlog(msg, *args):
+            self.kmlog(keyname, msg, *args)
+        #non-native: try harder to find matching keysym
+        #first, try to honour shift state:
+        lock = ("lock" in modifiers) and (SHIFT_LOCK or (bool(keystr) and keystr.isalpha()))
+        shift = ("shift" in modifiers) ^ lock
+        mode = 0
+        numlock = 0
+        numlock_modifier = None
+        for mod, keynames in self.keynames_for_mod.items():
+            if "Num_Lock" in keynames:
+                numlock_modifier = mod
+                break
+        for mod in modifiers:
+            names = self.keynames_for_mod.get(mod, [])
+            if "Num_Lock" in names:
+                numlock = 1
+            for name in names:
+                if name in ("ISO_Level3_Shift", "Mode_switch"):
+                    mode = 1
+                    break
+        levels = []
+        #try to preserve the mode (harder to toggle):
+        for m in (int(bool(mode)), int(not mode)):
+            #try to preserve shift state:
+            for s in (int(bool(shift)), int(not shift)):
+                #group is comparatively easier to toggle (one function call):
+                for g in (int(bool(group)), int(not group)):
+                    level = int(g)*4 + int(m)*2 + int(s)*1
+                    levels.append(level)
+        kmlog("will try levels: %s", levels)
+        for level in levels:
+            keycode = self.keycode_translation.get((keyname, level))
+            if keycode:
+                keysyms = self.keycode_mappings.get(keycode)
+                klog("=%i (level=%i, shift=%s, mode=%i, keysyms=%s)", keycode, level, shift, mode, keysyms)
+                level0 = levels[0]
+                uq_keysyms = set(keysyms)
+                if len(uq_keysyms)<=1 or (len(keysyms)>level0 and keysyms[level0]==""):
+                    #if the keysym we would match for this keycode is 'NoSymbol',
+                    #then we can probably ignore it ('NoSymbol' shows up as "")
+                    #same if there's only one actual keysym for this keycode
+                    kmlog("not toggling any modifiers state for keysyms=%s", keysyms)
+                    break
+                def toggle_modifier(mod):
+                    keynames = self.keynames_for_mod.get(mod, ())
+                    if keyname in keynames:
+                        kmlog("not toggling '%s' since '%s' should deal with it", mod, keyname)
+                        #the keycode we're returning is for this modifier,
+                        #assume that this will end up doing what is needed
+                        return
+                    if mod in modifiers:
+                        kmlog("removing '%s' from modifiers", mod)
+                        modifiers.remove(mod)
+                    else:
+                        kmlog("adding '%s' to modifiers", mod)
+                        modifiers.append(mod)
+                #keypad overrules shift state (see #2702):
+                if keyname.startswith("KP_"):
+                    if numlock_modifier and not numlock:
+                        toggle_modifier(numlock_modifier)
+                elif (level & 1) ^ shift:
+                    #shift state does not match
+                    toggle_modifier("shift")
+                if int(bool(level & 2)) ^ mode:
+                    #try to set / unset mode:
+                    for mod, keynames in self.keynames_for_mod.items():
+                        if "ISO_Level3_Shift" in keynames or "Mode_switch" in keynames:
+                            #found mode switch modified
+                            toggle_modifier(mod)
+                            break
+                rgroup = level//4
+                if rgroup!=group:
+                    kmlog("switching group from %i to %i", group, rgroup)
+                break
+        #this should not find anything new?:
+        if keycode is None:
+            keycode = self.keycode_translation.get(keyname, -1)
+            klog("=%i, %i (keyname translation)", keycode, rgroup)
+        if keycode is None:
+            #last resort, find using the keyval:
+            display = Gdk.Display.get_default()
+            keymap = Gdk.Keymap.get_for_display(display)
+            b, entries = keymap.get_entries_for_keyval(keyval)
+            if b and entries:
+                for entry in entries:
+                    if entry.group==group:
+                        return entry.keycode, entry.group
+        return keycode, rgroup
 
     def get_current_mask(self):
         current_mask = get_default_root_window().get_pointer()[-1]
@@ -604,11 +619,14 @@ class KeyboardConfig(KeyboardConfigBase):
             failed = []
             for modifier in modifiers:
                 modifier = bytestostr(modifier)
-                keynames = self.keynames_for_mod.get(modifier)
-                if not keynames:
+                keynames = self.keynames_for_mod.get(modifier, None)
+                if keynames is None:
                     log.error("Error: unknown modifier '%s'", modifier)
                     log.error(" known modifiers: %s", csv(self.keynames_for_mod.keys()))
                     continue
+                if not keynames:
+                    log.warn("Warning: found no keynames for '%s'", modifier)
+                    continue
                 #find the keycodes that match the keynames for this modifier
                 keycodes = []
                 #log.info("keynames(%s)=%s", modifier, keynames)
