@@ -1,5 +1,5 @@
 # This file is part of Xpra.
-# Copyright (C) 2010-2018 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -8,25 +8,48 @@
 #  http://lists.partiwm.org/pipermail/parti-discuss/2008-September/000041.html
 #  http://lists.partiwm.org/pipermail/parti-discuss/2008-September/000042.html
 # (also do not import anything that imports gtk)
-import subprocess
+import signal
+from subprocess import Popen, PIPE, call
 import os.path
 
 from xpra.scripts.config import InitException, get_Xdummy_confdir
 from xpra.util import envbool, envint
 from xpra.os_util import (
-    setsid, shellsub, close_fds,
+    shellsub,
     setuidgid, getuid, getgid,
-    strtobytes, osexpand, monotonic_time,
-    POSIX, OSX, PYTHON3,
+    strtobytes, bytestostr, osexpand, monotonic_time,
+    pollwait,
+    POSIX, OSX,
     )
 from xpra.platform.displayfd import read_displayfd, parse_displayfd
+from xpra.log import Logger
 
 
 VFB_WAIT = envint("XPRA_VFB_WAIT", 3)
-DEFAULT_VFB_RESOLUTION = tuple(int(x) for x in os.environ.get("XPRA_DEFAULT_VFB_RESOLUTION", "8192x4096").replace(",", "x").split("x", 1))
-assert len(DEFAULT_VFB_RESOLUTION)==2
-DEFAULT_DESKTOP_VFB_RESOLUTION = tuple(int(x) for x in os.environ.get("XPRA_DEFAULT_DESKTOP_VFB_RESOLUTION", "1280x1024").replace(",", "x").split("x", 1))
-assert len(DEFAULT_DESKTOP_VFB_RESOLUTION)==2
+
+RESOLUTION_ALIASES = {
+    "QVGA"  : (320, 240),
+    "VGA"   : (640, 480),
+    "SVGA"  : (800, 600),
+    "XGA"   : (1024, 768),
+    "1080P" : (1920, 1080),
+    "FHD"   : (1920, 1080),
+    "4K"    : (3840, 2160),
+    }
+
+def parse_resolution(s):
+    v = RESOLUTION_ALIASES.get(s.upper())
+    if v:
+        return v
+    res = tuple(int(x) for x in s.replace(",", "x").split("x", 1))
+    assert len(res)==2, "invalid resolution string '%s'" % s
+    return res
+def parse_env_resolution(envkey="XPRA_DEFAULT_VFB_RESOLUTION", default_res="8192x4096"):
+    s = os.environ.get(envkey, default_res)
+    return parse_resolution(s)
+
+DEFAULT_VFB_RESOLUTION = parse_env_resolution()
+DEFAULT_DESKTOP_VFB_RESOLUTION = parse_env_resolution("XPRA_DEFAULT_DESKTOP_VFB_RESOLUTION", "1280x1024")
 PRIVATE_XAUTH = envbool("XPRA_PRIVATE_XAUTH", False)
 XAUTH_PER_DISPLAY = envbool("XPRA_XAUTH_PER_DISPLAY", True)
 
@@ -35,14 +58,13 @@ vfb_logger = None
 def get_vfb_logger():
     global vfb_logger
     if not vfb_logger:
-        from xpra.log import Logger
-        vfb_logger = Logger("server", "x11")
+        vfb_logger = Logger("server", "x11", "screen")
     return vfb_logger
 
 def osclose(fd):
     try:
         os.close(fd)
-    except:
+    except OSError:
         pass
 
 def create_xorg_device_configs(xorg_conf_dir, device_uuid, uid, gid):
@@ -90,11 +112,11 @@ def create_xorg_device_configs(xorg_conf_dir, device_uuid, uid, gid):
 #create individual device files:
 def save_input_conf(xorg_conf_dir, i, dev_type, device_uuid, uid, gid):
     upper_dev_type = dev_type[:1].upper()+dev_type[1:]   #ie: Pointer
-    product_name = "Xpra Virtual %s %s" % (upper_dev_type, device_uuid)
+    product_name = "Xpra Virtual %s %s" % (upper_dev_type, bytestostr(device_uuid))
     identifier = "xpra-virtual-%s" % dev_type
     conf_file = os.path.join(xorg_conf_dir, "%02i-%s.conf" % (i, dev_type))
     with open(conf_file, "wb") as f:
-        f.write(b"""Section "InputClass"
+        f.write(strtobytes("""Section "InputClass"
 Identifier "%s"
 MatchProduct "%s"
 MatchUSBID "ffff:ffff"
@@ -103,7 +125,7 @@ Driver "libinput"
 Option "AccelProfile" "flat"
 Option "Ignore" "False"
 EndSection
-""" % (identifier, product_name, upper_dev_type))
+""" % (identifier, product_name, upper_dev_type)))
         os.fchown(f.fileno(), uid, gid)
     #Option "AccelerationProfile" "-1"
     #Option "AccelerationScheme" "none"
@@ -132,7 +154,7 @@ def get_xauthority_path(display_name, username, uid, gid):
         filename = ".Xauthority"
     return os.path.join(pathexpand(d), filename)
 
-def start_Xvfb(xvfb_str, pixel_depth, display_name, cwd, uid, gid, username, xauth_data, uinput_uuid=None):
+def start_Xvfb(xvfb_str, vfb_geom, pixel_depth, display_name, cwd, uid, gid, username, uinput_uuid=None):
     if not POSIX:
         raise InitException("starting an Xvfb is not supported on %s" % os.name)
     if OSX:
@@ -142,29 +164,19 @@ def start_Xvfb(xvfb_str, pixel_depth, display_name, cwd, uid, gid, username, xau
 
     cleanups = []
     log = get_vfb_logger()
-    log("start_Xvfb%s", (xvfb_str, pixel_depth, display_name, cwd, uid, gid, username, xauth_data, uinput_uuid))
-    xauthority = get_xauthority_path(display_name, username, uid, gid)
-    os.environ["XAUTHORITY"] = xauthority
-    if not os.path.exists(xauthority):
-        log("creating XAUTHORITY=%s with data=%s", xauthority, xauth_data)
-        try:
-            with open(xauthority, 'wa') as f:
-                if getuid()==0 and (uid!=0 or gid!=0):
-                    os.fchown(f.fileno(), uid, gid)
-        except Exception as e:
-            #trying to continue anyway!
-            log.error("Error trying to create XAUTHORITY file %s:", xauthority)
-            log.error(" %s", e)
-    else:
-        log("found existing XAUTHORITY file '%s'", xauthority)
+    log("start_Xvfb%s", (xvfb_str, vfb_geom, pixel_depth, display_name, cwd, uid, gid, username, uinput_uuid))
     use_display_fd = display_name[0]=='S'
 
     subs = {}
     def pathexpand(s):
         return osexpand(s, actual_username=username, uid=uid, gid=gid, subs=subs)
+    etc_prefix = os.environ.get("XPRA_INSTALL_PREFIX", "")
+    if etc_prefix.endswith("/usr"):
+        etc_prefix = etc_prefix[:-4]
     subs.update({
         "DISPLAY"       : display_name,
         "XPRA_LOG_DIR"  : pathexpand(os.environ.get("XPRA_LOG_DIR")),
+        "XORG_CONFIG_PREFIX" : os.environ.get("XORG_CONFIG_PREFIX", etc_prefix),
         })
 
     #identify logfile argument if it exists,
@@ -176,9 +188,26 @@ def start_Xvfb(xvfb_str, pixel_depth, display_name, cwd, uid, gid, username, xau
     #make sure all path values are expanded:
     xvfb_cmd = [pathexpand(s) for s in xvfb_cmd]
 
+    #try to honour initial geometries if specified:
+    if vfb_geom and xvfb_cmd[0].endswith("Xvfb"):
+        #find the '-screen' arguments:
+        #"-screen 0 8192x4096x24+32"
+        try:
+            no_arg = xvfb_cmd.index("0")
+        except ValueError:
+            no_arg = -1
+        geom_str = "%sx%s" % ("x".join(str(x) for x in vfb_geom), pixel_depth)
+        if no_arg>0 and xvfb_cmd[no_arg-1]=="-screen" and len(xvfb_cmd)>(no_arg+1):
+            #found an existing "-screen" argument for this screen number,
+            #replace it:
+            xvfb_cmd[no_arg+1] = geom_str
+        else:
+            #append:
+            xvfb_cmd += ["-screen", "0", geom_str]
     try:
         logfile_argindex = xvfb_cmd.index('-logfile')
-        assert logfile_argindex+1<len(xvfb_cmd), "invalid xvfb command string: -logfile should not be last (found at index %i)" % logfile_argindex
+        if logfile_argindex+1>=len(xvfb_cmd):
+            raise InitException("invalid xvfb command string: -logfile should not be last")
         xorg_log_file = xvfb_cmd[logfile_argindex+1]
     except ValueError:
         xorg_log_file = None
@@ -196,10 +225,10 @@ def start_Xvfb(xvfb_str, pixel_depth, display_name, cwd, uid, gid, username, xau
                 if POSIX and uid!=getuid() or gid!=getgid():
                     try:
                         os.lchown(xorg_log_dir, uid, gid)
-                    except:
-                        pass
+                    except OSError:
+                        log("lchown(%s, %i, %i)", xorg_log_dir, uid, gid, exc_info=True)
             except OSError as e:
-                raise InitException("failed to create the Xorg log directory '%s': %s" % (xorg_log_dir, e))
+                raise InitException("failed to create the Xorg log directory '%s': %s" % (xorg_log_dir, e)) from None
 
     if uinput_uuid:
         #use uinput:
@@ -210,7 +239,8 @@ def start_Xvfb(xvfb_str, pixel_depth, display_name, cwd, uid, gid, username, xau
             log.warn("Warning: cannot use uinput")
             log.warn(" '-config' argument not found in the xvfb command")
         else:
-            assert config_argindex+1<len(xvfb_cmd), "invalid xvfb command string: -config should not be last (found at index %i)" % config_argindex
+            if config_argindex+1>=len(xvfb_cmd):
+                raise InitException("invalid xvfb command string: -config should not be last")
             xorg_conf = xvfb_cmd[config_argindex+1]
             if xorg_conf.endswith("xorg.conf"):
                 xorg_conf = xorg_conf.replace("xorg.conf", "xorg-uinput.conf")
@@ -232,21 +262,19 @@ def start_Xvfb(xvfb_str, pixel_depth, display_name, cwd, uid, gid, username, xau
                 raise InitException("%s: %s" % (xvfb_executable, msg))
             r_pipe, w_pipe = os.pipe()
             try:
-                if PYTHON3:
-                    os.set_inheritable(w_pipe, True)        #@UndefinedVariable
+                os.set_inheritable(w_pipe, True)        #@UndefinedVariable
                 xvfb_cmd += ["-displayfd", str(w_pipe)]
-                xvfb_cmd[0] = "%s-for-Xpra-%s" % (xvfb_executable, display_name)
+                xvfb_cmd[0] = "%s-for-Xpra-%s" % (xvfb_executable, display_name.lstrip(":"))
                 def preexec():
-                    setsid()
+                    os.setpgrp()
                     if getuid()==0 and uid:
                         setuidgid(uid, gid)
-                    close_fds([0, 1, 2, r_pipe, w_pipe])
                 try:
-                    xvfb = subprocess.Popen(xvfb_cmd, executable=xvfb_executable, close_fds=False,
-                                            stdin=subprocess.PIPE, preexec_fn=preexec, cwd=cwd)
+                    xvfb = Popen(xvfb_cmd, executable=xvfb_executable,
+                                 preexec_fn=preexec, cwd=cwd, pass_fds=(w_pipe,))
                 except OSError as e:
                     log("Popen%s", (xvfb_cmd, xvfb_executable, cwd), exc_info=True)
-                    raise InitException("failed to execute xvfb command %s: %s" % (xvfb_cmd, e))
+                    raise InitException("failed to execute xvfb command %s: %s" % (xvfb_cmd, e)) from None
                 assert xvfb.poll() is None, "xvfb command failed"
                 # Read the display number from the pipe we gave to Xvfb
                 try:
@@ -260,7 +288,7 @@ def start_Xvfb(xvfb_str, pixel_depth, display_name, cwd, uid, gid, username, xau
             n = parse_displayfd(buf, displayfd_err)
             new_display_name = ":%s" % n
             log("Using display number provided by %s: %s", xvfb_executable, new_display_name)
-            if tmp_xorg_log_file != None:
+            if tmp_xorg_log_file:
                 #ie: ${HOME}/.xpra/Xorg.${DISPLAY}.log -> /home/antoine/.xpra/Xorg.S14700.log
                 f0 = shellsub(tmp_xorg_log_file, subs)
                 subs["DISPLAY"] = new_display_name
@@ -282,12 +310,10 @@ def start_Xvfb(xvfb_str, pixel_depth, display_name, cwd, uid, gid, username, xau
                 if getuid()==0 and (uid!=0 or gid!=0):
                     setuidgid(uid, gid)
                 else:
-                    setsid()
+                    os.setsid()
             log("xvfb_cmd=%s", xvfb_cmd)
-            xvfb = subprocess.Popen(xvfb_cmd, executable=xvfb_executable, close_fds=True,
-                                    stdin=subprocess.PIPE, preexec_fn=preexec)
-
-        xauth_add(xauthority, display_name, xauth_data, uid, gid)
+            xvfb = Popen(xvfb_cmd, executable=xvfb_executable,
+                         stdin=PIPE, preexec_fn=preexec)
     except Exception as e:
         if xvfb and xvfb.poll() is None:
             log.error(" stopping vfb process with pid %i", xvfb.pid)
@@ -298,32 +324,43 @@ def start_Xvfb(xvfb_str, pixel_depth, display_name, cwd, uid, gid, username, xau
     return xvfb, display_name, cleanups
 
 
-def set_initial_resolution(desktop=False):
+def kill_xvfb(xvfb_pid):
+    log = get_vfb_logger()
+    log.info("killing xvfb with pid %s", xvfb_pid)
+    try:
+        os.kill(xvfb_pid, signal.SIGTERM)
+    except OSError as e:
+        log.info("failed to kill xvfb process with pid %s:", xvfb_pid)
+        log.info(" %s", e)
+    xauthority = os.environ.get("XAUTHORITY")
+    if PRIVATE_XAUTH and xauthority and os.path.exists(xauthority):
+        os.unlink(xauthority)
+
+
+def set_initial_resolution(res=DEFAULT_VFB_RESOLUTION):
     try:
         log = get_vfb_logger()
-        log("set_initial_resolution")
-        if desktop:
-            res = DEFAULT_DESKTOP_VFB_RESOLUTION
-        else:
-            res = DEFAULT_VFB_RESOLUTION
+        log("set_initial_resolution(%s)", res)
         from xpra.x11.bindings.randr_bindings import RandRBindings      #@UnresolvedImport
         #try to set a reasonable display size:
         randr = RandRBindings()
         if not randr.has_randr():
-            l = log
-            if desktop:
-                l = log.warn
-            l("Warning: no RandR support,")
-            l(" default virtual display size unchanged")
+            log.warn("Warning: no RandR support,")
+            log.warn(" default virtual display size unchanged")
             return
         sizes = randr.get_xrr_screen_sizes()
         size = randr.get_screen_size()
         log("RandR available, current size=%s, sizes available=%s", size, sizes)
-        if res in sizes:
+        if res not in sizes:
+            log.warn("Warning: cannot set resolution to %s", res)
+            log.warn(" (this resolution is not available)")
+        elif res==size:
+            log("initial resolution already set: %s", res)
+        else:
             log("RandR setting new screen size to %s", res)
             randr.set_screen_size(*res)
     except Exception as e:
-        log("set_initial_resolution(%s)", desktop, exc_info=True)
+        log("set_initial_resolution(%s)", res, exc_info=True)
         log.error("Error: failed to set the default screen size:")
         log.error(" %s", e)
 
@@ -332,22 +369,24 @@ def xauth_add(filename, display_name, xauth_data, uid, gid):
     xauth_args = ["-f", filename, "add", display_name, "MIT-MAGIC-COOKIE-1", xauth_data]
     try:
         def preexec():
-            setsid()
+            os.setsid()
             if getuid()==0 and uid:
                 setuidgid(uid, gid)
         xauth_cmd = ["xauth"]+xauth_args
         start = monotonic_time()
-        code = subprocess.call(xauth_cmd, preexec_fn=preexec, close_fds=True)
+        log = get_vfb_logger()
+        log("xauth command: %s", xauth_cmd)
+        code = call(xauth_cmd, preexec_fn=preexec)
         end = monotonic_time()
         if code!=0 and (end-start>=10):
-            log = get_vfb_logger()
             log.warn("Warning: xauth command took %i seconds and failed" % (end-start))
             #took more than 10 seconds to fail, check for stale locks:
             import glob
             if glob.glob("%s-*" % filename):
                 log.warn("Warning: trying to clean some stale xauth locks")
                 xauth_cmd = ["xauth", "-b"]+xauth_args
-                code = subprocess.call(xauth_cmd, preexec_fn=preexec, close_fds=True)
+                log("xauth command: %s", xauth_cmd)
+                code = call(xauth_cmd, preexec_fn=preexec)
         if code!=0:
             raise OSError("non-zero exit code: %s" % code)
     except OSError as e:
@@ -357,11 +396,11 @@ def xauth_add(filename, display_name, xauth_data, uid, gid):
         log.error(" using command \"%s\":" % (" ".join(xauth_cmd)))
         log.error(" %s" % (e,))
 
-def check_xvfb_process(xvfb=None, cmd="Xvfb"):
+def check_xvfb_process(xvfb=None, cmd="Xvfb", timeout=0, command=None):
     if xvfb is None:
         #we don't have a process to check
         return True
-    if xvfb.poll() is None:
+    if pollwait(xvfb, timeout) is None:
         #process is running
         return True
     log = get_vfb_logger()
@@ -369,28 +408,48 @@ def check_xvfb_process(xvfb=None, cmd="Xvfb"):
     log.error("%s command has terminated! xpra cannot continue", cmd)
     log.error(" if the display is already running, try a different one,")
     log.error(" or use the --use-display flag")
+    if command:
+        log.error(" full command: %s", command)
     log.error("")
     return False
 
-def verify_display_ready(xvfb, display_name, shadowing_check=True):
+def verify_display_ready(xvfb, display_name, shadowing_check=True, log_errors=True, timeout=VFB_WAIT):
     from xpra.x11.bindings.wait_for_x_server import wait_for_x_server        #@UnresolvedImport
     # Whether we spawned our server or not, it is now running -- or at least
     # starting.  First wait for it to start up:
     try:
-        wait_for_x_server(strtobytes(display_name), VFB_WAIT)
+        wait_for_x_server(strtobytes(display_name), timeout)
     except Exception as e:
         log = get_vfb_logger()
         log("verify_display_ready%s", (xvfb, display_name, shadowing_check), exc_info=True)
-        log.error("Error: failed to connect to display %s" % display_name)
-        log.error(" %s", e)
+        if log_errors:
+            log.error("Error: failed to connect to display %s" % display_name)
+            log.error(" %s", e)
         return False
     if shadowing_check and not check_xvfb_process(xvfb):
         #if we're here, there is an X11 server, but it isn't the one we started!
         log = get_vfb_logger()
-        log.error("There is an X11 server already running on display %s:" % display_name)
-        log.error("You may want to use:")
-        log.error("  'xpra upgrade %s' if an instance of xpra is still connected to it" % display_name)
-        log.error("  'xpra --use-display start %s' to connect xpra to an existing X11 server only" % display_name)
-        log.error("")
+        log("verify_display_ready%s display exists, but the vfb process has terminated",
+                (xvfb, display_name, shadowing_check, log_errors))
+        if log_errors:
+            log.error("There is an X11 server already running on display %s:" % display_name)
+            log.error("You may want to use:")
+            log.error("  'xpra upgrade %s' if an instance of xpra is still connected to it" % display_name)
+            log.error("  'xpra --use-display start %s' to connect xpra to an existing X11 server only" % display_name)
+            log.error("")
         return False
     return True
+
+
+def main():
+    import sys
+    display = None
+    if len(sys.argv)>1:
+        display = strtobytes(sys.argv[1])
+    from xpra.x11.bindings.wait_for_x_server import wait_for_x_server        #@UnresolvedImport
+    wait_for_x_server(display, VFB_WAIT)
+    print("OK")
+
+
+if __name__ == "__main__":
+    main()
