@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2018 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2021 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -9,10 +9,12 @@
 import os
 
 from xpra.os_util import bytestostr, strtobytes, hexstr
-from xpra.util import nonl, typedict, envbool, iround
+from xpra.util import typedict, envbool, iround
 from xpra.gtk_common.error import xswallow, xsync, xlog
+from xpra.scripts.config import parse_bool
 from xpra.x11.x11_server_core import X11ServerCore, XTestPointerDevice
 from xpra.x11.bindings.keyboard_bindings import X11KeyboardBindings #@UnresolvedImport
+from xpra.x11.xsettings_prop import XSettingsTypeInteger, XSettingsTypeString, BLACKLISTED_XSETTINGS
 from xpra.log import Logger
 
 log = Logger("x11", "server")
@@ -49,9 +51,8 @@ class X11ServerBase(X11ServerCore):
         (see XpraServer or DesktopServer for actual implementations)
     """
 
-    def __init__(self, clobber):
-        self.clobber = clobber
-        X11ServerCore.__init__(self)
+    def __init__(self):
+        super().__init__()
         self._default_xsettings = {}
         self._settings = {}
         self.double_click_time = 0
@@ -61,24 +62,32 @@ class X11ServerBase(X11ServerCore):
         self._xsettings_manager = None
         self._xsettings_enabled = False
         self.display_pid = 0
+        self.icc_profile = b""
 
     def do_init(self, opts):
-        X11ServerCore.do_init(self, opts)
-        self._xsettings_enabled = opts.xsettings
+        super().do_init(opts)
+        #the server class sets the default value for 'xsettings_enabled'
+        #it is overriden in the seamless server (enabled by default),
+        #and we let the options have the final say here:
+        self._xsettings_enabled = parse_bool("xsettings", opts.xsettings, self._xsettings_enabled)
+        log("xsettings_enabled(%s)=%s", opts.xsettings, self._xsettings_enabled)
         if self._xsettings_enabled:
             from xpra.x11.xsettings import XSettingsHelper
             self._default_xsettings = XSettingsHelper().get_settings()
             log("_default_xsettings=%s", self._default_xsettings)
             self.init_all_server_settings()
 
+    def save_pid(self):
+        from xpra.scripts.server import _save_int
+        _save_int(b"XPRA_SERVER_gPID", os.getpid())
 
     def init_display_pid(self, pid):
         if pid:
             from xpra.scripts.server import _save_int
-            _save_int(b"_XPRA_SERVER_PID", pid)
-        elif self.clobber:
+            _save_int(b"XPRA_XVFB_PID", pid)
+        else:
             from xpra.scripts.server import _get_int
-            pid = _get_int(b"_XPRA_SERVER_PID")
+            pid = _get_int(b"XPRA_XVFB_PID") or _get_int(b"_XPRA_SERVER_PID")
             if not pid:
                 log.info("xvfb pid not found")
             else:
@@ -93,8 +102,11 @@ class X11ServerBase(X11ServerCore):
             elif self._upgrading:
                 log.info("upgrading: not cleaning up Xvfb")
             else:
+                from xpra.scripts.server import add_cleanup
                 from xpra.x11.vfb_util import kill_xvfb
-                kill_xvfb(self.display_pid)
+                def do_kill_display():
+                    kill_xvfb(self.display_pid)
+                add_cleanup(do_kill_display)
 
 
     def do_cleanup(self):
@@ -111,25 +123,15 @@ class X11ServerBase(X11ServerCore):
 
 
     def init_dbus(self, dbus_pid, dbus_env):
-        from xpra.server.dbus.dbus_start import (
-            get_saved_dbus_pid, get_saved_dbus_env,
-            save_dbus_pid, save_dbus_env,
-            )
-        super(X11ServerBase, self).init_dbus(dbus_pid, dbus_env)
-        if self.clobber:
-            #get the saved pids and env
-            self.dbus_pid = get_saved_dbus_pid()
-            self.dbus_env = get_saved_dbus_env()
-            dbuslog("retrieved existing dbus attributes: %s, %s", self.dbus_pid, self.dbus_env)
-            if self.dbus_env:
-                os.environ.update(self.dbus_env)
-        else:
+        dbuslog("init_dbus(%s, %s)", dbus_pid, dbus_env)
+        if dbus_pid and dbus_env:
+            os.environ.update(dbus_env)
+            self.dbus_pid = dbus_pid
+            self.dbus_env = dbus_env
             #now we can save values on the display
             #(we cannot access gtk3 until dbus has started up)
-            if self.dbus_pid:
-                save_dbus_pid(self.dbus_pid)
-            if self.dbus_env:
-                save_dbus_env(self.dbus_env)
+            from xpra.server.dbus.dbus_start import save_dbus_env
+            save_dbus_env(dbus_env)
 
 
     def last_client_exited(self):
@@ -189,6 +191,22 @@ class X11ServerBase(X11ServerCore):
         self.update_server_settings()
 
 
+    def get_info(self, proto=None, client_uuids=None):
+        info = super().get_info(proto=proto, client_uuids=client_uuids)
+        display_info = info.setdefault("display", {})
+        if self.display_pid:
+            display_info["pid"] = self.display_pid
+        display_info["icc"] = self.get_icc_info()
+        return info
+
+    def get_icc_info(self) -> dict:
+        icc_info = {
+            "sync"  : SYNC_ICC,
+            }
+        if SYNC_ICC:
+            icc_info["profile"] = hexstr(self.icc_profile)
+        return icc_info
+
     def set_icc_profile(self):
         if not SYNC_ICC:
             return
@@ -209,14 +227,9 @@ class X11ServerBase(X11ServerCore):
             return
         screenlog("set_icc_profile() icc data for %s: %s (%i bytes)",
                   ui_clients[0], hexstr(data or ""), len(data or ""))
+        self.icc_profile = data
         from xpra.x11.gtk_x11.prop import prop_set
-        #each CARD32 contains just one 8-bit value - don't ask me why
-        def o(x):
-            try:
-                return ord(x)
-            except:
-                return x
-        prop_set(self.root_window, "_ICC_PROFILE", ["u32"], [o(x) for x in data])
+        prop_set(self.root_window, "_ICC_PROFILE", ["u32"], [ord(x) for x in data])
         prop_set(self.root_window, "_ICC_PROFILE_IN_X_VERSION", "u32", 0*100+4) #0.4 -> 0*100+4*1
 
     def reset_icc_profile(self):
@@ -224,6 +237,7 @@ class X11ServerBase(X11ServerCore):
         from xpra.x11.gtk_x11.prop import prop_del
         prop_del(self.root_window, "_ICC_PROFILE")
         prop_del(self.root_window, "_ICC_PROFILE_IN_X_VERSION")
+        self.icc_profile = b""
 
 
     def reset_settings(self):
@@ -280,12 +294,11 @@ class X11ServerBase(X11ServerCore):
                     log.warn("Warning: failed to parse default XSettings:")
                     log.warn(" %s", e)
         old_settings = dict(self._settings)
-        log("server_settings: old=%s, updating with=%s", nonl(old_settings), nonl(settings))
+        log("server_settings: old=%r, updating with=%r", old_settings, settings)
         log("overrides: dpi=%s, double click time=%s, double click distance=%s",
             dpi, double_click_time, double_click_distance)
         log("overrides: antialias=%s", antialias)
         self._settings.update(settings)
-        from xpra.x11.xsettings_prop import XSettingsTypeInteger, XSettingsTypeString, BLACKLISTED_XSETTINGS
         for k, v in settings.items():
             #cook the "resource-manager" value to add the DPI and/or antialias values:
             if k==b"resource-manager" and (dpi>0 or antialias or cursor_size>0):
@@ -327,7 +340,7 @@ class X11ServerBase(X11ServerCore):
                                    "Xft.hinting"    : ad.intget("hinting", -1),
                                    "Xft.rgba"       : subpixel_order,
                                    "Xft.hintstyle"  : _get_antialias_hintstyle(ad)})
-                log("server_settings: resource-manager values=%s", nonl(values))
+                log("server_settings: resource-manager values=%r", values)
                 #convert the dict back into a resource string:
                 value = ''
                 for vk, vv in values.items():
@@ -389,7 +402,7 @@ class X11ServerBase(X11ServerCore):
                 elif k == b"resource-manager":
                     from xpra.x11.gtk_x11.prop import prop_set
                     p = "RESOURCE_MANAGER"
-                    log("server_settings: setting %s to %s", nonl(p), nonl(v))
+                    log("server_settings: setting %s to %r", p, v)
                     prop_set(self.root_window, p, "latin1", strtobytes(v).decode("latin1"))
                 else:
                     log.warn("Warning: unexpected setting '%s'", bytestostr(k))
